(* 
   file        : $CAT_DIR/ExtBin/ExtBin.ML
   logic image : HOL
   platform    : PolyML 3.2 and Isabelle94-8/HOL
   version     : 1.0b  (16.07.98)
   notes       : CATs arithmetical lemmas and tactics to interface to
                 an external simplifier and a decision procedure.
*)

(* tactics for computer arithmetic *)

local

  open Basics;
  open ExtBinOracle;


(* call external reasoners *)

  fun extbin_oracle sign data =
        invoke_oracle (ExtBin.thy, sign, CompArithOracleExn data);


(* collects all free variables of a term *)

  fun collect_free_vars term =
        case term of
            Free _        => [term]
          | Abs (_, _, u) => collect_free_vars u
          | t1 $ t2       => (collect_free_vars t1) @ (collect_free_vars t2)
          | _             => [];


(* forms abstraction of `term' over each element of `l' successively *)

  fun abstractions_over l term =
        foldl
          (fn (x, y) => case y of
                            Free (nm, T) =>
                              Const ("All", (T --> BOOL) --> BOOL) $
                              Abs (nm, T, abstract_over (y, x))
                          | _ => raise Error "abstraction failed")
          (term, l);


(* removes forall quantifiers; innermost first *)

  fun rm_forall term =
    let
      fun rm_forall' term' =
            case term' of
                Const ("All", _) $ Abs (a, T, u) =>
                  (case variant_abs (a, T, rm_forall' u) of (_, res) => res)
              | Abs (a, T, u) =>
                  Abs (a, T, rm_forall' u)
              | t1 $ t2 =>
                  (rm_forall' t1) $ (rm_forall' t2)
              | _ => term';
    in
      rm_forall' (abstractions_over (collect_free_vars term) term)
    end;


(* outermost quantifiers are removed and a list of free vars is returned *)

  fun rm_quantifiers sign term =
    let
      fun rm_quantifiers' c (Const ("all",_) $ Abs (a,T,u)) =
        let
          val (nm, res) = variant_abs (a,T,u);
        in
            rm_quantifiers' ((cterm_of sign (Free (nm, T)))::c) res
        end
        | rm_quantifiers' c result = (result, c);
    in
      rm_quantifiers' [] term
    end;

(* this tactic put subgoal premises to the object level and the tries
   to solve the subgoal by Shostak's SupInf method, calling the oracle. *)

  fun SUPINF_ASM_TAC i state = 
    let
      val sign   = #sign (rep_thm state);
    in
      case drop (i - 1, prems_of state) of
         []           => Sequence.null
       | subgoal :: _ =>
           let
             val (free, xs) = rm_quantifiers sign subgoal;

             val concl = Logic.strip_imp_concl free;

             val concl' = case concl of
                             Const ("Trueprop", _) $ c => c
                           | _ => raise error "supinf_tac failed";

             val prems = Logic.strip_imp_prems free;

             fun mk_impl (p, c) =
               let
                 val p' = case p of
                              Const ("op ==>", _) $ p1 $ p2 => 
                                mk_impl (p1, p2)
                            | Const ("Trueprop", _) $ p1 =>
                                p1
                            | _ => p;

                 val c' = case c of
                              Const ("op ==>", _) $ c1 $ c2 => 
                                mk_impl (c1, c2)
                            | Const ("Trueprop", _) $ c1 =>
                                c1
                            | _ => c;
               in
                 Const ("op -->", BOOL --> BOOL --> BOOL) $ p' $ c'
               end;

             fun mk_conj (c1, c2) =
               let
                 val c1' = case c1 of
                               Const ("Trueprop", _) $ c =>
                                c
                             | _ => c1;
 
                 val c2' = case c2 of
                               Const ("op ==>", _) $ c3 $ c4 => 
                                mk_impl (c3, c4)
                             | Const ("Trueprop", _) $ c3 =>
                                c3
                             | _ => c2;
               in
                 Const ("op &", BOOL --> BOOL --> BOOL) $ c1' $ c2'
               end;

             val data =
                   if prems = []
                     then
                       Const ("Trueprop", BOOL --> PROP) $ concl'
                     else
                       Const ("Trueprop", BOOL --> PROP) $
                        (Const ("op -->", BOOL --> BOOL --> BOOL) $
                           (foldl mk_conj (hd prems, tl prems)) $
                            concl');

             val data' = rm_forall data;

             val oracle_result =
                   extbin_oracle sign (SupInfData (data', concl))
                   handle Error s => error s;

             val result = forall_elim_vars
                             0
                            (forall_intr_list xs oracle_result);
           in
             (rtac result i) state
           end
  end;


(* this tactic put subgoal premises to the object level and the tries
   to solve the subgoal by Shostak's SupInf method, calling the oracle. *)

  fun SUPINF_TAC i state = 
    let
      val sign   = #sign (rep_thm state);
    in
      case drop (i - 1, prems_of state) of
         []           => Sequence.null
       | subgoal :: _ =>
           let
             val (free, xs) = rm_quantifiers sign subgoal;

             val concl = Logic.strip_imp_concl free;

             val concl' = rm_forall concl;

             val oracle_result =
                   extbin_oracle sign (SupInfData (concl', concl))
                   handle Error s => error s;

             val result = forall_elim_vars
                             0
                            (forall_intr_list xs oracle_result);
           in
             (rtac result i) state
           end
  end;


(* auxiliary theorem *)

  val swapped_mp = prove_goal HOL.thy "[| P; P --> Q |] ==> Q"
        (fn p => [cut_facts_tac p 1, Fast_tac 1]);

in

(* this tactic decides the validity of an expression in quantifier-free
   (extended) Presburger arithmetic using Shostak's method (see JACM 24(4),
   1977, pp. 529-543 and also JACM 26(2), 1979, pp. 351-360);
   *all* assumptions are used! this might have unwanted effects.  *)

  fun supinf_asm_tac i state =
        (TRY (EVERY [SUPINF_ASM_TAC i,
                     rtac TrueI i])) state;

  fun supinf_tac i state =
        (TRY (EVERY [SUPINF_TAC i,
                     rtac TrueI i])) state;


(* this tactic rewrites the integer arithmetic part of a subgoal;
   assumptions are not modified *)

  fun extbin_simp_tac i state = 
    let
      val sign = #sign (rep_thm state);
    in
      case drop (i - 1, prems_of state) of
         []           => Sequence.null
       | subgoal :: _ =>
           let
             val (free, xs) = rm_quantifiers sign subgoal;
             val concl      = Logic.strip_imp_concl free;

             val oracle_result = extbin_oracle sign (SimpData concl)
                   handle Error s => error s;

             val result = forall_elim_vars
                             0
                            (forall_intr_list xs oracle_result);
           in
             (rtac result i) state
           end

  end;

  fun extbin_asm_simp_tac n = 
    (REPEAT (etac swapped_mp n)) THEN
    (extbin_simp_tac n) THEN
    (REPEAT (rtac impI n));

  fun extbin_ss_tac ss i = (EVERY' [extbin_asm_simp_tac, asm_simp_tac ss] i);

  fun extbin_tac i = (EVERY' [extbin_asm_simp_tac, Asm_simp_tac] i);

end;


(* ------------------------------------------------------------------------- *)

(* Instantiating the Simplifier *)

Addsimps [zeqPlus_def,
	  zeqMinus_def,
	  zbool_def,
	  zsub_def,
	  zpositive_def,
	  zmagnitude_def,
	  zless_def,
	  zle_def,
	  zgt_def,
	  zge_def,
	  zmod_def,
	  zxor_def,
	  zor_def,
	  zasl_def,
	  zrol_def];

(* ------------------------------------------------------------------------- *)

(* Proof of basic properties *)

qed_goal "zbool_inj" ExtBin.thy 
    "inj(zbool)"
    (fn _ => [rtac injI 1, Asm_full_simp_tac 1]);

qed_goal "normBcons_eq'" ExtBin.thy 
    "! b. normBcons x b = Bcons x b"
    (fn _ => [induct_tac "x" 1,
	      TRYALL ((rtac allI) THEN' (case_tac "b")),
	      ALLGOALS Asm_simp_tac]);

qed_goal "normBcons_eq" ExtBin.thy
    "normBcons x b = Bcons x b"
    (fn _ => [simp_tac (!simpset addsimps [normBcons_eq']) 1]);

qed_goal "normint_eq" ExtBin.thy
    "normint w = w"
    (fn _ => [induct_tac "w" 1,
	      (REPEAT o (asm_simp_tac (!simpset addsimps 
				       [normBcons_eq']))) 1]);
Addsimps [normint_eq];

qed_goal "znat_ZeroCongruence" ExtBin.thy 
    "znat 0 = #0"
    (fn _ => [Simp_tac 1]);

qed_goal "znat_SucCongruence" ExtBin.thy 
    "znat (Suc x) = zsuc (znat x)"
    (fn _ => [Simp_tac 1]);

fun CAT_internal_induct x b =
  (induct_tac x 1) THEN
  (case_tac b 3) THEN
  (ALLGOALS Asm_simp_tac);

qed_goal "zsuc_inj'" ExtBin.thy 
    "! x y. zsuc x = zsuc y --> x = y"
    (fn _ => [rtac allI 1,
	      CAT_internal_induct "x" "bool",
	      rtac allI 1,
	      CAT_internal_induct "y" "bool",
	      rtac impI 1,
	      hyp_subst_tac 1,
	      Simp_tac 1,
	      rtac allI 1,
	      CAT_internal_induct "y" "bool",
	      rtac allI 1,
	      CAT_internal_induct "y" "boola",
	      eres_inst_tac [("x","inta")] allE 1,
	      atac 1,
	      rtac allI 1,
	      CAT_internal_induct "y" "boola"]);
Addsimps [zsuc_inj'];

qed_goal "zsuc_inj" ExtBin.thy 
    "inj(zsuc)"
    (fn _ => [rtac injI 1,
	      cut_facts_tac [zsuc_inj'] 1,
	      eres_inst_tac [("x","x")] allE 1,
	      eres_inst_tac [("x","y")] allE 1,
	      Fast_tac 1]);

qed_goal "zsuc_zadd_def" ExtBin.thy 
    "zsuc x = x + #1"
    (fn _ => [CAT_internal_induct "x" "bool",
	      CAT_internal_induct "int" "b"]);

qed_goal "zpred_zsub_def" ExtBin.thy 
    "zpred x = x - #1"
    (fn _ => [CAT_internal_induct "x" "bool"]);

qed_goal "zsuc_inverse" ExtBin.thy 
    "zsuc (zpred x) = x"
    (fn _ => [CAT_internal_induct "x" "bool"]);
Addsimps [zsuc_inverse];

qed_goal "zpred_inverse" ExtBin.thy 
    "zpred (zsuc x) = x"
    (fn _ => [CAT_internal_induct "x" "bool"]);
Addsimps [zpred_inverse];

qed_goal "zpred_inj'" ExtBin.thy 
    "! x y. zpred x = zpred y --> x = y"
    (fn _ => [rtac allI 1,
	      CAT_internal_induct "x" "bool",
	      rtac allI 1,
	      CAT_internal_induct "y" "bool",
	      rtac allI 1,
	      CAT_internal_induct "y" "bool",
	      rtac impI 1,
	      hyp_subst_tac 1,
	      Simp_tac 1,
	      rtac allI 1,
	      CAT_internal_induct "y" "boola",
	      rtac allI 1,
	      CAT_internal_induct "y" "boola",
	      eres_inst_tac [("x","inta")] allE 1,
	      atac 1]);
Addsimps [zpred_inj'];

qed_goal "zpred_inj" ExtBin.thy 
    "inj(zpred)"
    (fn _ => [rtac injI 1,
	      cut_facts_tac [zpred_inj'] 1,
	      eres_inst_tac [("x","x")] allE 1,
	      eres_inst_tac [("x","y")] allE 1,
	      Fast_tac 1]);

qed_goal "zpred_zminus" ExtBin.thy 
    "zpred (zminus x) = zminus (zsuc x)"
    (fn _ => [CAT_internal_induct "x" "bool"]);

qed_goal "zsuc_zminus" ExtBin.thy 
    "zsuc (zminus x) = zminus (zpred x)"
    (fn _ => [CAT_internal_induct "x" "bool",
              asm_simp_tac (!simpset addsimps [zpred_zminus]) 1]);
Addsimps [zpred_zminus,
	  zsuc_zminus];

qed_goal "zminus_inverse" ExtBin.thy 
    "zminus (zminus x) = x"
    (fn _ => [CAT_internal_induct "x" "bool"]);
Addsimps [zminus_inverse];

qed_goal "zminus_inj'" ExtBin.thy 
    "! x y. zminus x = zminus y --> x = y"
    (fn _ => [Step_tac 1,
	      res_inst_tac [("P","% w. x = w")] subst 1,
	      rtac zminus_inverse 1,
	      res_inst_tac [("P","% w. w = zminus (zminus y)")] subst 1,
	      rtac zminus_inverse 1,
	      etac arg_cong 1]);
Addsimps [zminus_inj'];

qed_goal "zminus_inj" ExtBin.thy 
    "inj(zminus)"
    (fn _ => [rtac injI 1,
	      cut_facts_tac [zminus_inj'] 1,
	      eres_inst_tac [("x","x")] allE 1,
	      eres_inst_tac [("x","y")] allE 1,
	      Fast_tac 1]);

qed_goal "zadd_Zero" ExtBin.thy 
    "! x. x + #0 = x"
    (fn _ => [rtac allI 1,
	      CAT_internal_induct "x" "bool"]);
Addsimps [zadd_Zero];

qed_goal "zadd_inverse" ExtBin.thy
    "! x. x + zminus x = #0"
    (fn _ => [rtac allI 1,
	      CAT_internal_induct "x" "bool"]);
Addsimps [zadd_inverse];

(* ------------------------------------------------------------------------- *)

(* Axioms in the Calculation Engine *)

(* CEAxiom0 is the general case which state that znat preserve operations
   on natural numbers: the lemmas are proved on the top of this file       *)
val CEAxiom1 = store_thm("CEAxiom1"    , zaddPlus);
val CEAxiom2 = store_thm("CEAxiom2"    , zadd_Zero);
qed_goal "CEAxiom3" ExtBin.thy
    "x = y ==> x + y = #2 * x"
    (fn p => [cut_facts_tac p 1,
	      Asm_simp_tac 1,
	      CAT_internal_induct "y" "bool",
	      res_inst_tac [("s","zsuc (Bcons int False)")] trans 1,
	      res_inst_tac [("s","zsuc (int + int)")] trans 1,
	      Asm_simp_tac 2,
	      Simp_tac 2,
	      CAT_internal_induct "int" "boola"]);
qed_goal "CEAxiom4" ExtBin.thy
    "x = y ==> x + $~ y = #0"
    (fn p => [cut_facts_tac p 1, 
	      Asm_simp_tac 1]);
qed_goal "CEAxiom5" ExtBin.thy
    "x = y ==> $~ x + y = #0"
    (fn p => [cut_facts_tac p 1, 
	      res_inst_tac [("s","$~ x + $~ $~ y")] trans 1,
	      Simp_tac 1,
	      rtac CEAxiom4 1,
	      Fast_tac 1]);
qed_goal "CEAxiom6" ExtBin.thy 
    "x = y ==> $~ x + $~ y = #~2 * x"
    (fn p => [cut_facts_tac p 1,
	      subgoal_tac "$~ x = $~ y" 1,
	      Fast_tac 2,
	      subgoal_tac "#~2 * x = #2 * $~ x" 1,
	      Simp_tac 2,
	      rotate_tac 1 1,
	      etac (make_elim CEAxiom3) 1,
	      Asm_full_simp_tac 1]);
qed_goal "CEAxiom7" ExtBin.thy 
    "x = y ==> x + (y + z) = #2 * x + z"
    (fn p => [cut_facts_tac p 1,
	      rtac trans 1,
	      rtac zaddAssoc 1,
	      subgoal_tac "x + y = #2 * x" 1,
	      etac CEAxiom3 2,
	      Asm_full_simp_tac 1]);
qed_goal "CEAxiom8" ExtBin.thy 
    "x = y ==> x + ($~ y + z) = z"
    (fn p => [cut_facts_tac p 1,
	      res_inst_tac [("s","(x + $~ y) + z")] trans 1,
	      rtac zaddAssoc 1,
	      res_inst_tac [("s","#0"),("t","x + $~ y")] subst 1,
	      etac (CEAxiom4 RS sym) 1,
	      Simp_tac 1]);
qed_goal "CEAxiom9" ExtBin.thy 
    "x = y ==> ($~ x) + (y + z) = z"
    (fn p => [cut_facts_tac p 1,
	      res_inst_tac [("s","($~ x + y) + z")] trans 1,
	      rtac zaddAssoc 1,
	      res_inst_tac [("s","#0"),("t","$~ x + y")] subst 1,
	      etac (CEAxiom5 RS sym) 1,
	      Simp_tac 1]);
qed_goal "CEAxiom10" ExtBin.thy 
    "x = y ==> ($~ x) + ($~ y + z) = #~2 * x + z"
    (fn p => [cut_facts_tac p 1,
	      res_inst_tac [("s","($~ x + $~ y) + z")] trans 1,
	      rtac zaddAssoc 1,
	      res_inst_tac [("s","#~2 * x"),("t","$~ x + $~ y")] subst 1,
	      etac (CEAxiom6 RS sym) 1,
	      Simp_tac 1]);
qed_goal "CEAxiom11" ExtBin.thy
    "(x::int) = y ==> n * x + m * y = (n + m) * x"
    (fn p => [cut_facts_tac p 1,
	      res_inst_tac [("s","x * (n + m)")] trans 1,
	      rtac zmultComm 2,
	      res_inst_tac [("s","x * n + y * m")] trans 1,
	      res_inst_tac [("s","x * n + m * y")] trans 1,
	      rtac (zmultComm RS arg_cong) 1,
	      rtac (zmultComm RS arg_cong) 1,
	      Asm_simp_tac 1,
	      rtac (zmultDistrib RS sym) 1]);
qed_goal "CEAxiom12" ExtBin.thy
     "(x::int) = y ==> n * x + (m * y + z) = (n + m) * x + z"
     (fn p => [res_inst_tac [("s","(n * x + m * y) + z")] trans 1,
	       rtac zaddAssoc 1,
	       rtac (CEAxiom11 RS arg_cong) 1,
	       rtac (hd p) 1]);
qed_goal "CEAxiom13" ExtBin.thy
    "x = y ==> x + n * y = (#1 + n) * x"
    (fn p => [res_inst_tac [("s","#1 * x + n * y")] trans 1,
	      Simp_tac 1,
	      rtac CEAxiom11 1,
	      rtac (hd p) 1]);
qed_goal "CEAxiom14" ExtBin.thy
     "x = y ==> x + (n * y + z) = (#1 + n) * x + z"
     (fn p => [res_inst_tac [("s","#1 * x + (n * y + z)")] trans 1,
	       Simp_tac 1,
	       rtac CEAxiom12 1,
	       rtac (hd p) 1]);
qed_goal "CEAxiom15" ExtBin.thy
     "x = y ==> n * x + y = (n + #1) * x"
     (fn p => [res_inst_tac [("s","n * x + #1 * y")] trans 1,
	       Simp_tac 1,
	       rtac CEAxiom11 1,
	       rtac (hd p) 1]);
qed_goal "CEAxiom16" ExtBin.thy
     "x = y ==> n * x + (y + z) = (n + #1) * x + z"
     (fn p => [res_inst_tac [("s","n * x + (#1 * y + z)")] trans 1,
	       Simp_tac 1,
	       rtac CEAxiom12 1,
	       rtac (hd p) 1]);
val CEAxiom17 = store_thm("CEAxiom17"   , zsub_def);
val CEAxiom20 = store_thm("CEAxiom20"   , zminus_inverse);
(* CEAxiom21 is performed automatically in the representation *)
val CEAxiom30 = store_thm("CEAxiom30"   , zmultPlus);
qed_goal "CEAxiom31" ExtBin.thy 
    "x * #0 = #0"
    (fn _ => [CAT_internal_induct "x" "bool"]);
qed_goal "CEAxiom32" ExtBin.thy 
    "#1 * x = x"
    (fn _ => [Simp_tac 1]);
qed_goal "CEAxiom33" ExtBin.thy 
    "x * #1 = x"
    (fn _ => [CAT_internal_induct "x" "bool"]);
val CEAxiom34 = store_thm("CEAxiom34"   , zmultMinus);
qed_goal "CEAxiom35" ExtBin.thy 
    "x * #~1 = $~ x"
    (fn _ => [CAT_internal_induct "x" "bool"]);
(* store_thm("CEAxiom40"   , zaddPlus);
val p = goal ExtBin.thy 
    "[| x = y; y ~= #0 |] ==> x zdiv y = #1";
by (cut_facts_tac p 1);
by (Asm_simp_tac 1);
by (subgoal_tac "y < #0 | y = #0 | #0 < y" 1);
by (etac disjE 1);
by (etac disjE 2);
by (Fast_tac 2);

by (res_inst_tac [("n","#1"),("P","% x. x zdiv x = #1")] intInduct 1);

by (induct_tac "y" 1);
by (Simp_tac 1);
by (Simp_tac 1);
by (case_tac "bool" 1);
by (Asm_simp_tac 1);
*)
(* store_thm("CEAxiom41"   , zaddPlus); *)
(* store_thm("CEAxiom42"   , zaddPlus); *)
(* store_thm("CEAxiom43"   , zaddPlus); *)
(* store_thm("CEAxiom44"   , zaddPlus); *)
(* store_thm("CEAxiom45"   , zaddPlus); *)
(* store_thm("CEAxiom46"   , zaddPlus);
goal thy "x zdiv #1 = x";
by (rtac intInduct 1);
by (rtac zdivZero_def 1);
by (Simp_tac 1);

*)
(* store_thm("CEAxiom47"   , zaddPlus); *)
(* store_thm("CEAxiom48"   , zaddPlus); *)
(* store_thm("CEAxiom50"   , zaddPlus); *)
(* store_thm("CEAxiom51"   , zaddPlus); *)
(* store_thm("CEAxiom52"   , zaddPlus); *)
(* store_thm("CEAxiom53"   , zaddPlus); *)
(* store_thm("CEAxiom54"   , zaddPlus); *)
(* store_thm("CEAxiom55"   , zaddPlus); *)
(* store_thm("CEAxiom56"   , zaddPlus); *)
(* store_thm("CEAxiom57"   , zaddPlus); *)
(* store_thm("CEAxiom58"   , zaddPlus); *)
(* store_thm("CEAxiom59"   , zaddPlus); *)
(* store_thm("CEAxiom60"   , zaddPlus); *)
(* store_thm("CEAxiom61"   , zaddPlus); *)
(* store_thm("CEAxiom62"   , zaddPlus); *)
val CEAxiom70 = store_thm("CEAxiom70"   , zexpZero_def);
val CEAxiom71 = store_thm("CEAxiom71"   , zexpEOne_def);
val CEAxiom72 = store_thm("CEAxiom72"   , zexpOne_def);
qed_goal "CEAxiom73" ExtBin.thy 
    "[| n < #0; x ~= #0 |] ==> x zexp n = #0"
    (fn p => [cut_facts_tac p 1,
	      res_inst_tac [("t","x zexp n")] ssubst 1,
	      rtac zexp_def 1,
	      Asm_full_simp_tac 1]);
(* store_thm("CEAxiom74"   , zaddPlus);
val p = goal ExtBin.thy
    "x = y ==> x zexp z * y zexp z' = x zexp (z + z')";
by (cut_facts_tac p 1);
by (Asm_simp_tac 1);
 *)
(* store_thm("CEAxiom75"   , zaddPlus); *)
(* store_thm("CEAxiom76"   , zaddPlus); *)
(* store_thm("CEAxiom77"   , zaddPlus); *)
val CEAxiom100  = store_thm("CEAxiom100"  , zaddAssoc RS sym);
val CEAxiom100a = store_thm("CEAxiom100a" , zaddAssoc);
val CEAxiom101  = store_thm("CEAxiom101"  , zmultAssoc);
val CEAxiom102  = store_thm("CEAxiom102"  , zaddComm);
qed_goal "CEAxiom103" ExtBin.thy
    "(x::int) + (y + z) = y + (x + z)"
    (fn _ => [res_inst_tac [("s","(x + y) + z")] trans 1,
	      rtac zaddAssoc 1,
	      res_inst_tac [("s","(y + x) + z")] trans 1,
	      rtac (zaddComm RS arg_cong) 1,
	      rtac (zaddAssoc RS sym) 1]);
val CEAxiom104 = store_thm("CEAxiom104"  , zmultComm);
qed_goal "CEAxiom105" ExtBin.thy
    "(x::int) * (y * z) = y * (x * z)"
    (fn _ => [res_inst_tac [("s","(x * y) * z")] trans 1,
	      rtac zmultAssoc 1,
	      res_inst_tac [("s","(y * x) * z")] trans 1,
	      rtac (zmultComm RS arg_cong) 1,
	      rtac (zmultAssoc RS sym) 1]);
val CEAxiom106 = store_thm("CEAxiom106"  , zmultDistrib);
qed_goal "CEAxiom107" ExtBin.thy
    "(y + z) * (x::int) = y * x + z * x"
    (fn _ => [rtac (CEAxiom11 RS sym) 1,
	      rtac refl 1]);
qed_goal "CEAxiom108" ExtBin.thy
    "! x y. $~ (x + y) = $~ x + $~ y"
    (fn _ => [rtac allI 1,
(*
	      CAT_internal_induct "x" "bool",
	      rtac allI 1,
	      CAT_internal_induct "y" "bool",
	      CAT_internal_induct "int" "boola",
	      rtac allI 1,
	      CAT_internal_induct "y" "boola",
	      CAT_internal_induct "int" "boola",
*)
	      induct_tac "x" 1,
	      Simp_tac 1,
	      Simp_tac 1,
	      rtac allI 1,
	      induct_tac "y" 1,
	      Simp_tac 1,
	      Simp_tac 1,
	      case_tac "bool" 1,
	      Asm_simp_tac 1,
	      Asm_simp_tac 1,
	      res_inst_tac [("s","zpred ($~ zpred int)")] trans 1,
	      Asm_simp_tac 1,
	      Simp_tac 1,
	      case_tac "bool" 1,
	      rtac allI 1,
	      induct_tac "y" 1,
	      Simp_tac 1,
	      Asm_simp_tac 1,
	      res_inst_tac [("s","zpred ($~ int) + #1")] trans 1,
	      Simp_tac 2,
	      res_inst_tac [("s","($~ int - #1) + #1")] trans 1,
	      rtac (zpred_zsub_def RS arg_cong RS sym) 2,
	      res_inst_tac [("s","$~ int + (#~1 + #1)")] trans 1,
	      Simp_tac 1,
	      rewtac zsub_def,
	      res_inst_tac [("s","$~ int + ($~ #1 + #1)")] trans 1,
	      Simp_tac 1,
	      rtac zaddAssoc 1,
	      case_tac "boola" 1,
	      Asm_simp_tac 1,
	      res_inst_tac [("s","$~ int + zpred ($~ inta)")] trans 1,
	      Simp_tac 1,
	      res_inst_tac [("s","$~ int + ($~ inta - #1)")] trans 1,
	      rtac (zpred_zsub_def RS arg_cong) 1,
	      res_inst_tac [("s","$~ int + ($~ inta + $~ #1)")] trans 1,
	      Simp_tac 1,
	      res_inst_tac [("s","$~ int + ($~ #1 + $~ inta)")] trans 1,
	      rtac (zaddComm RS arg_cong) 1,
	      res_inst_tac [("s","($~ int + $~ #1) + $~ inta")] trans 1,
	      rtac zaddAssoc 1,
	      res_inst_tac [("s","($~ int - #1) + $~ inta")] trans 1,
	      Simp_tac 1,
	      res_inst_tac [("s","zpred ($~ int) + $~ inta")] trans 1,
	      rtac (zpred_zsub_def RS arg_cong RS sym) 1,
	      Simp_tac 1,
	      Asm_simp_tac 1,
	      res_inst_tac [("s","($~ int + $~ inta) - #1")] trans 1,
	      rtac zpred_zsub_def 1,
	      res_inst_tac [("s","($~ int + $~ inta) + $~ #1")] trans 1,
	      Simp_tac 1,
	      res_inst_tac [("s","$~ int + ($~ inta + $~ #1)")] trans 1,
	      rtac (zaddAssoc RS sym) 1,
	      res_inst_tac [("s","$~ int + ($~ #1 + $~ inta)")] trans 1,
	      rtac (zaddComm RS arg_cong) 1,
	      res_inst_tac [("s","($~ int + $~ #1) + $~ inta")] trans 1,
	      rtac zaddAssoc 1,
	      res_inst_tac [("s","($~ int - #1) + $~ inta")] trans 1,
	      Simp_tac 1,
	      res_inst_tac [("s","zpred ($~ int) + $~ inta")] trans 1,
	      rtac (zpred_zsub_def RS arg_cong RS sym) 1,
	      Simp_tac 1,
	      rtac allI 1,
	      induct_tac "y" 1,
	      Simp_tac 1,
	      Asm_simp_tac 1,
	      case_tac "boola" 1,
	      Asm_simp_tac 1,
	      res_inst_tac [("s","($~ int + $~ inta) - #1")] trans 1,
	      rtac zpred_zsub_def 1,
	      res_inst_tac [("s","($~ int + $~ inta) + $~ #1")] trans 1,
	      Simp_tac 1,
	      res_inst_tac [("s","$~ int + ($~ inta + $~ #1)")] trans 1,
	      rtac (zaddAssoc RS sym) 1,
	      res_inst_tac [("s","$~ int + ($~ inta - #1)")] trans 1,
	      Simp_tac 1,
	      res_inst_tac [("s","$~ int + zpred ($~ inta)")] trans 1,
	      rtac (zpred_zsub_def RS arg_cong RS sym) 1,
	      Simp_tac 1,
	      Asm_simp_tac 1]);
(* store_thm("CEAxiom109"  , zaddPlus);
goal ExtBin.thy "! x y. $~ (x * y) = ($~ x) * y";
by (rtac allI 1);
by (induct_tac "x" 1);
by (Simp_tac 1);
by (Simp_tac 1);
by (rtac allI 1);
by (induct_tac "y" 1);
by (asm_simp_tac (!simpset addsimps [CEAxiom31]) 1);
by (case_tac "bool" 1);
by (Asm_simp_tac 1);
by (res_inst_tac [("s","$~ int * #~1 + (#1 + #~1)")] trans 1);
by (Simp_tac 1);
by (res_inst_tac [("s","$~ int * #~1 + #1 + #~1")] trans 1);
by (rtac zaddAssoc 1);
by (res_inst_tac [("s","$~ int * #~1 + #~1 * #~1 + #~1")] trans 1);
by (Simp_tac 1);
by (res_inst_tac [("s","($~ int + #~1) * #~1 + #~1")] trans 1);
by (rtac (CEAxiom107 RS sym RS arg_cong) 1);
by (res_inst_tac [("s","($~ int - #1) * #~1 + #~1")] trans 1);
by (Simp_tac 1);
by (res_inst_tac [("s","zpred ($~ int) * #~1 + #~1")] trans 1);
by (rtac (zpred_zsub_def RS sym RS arg_cong) 1);
by (res_inst_tac [("s","$~ zsuc int * #~1 + #~1")] trans 1);
by (Simp_tac 1);
by (res_inst_tac [("s","$~ zsuc int * #~1 - #1")] trans 1);
by (Simp_tac 1);
by (rtac (zpred_zsub_def RS sym) 1);
by (Asm_simp_tac 1);

by (case_tac "boola" 1);
by (case_tac "bool" 1);
by (rotate_tac ~2 1);
by (Asm_full_simp_tac 1);

by (Asm_simp_tac 1);
by (res_inst_tac [("s","$~ zsuc (Bcons inta True * int + inta)")] trans 1);
by (rtac (zmultComm RS arg_cong) 1);
by (res_inst_tac [("s","Bcons inta True * $~ zsuc int + inta")] trans 1);
by (rtac (zmultComm RS arg_cong) 2);

*)
(* store_thm("CEAxiom110"  , zaddPlus); *)
(* store_thm("CEAxiom111"  , zaddPlus); *)
(* store_thm("CEAxiom112"  , zaddPlus); *)
(* store_thm("CEAxiom113"  , zaddPlus); *)
(* store_thm("CEAxiom114"  , zaddPlus); *)
(* store_thm("CEAxiomE1"   , zaddPlus);
qed_goal "CEAxiomE1" ExtBin.thy
    "((x::int) = y) = ((x <= y) & (y <= x))"
    (fn _ => [rtac iffI 1,
	      Asm_simp_tac 1,
	      Asm_full_simp_tac 1,
--->
	      ]);
*)
qed_goal "CEAxiomE2" ExtBin.thy
    "(x >= y) = (y <= x)"
    (fn _ => [Simp_tac 1]);
(* store_thm("CEAxiomE3"   , zaddPlus);
val p = goal ExtBin.thy
    "x < y ==> (x + #1) <= y";
by (cut_facts_tac p 1);
 
*)
(* store_thm("CEAxiomE4"   , zaddPlus); *)
qed_goal "CEAxiomF1" ExtBin.thy
    "(x & True) = x"
    (fn _ => [Fast_tac 1]);
qed_goal "CEAxiomF2" ExtBin.thy
    "(True & x) = x"
    (fn _ => [Fast_tac 1]);
qed_goal "CEAxiomF3" ExtBin.thy
    "(x & False) = False"
    (fn _ => [Fast_tac 1]);
qed_goal "CEAxiomF4" ExtBin.thy
    "(False & x) = False"
    (fn _ => [Fast_tac 1]);
qed_goal "CEAxiomF5" ExtBin.thy
    "(x | False) = x"
    (fn _ => [Fast_tac 1]);
qed_goal "CEAxiomF6" ExtBin.thy
    "(False | x) = x"
    (fn _ => [Fast_tac 1]);
qed_goal "CEAxiomF7" ExtBin.thy
    "(x | True) = True"
    (fn _ => [Fast_tac 1]);
qed_goal "CEAxiomF8" ExtBin.thy
    "(True | x) = True"
    (fn _ => [Fast_tac 1]);
qed_goal "CEAxiomF10" ExtBin.thy
    "(x --> y) = (~x | y)"
    (fn _ => [Fast_tac 1]);
qed_goal "CEAxiomF20" ExtBin.thy
    "(~~x) = x"
    (fn _ => [Fast_tac 1]);
(* store_thm("CEAxiomF21"  , zaddPlus);
qed_goal "CEAxiomF21" ExtBin.thy
    "! x y. (~(x <= y)) = (zsuc y <= x)"
    (fn _ => [Simp_tac 1,
	      rtac allI 1,
	      induct_tac "x" 1,
	      Simp_tac 1,
--->
	      ]);
*)
qed_goal "CEAxiomF100" ExtBin.thy
    "(x & (y & z)) = ((x & y) & z)"
    (fn _ => [Fast_tac 1]);
qed_goal "CEAxiomF101" ExtBin.thy
    "(x | (y | z)) = ((x | y) | z)"
    (fn _ => [Fast_tac 1]);
qed_goal "CEAxiomF102" ExtBin.thy
    "(x & (y | z)) = ((x & y) | (x & z))"
    (fn _ => [Fast_tac 1]);
qed_goal "CEAxiomF103" ExtBin.thy
    "((y | z) & x) = ((y & x) | (z & x))"
    (fn _ => [Fast_tac 1]);
qed_goal "CEAxiomF104" ExtBin.thy
    "(~(x & y)) = (~x | ~ y)"
    (fn _ => [Fast_tac 1]);
qed_goal "CEAxiomF105" ExtBin.thy
    "(~(x | y)) = (~x & ~ y)"
    (fn _ => [Fast_tac 1]);
(* store_thm("CEAxiomF106" , zaddPlus); *)
(* store_thm("CEAxiomF107" , zaddPlus); *)
(* store_thm("CEAxiomF108" , zaddPlus); *)


(* ------------------------------------------------------------------------- *)

(* Tests *)

(*
goal ExtBin.thy "#13  +  #19 = #32";
by (Simp_tac 1);
result();

goal ExtBin.thy "#1234  +  #5678 = #6912";
by (Simp_tac 1);
result();

goal ExtBin.thy "#1359  +  #~2468 = #~1109";
by (Simp_tac 1);
result();

goal ExtBin.thy "#93746 +  #~46375 = #47371";
by (Simp_tac 1);
result();

goal ExtBin.thy "$~ #65745 = #~65745";
by (Simp_tac 1);
result();

goal ExtBin.thy "#13  *  #19 = #247";
by (Simp_tac 1);
result();

goal ExtBin.thy "#~84  *  #51 = #~4284";
by (Simp_tac 1);
result();

goal ExtBin.thy "#1359  *  #~2468 = #~3354012";
by (Simp_tac 1);
result();

goal ExtBin.thy "#13 < #18 - #4";  
by (Simp_tac 1); 
result();

goal ExtBin.thy "#~345 < #~242 + #~100";  
by (Simp_tac 1); 
result();

goal ExtBin.thy "#13557456 < #18678654";  
by (Simp_tac 1); 
result();

*)

(* ------------------------------------------------------------------------- *)

(* Delete rules from the simplifier, so not to make visible the internal 
   translations we do: these definitions are useful ONLY to certify the
   validity of the rules we use in our external reasoners.

   They are available in the ExtBin_ss simplifier set, anyway            *)

val ExtBin_ss = !simpset;

Delsimps [normintPlus,
	  normintMinus,
	  normintBcons,
	  normPlus,
	  normMinus,
	  normBcons,
	  zsucBcons,
	  zpredBcons,
	  zminusBcons,
	  zaddcarryPlus,
	  zaddcarryMinus,
	  zaddcarryBcons,
	  zaddMinus,
	  zaddBcons,
	  zmultBcons,
	  zmod_def,
	  zless_def,
	  zle_def,
	  zgt_def,
	  zge_def,
	  zandBcons,
	  znotBcons,
          zasrBcons,
          zrorBcons,
          zor_def,
          zxor_def,
          zasl_def,
          zrol_def
	  ];
