(* 
   file        : $CAT_DIR/ExtBin/ExtBin.thy
   platform    : PolyML 3.2 and Isabelle94-8/HOL
   version     : 1.0b  (16.07.98)
   notes       : The Isabelle theory for CAT; it defines an integer type
                 with operations necessary to model computer arithmetic.
                 This theory also interfaces to an external simplifier and
                 a decision procedure via an oracle.
*)

ExtBin  =  Arith +


(* Oracle for external reasoners (defined in `ExtBinOracle.ML') *)

oracle 
  computer_arithmetic_reasoner

(* Definition of type "int" *)

datatype int = PlusSign
             | MinusSign
             | Bcons int bool

instance
  int :: {ord, plus, times, minus}


syntax
  "_Int"  :: xnum => int     ("_")

rules
  zeqPlus_def  "Bcons PlusSign False == PlusSign"
  zeqMinus_def "Bcons MinusSign True == MinusSign"


(* Conversions and normalization *)
  
consts
  znat      :: nat  => int       ("$# _" [80] 80)
  zbool     :: bool => int
  normint   :: int  => int
  
consts
  normBcons :: [int, bool] => int

primrec normint int
  normintPlus  "normint PlusSign    = PlusSign"
  normintMinus "normint MinusSign   = MinusSign"
  normintBcons "normint (Bcons x y) = normBcons (normint x) y"

primrec normBcons int
  normPlus  "normBcons PlusSign  b = (if b then (Bcons PlusSign b) \
\				      else PlusSign)"
  normMinus "normBcons MinusSign b = (if b then MinusSign else \
\				      (Bcons MinusSign b))"
  normBcons "normBcons (Bcons w' x') b = Bcons (normBcons w' x') b"
  
defs
  zbool_def   "zbool x == Bcons PlusSign x"
  

(* Arithmetical operations *)
  
consts
  zsuc      :: int => int
  zpred     :: int => int
  zminus    :: int => int        ("$~ _" [80] 80)
  zdiv      :: [int, int] => int (infixl 70)
  zmod      :: [int, int] => int (infixl 70)
  zexp      :: [int, int] => int  (infixr 75)

consts
  zaddcarry :: [int, bool, int] => int

primrec zsuc int
  zsucPlus  "zsuc PlusSign   = Bcons PlusSign True" 
  zsucMinus "zsuc MinusSign  = PlusSign"
  zsucBcons "zsuc(Bcons w x) = (if x then (Bcons (zsuc w) False) \
\				else (Bcons w True))"

primrec zpred int
  zpredPlus  "zpred PlusSign    = MinusSign"
  zpredMinus "zpred MinusSign   = Bcons MinusSign False"
  zpredBcons "zpred (Bcons w x) = (if x then (Bcons w False) \
\				   else (Bcons (zpred w) True))"

primrec zminus int
  zminusPlus  "zminus PlusSign    = PlusSign"
  zminusMinus "zminus MinusSign   = Bcons PlusSign True"
  zminusBcons "zminus (Bcons w x) = (if x \
\				     then (Bcons (zpred (zminus w)) True) \
\				     else (Bcons (zminus w) False))"

primrec zaddcarry int
  zaddcarryPlus  "zaddcarry v x PlusSign    = Bcons v x"
  zaddcarryMinus "zaddcarry v x MinusSign   = zpred (Bcons v x)"
  zaddcarryBcons "zaddcarry v x (Bcons w y) = \
\	          Bcons (v + (if (x & y) then zsuc w else w)) (x ~= y)" 

primrec "op +" int
  zaddPlus  "PlusSign    + w = w"
  zaddMinus "MinusSign   + w = zpred w"
  zaddBcons "(Bcons v x) + w = zaddcarry v x w"

primrec "op *" int
  zmultPlus  "PlusSign    * w = PlusSign"
  zmultMinus "MinusSign   * w = zminus w"
  zmultBcons "(Bcons v x) * w = \
\	      (if x then ((Bcons (v * w) False) + w) \
\	       else (Bcons (v * w) False))"

defs
  zsub_def	"x - y == x + ($~ y)"
  

(* Additional conversions *)
  
primrec znat nat
  znatZero "znat 0       = PlusSign"
  znatSuc  "znat (Suc n) = zsuc (znat n)"

  
(* Ordering relations, positive and negative, absolute value *)
  
consts
  znegative  :: int => bool
  zpositive  :: int => bool
  zmagnitude :: int => int
  zgt        :: [int,int] => bool  (infixr ">"  50)
  zge        :: [int,int] => bool  (infixr ">=" 50)

primrec znegative int
  znegativePlus  "znegative PlusSign    = False"
  znegativeMinus "znegative MinusSign   = True"
  znegativeBcons "znegative (Bcons x y) = znegative x"
  
defs
  zpositive_def  "zpositive x == ~ znegative x"
  zmagnitude_def "zmagnitude x == (if (zpositive x) then x else $~ x)"
  zless_def      "x < y == znegative(x - y)"
  zle_def        "(x::int) <= y == x < y | x = y"
  zgt_def        "x > y  == y < x"
  zge_def        "x >= y == y <= x"

(* Additional arithmetical definitions *)

defs
  zmod_def "x zmod y == x - y * (x zdiv y)"
  
rules
  zexpZero_def  "x ~= #0 ==> x zexp #0 = #1"
  zexpOne_def   "#1 zexp y  = #1"
  zexpNeg_deg   "#~1 zexp y = (if (y zmod #2 = #0) then #1 else #~1)"
  zexpEOne_def  "x zexp #1 = x"
  zexp_def      "x zexp y = (if (y < #0) then #0 else \
\			     (if (y zmod #2 = #0) then #1 else x)) * \
\		            (x zexp (y zdiv #2)) * (x zexp (y zdiv #2))"
  zdivZero_def  "#0 ~= x              ==> #0 zdiv x = #0"
  zdivPosGe_def "[| #0 < y; y <= x |] ==> x zdiv y = zsuc ((x - y) zdiv y)"
  zdivPosLt_def "[| #0 < x; x < y  |] ==> x zdiv y = #0"
  zdivNeg_def   "[| x < #0; y < #0 |] ==> x zdiv y = ($~ x) zdiv ($~ y)"
  zdivNP_def    "[| x < #0; #0 < y |] ==> x zdiv y = $~ zsuc (($~ x) zdiv y)"
  zdivPN_def    "[| y < #0; #0 < x |] ==> x zdiv y = ($~ x) zdiv ($~ y)"

  
(* Bitwise operations *)
  
consts
  zand     :: [int, int] => int   (infixl 80)
  zor      :: [int, int] => int   (infixl 80)
  zxor     :: [int, int] => int   (infixl 80)
  znot     :: int => int
  zasl     :: int => int
  zasr     :: int => int
  zrol     :: int => int
  zror     :: int => int

consts
  zandhdtl :: [int, bool, int] => int

primrec zandhdtl int
  zandhdtlPlus  "zandhdtl v x PlusSign      = PlusSign"
  zandhdtlMinus "zandhdtl v x MinusSign     = Bcons v x"
  zandhdtlBcons "zandhdtl v x (Bcons v' x') = Bcons (v zand v') (x & x')"
  
primrec "op zand" int
  zandPlus  "PlusSign    zand w = PlusSign"
  zandMinus "MinusSign   zand w = w"
  zandBcons "(Bcons v x) zand w = zandhdtl v x w"

primrec znot int
  znotPlus  "znot PlusSign    = MinusSign"
  znotMinus "znot MinusSign   = PlusSign"
  znotBcons "znot (Bcons v x) = Bcons (znot v) (~x)"

primrec zasr int
  zasrPlus  "zasr PlusSign    = PlusSign"
  zasrMinus "zasr MinusSign   = MinusSign"
  zasrBcons "zasr (Bcons v x) = v"

primrec zror int
  zrorPlus  "zror PlusSign    = PlusSign"
  zrorMinus "zror MinusSign   = MinusSign"
  zrorBcons "zror (Bcons v x) = (if x then $~ (zmagnitude v) \
\				 else (zmagnitude v))"
  
defs
  zor_def	"x zor y  == znot((znot x) zand (znot y))"
  zxor_def      "x zxor y == (x zand (znot y)) zor ((znot x) zand y)"
  zasl_def      "zasl x   == Bcons x False"
  zrol_def      "zrol x   == Bcons x (znegative x)"

(* ------------------------------------------------------------------------- *)

(* Standard rules for arithmetic: most of them can be proven in our theory
   from definitions, but it is long and complex to do so.                    *)
  
rules
  zaddAssoc	"(x::int) + (y + z) = (x + y) + z"
  zaddComm      "(x::int) + y = y + x"
  zmultAssoc    "(x::int) * (y * z) = (x * y) * z"
  zmultComm     "(x::int) * y = y * x"
  zmultDistrib  "(x::int) * (y + z) = x * y + x * z"

  intInduct     "[| P n; \
\		    !! x. P x ==> P (zsuc x); \
\		    !! x. P x ==> P (zpred x) |] \
\		 ==> P t"
  intPosInduct  "[| P n; \
\		    !! x. [| n <= x; P x |] ==> P (zsuc x) |] \
\		 ==> n <= t --> P t"

  intNegInduct  "[| P n; \
\		    !! x. [| x <= n; P x |] ==> P (zpred x) |] \
\		 ==> t <= n --> P t"

  intRngInduct  "[| P n; \
\		     !! x. [| n <= x; x <= t; P x |] ==> P (zsuc x) |] \
\		  ==> P t"
  
  intRngInduct' "[| P n; \
\		     !! x. [| t <= x; x <= n; P x |] ==> P (zpred x) |] \
\		  ==> P t"

end

ML

(* Parsing and printing translations *)
  
local
  open Syntax;
    
  fun mk_bit 0 = const "False"
    | mk_bit 1 = const "True"
    | mk_bit _ = sys_error "mk_bit";

  fun dest_bit (Const ("False", _)) = 0
    | dest_bit (Const ("True", _))  = 1
    | dest_bit _                    = raise Match;

  fun mk_bin str =
    let
      val (sign, digs) =
        (case explode str of
          "#" :: "~" :: cs => (~1, cs)
        | "#" :: cs        => (1, cs)
        | _                => raise ERROR);

      fun bin_of 0  = []
        | bin_of ~1 = [~1]
        | bin_of n  = (n mod 2) :: bin_of (n div 2);

      fun term_of []        = const "PlusSign"
        | term_of [~1]      = const "MinusSign"
        | term_of (b :: bs) = const "Bcons" $ term_of bs $ mk_bit b;
    in
      term_of (bin_of (sign * (#1 (scan_int digs))))
    end;

  fun dest_bin tm =
    let
      fun bin_of (Const ("PlusSign", _))       = [0]
        | bin_of (Const ("MinusSign", _))      = [~1]
        | bin_of (Const ("Bcons", _) $ bs $ b) = dest_bit b :: bin_of bs
        | bin_of _                             = raise Match;

      fun int_of []        = 0
        | int_of (b :: bs) = b + 2 * int_of bs;

      val rev_digs   = bin_of tm;
      val (sign, zs) =
        (case rev rev_digs of
          ~1 :: bs => ("~", bs)
        | bs       => ("", bs));
      val num = string_of_int (abs (int_of rev_digs));
    in
      "#" ^ sign ^ num
    end;


  fun int_tr [t as Free (str, _)] =
        (mk_bin str handle ERROR => raise_term "int_tr" [t])
    | int_tr ts                   = raise_term "int_tr" ts;

  fun intPlus_tr'  _ = const "_Int" $ free ("#0");
  fun intMinus_tr' _ = const "_Int" $ free ("#~1");
  
  fun int_tr' [t,s] = const "_Int" $
                            free (dest_bin (Const ("Bcons",dummyT) $
                                            t $ s))
    | int_tr' _     = raise Match;

in
  val parse_translation = [("_Int", int_tr)];
  val print_translation = [("PlusSign",  intPlus_tr'),
                           ("MinusSign", intMinus_tr'),
                           ("Bcons", int_tr')];
end;
