(* 
   file     : $CAT_DIR/IntSimplifier/IntSimplifier.ML
   platform : PolyML 3.2 and Isabelle94-8/HOL
   version  : 1.0  (16.07.98)
   notes    : The CAT Simplifier is defined here.
*)

signature IntSIMP =
sig

  val int_simplify: IntSemantics.IntTerm -> IntSemantics.IntTerm

end;


structure IntSimplifier: IntSIMP =
struct

local

  open Basics;
  open IntSemantics;

  infix pow;
  infix zand;
  infix zor;
  infix zxor;


(* auxiliary functions for an ordering relation *)


(* type: IntTerm * unit -> bool *)

  infix is_number;
  fun (IntNat (NatValue _)) is_number () = true
    | (IntValue _)          is_number () = true
    | _                     is_number () = false;


(* type: IntTerm -> int *)

  fun root_weight iterm =
        case iterm of
            IntNat _      =>  2
          | IntConstant _ =>  1
          | IntValue _    =>  0
          | IntAbs _      =>  3
          | IntComp _     =>  4
          | IntPlus _     =>  5
          | IntMinus _    =>  6
          | IntTimes _    =>  7
          | IntDivide _   =>  8
          | IntModulus _  =>  9
          | IntPower _    => 10
          | IntAnd _      => 11
          | IntOr _       => 12
          | IntXOr _      => 13
          | IntNot _      => 14
          | IntASL _      => 15
          | IntASR _      => 16
          | IntRoL _      => 17
          | IntRoR _      => 18;


(* type: IntTerm -> string *)

  fun name_of_const (IntConstant (term, _)) = name_of_term term
    | name_of_const _ = raise Error "Attempt to take name of non-constant!";


(* type: IntTerm -> IntTerm *)

  fun rm_comp (IntComp x) = x
    | rm_comp x = x;


(* type: IntTerm -> IntTerm list *)

  fun pure_factors prod =
    let
      fun pure_factors' (IntTimes (x, y)) =
            (pure_factors' x) @ (pure_factors' y)
        | pure_factors' x =
            if x is_number ()
              then []
              else [x];
    in
      map rm_comp (pure_factors' prod)
    end;


(* type: IntTerm list * IntTerm list -> bool *)

  infix lit_ord;
  fun x lit_ord y =
    let
      fun ord' []           = false
        | ord' ((a, b)::fs) =
        let
          val w_a = root_weight a;
          val w_b = root_weight b;
        in
          if w_a = w_b
            then
              (case w_a of
                  1 => if (name_of_const a) = (name_of_const b)
                         then ord' fs
                         else (name_of_const a) < (name_of_const b)
                | _ => ord' fs)
            else
              w_a < w_b
        end;
    in
      if (length x) = (length y)
        then
          ord' (zip (x, y))
        else
          (length x) < (length y)
    end;


(* imposes a relation over IntTerm which sets IntValues right of IntConstants
   and IntConstants right of other IntTerms; IntConstants themself are
   ordered by `is_less_then', defined above.
   type: IntTerm * IntTerm -> bool  *)

  infix is_right_of;
  fun (x is_right_of y) =
    let
      val w_x = root_weight (rm_comp x);
      val w_y = root_weight (rm_comp y);
    in
      if w_x = w_y
        then
          (case w_x of
               1 => (name_of_const (rm_comp x)) < (name_of_const (rm_comp y))
             | 7 => (pure_factors x) lit_ord (pure_factors y)
             | _ => false)
        else
          w_x < w_y
    end;


(* syntactical equality relation over IntTerm;
   could be changed for a better one.
   type: IntTerm * IntTerm -> bool  *)

  infix is_equal_to;
  fun x is_equal_to y = x = y;


(* auxiliary axiom; identity.
   type: IntTerm -> IntTerm  *)

  fun DummyAxiom x = x;


(*** Calculation axiom ***)
 
  (* IntAxiom0: n Int<op> m = n' <op> m' *)
 
  fun IntAxiom0 (IntNat (NatValue n)) =
        IntValue n
    | IntAxiom0 (IntAbs (IntValue n)) =
        if n < 0 then IntValue (~n) else IntValue n
    | IntAxiom0 (IntComp (IntValue n)) =
        IntValue (~n)
    | IntAxiom0 (IntPlus (IntValue n, IntValue m)) =
        (IntValue (n + m) 
          handle e => IntPlus (IntValue n, IntValue m))
    | IntAxiom0 (IntMinus (IntValue n, IntValue m)) =
        (IntValue (n - m) 
          handle e => IntMinus (IntValue n, IntValue m))
    | IntAxiom0 (IntTimes (IntValue n, IntValue m)) =
        (IntValue (n * m)
          handle e => IntTimes (IntValue n, IntValue m))
    | IntAxiom0 (IntDivide (IntValue n, IntValue m)) =
        (IntValue (n div m)
          handle e => IntDivide (IntValue n, IntValue m))
    | IntAxiom0 (IntModulus (IntValue n, IntValue m)) =
        (IntValue (n mod m)
          handle e => IntModulus (IntValue n, IntValue m))
    | IntAxiom0 (IntPower (IntValue n, IntValue m)) =
        (IntValue (n pow m)
          handle e => IntPower (IntValue n, IntValue m))
    | IntAxiom0 (IntAnd (IntValue n, IntValue m)) =
        (IntValue (n zand m)
          handle e => IntAnd (IntValue n, IntValue m))
    | IntAxiom0 (IntOr (IntValue n, IntValue m)) =
        (IntValue (n zor m)
          handle e => IntOr (IntValue n, IntValue m))
    | IntAxiom0 (IntXOr (IntValue n, IntValue m)) =
        (IntValue (n zxor m)
          handle e => IntXOr (IntValue n, IntValue m))
    | IntAxiom0 (IntNot (IntValue n)) =
        (IntValue (znot n) 
          handle e => IntNot (IntValue n))
    | IntAxiom0 (IntASL (IntValue n)) =
        (IntValue (zasl n) 
          handle e => IntASL (IntValue n))
    | IntAxiom0 (IntASR (IntValue n)) =
        IntValue (zasr n) 
    | IntAxiom0 (IntRoL (IntValue n)) =
        (IntValue (zrol n) 
          handle e => IntRoL (IntValue n))
    | IntAxiom0 (IntRoR (IntValue n)) =
        IntValue (zror n) 
    | IntAxiom0 x = x;
 

(*** Addition axioms ***)
 
  (* IntAxiom1: 0 + x = x *)
 
  fun IntAxiom1 (IntPlus (IntValue 0, x)) =
        x
    | IntAxiom1 x = x;
 
  (* IntAxiom2: x + 0 = x *)
 
  fun IntAxiom2 (IntPlus (x, IntValue 0)) =
        x
    | IntAxiom2 x = x;
 
  (* IntAxiom3: x .=. y ==> x + y = 2x *)
 
  fun IntAxiom3 (IntPlus (x, y)) =
        if x is_equal_to y
          then IntTimes (IntValue 2, x)
          else IntPlus (x, y)
    | IntAxiom3 x = x;
 
  (* IntAxiom4: x .=. y ==> x + (- y) = 0 *)
 
  fun IntAxiom4 (IntPlus (x, IntComp y)) =
        if x is_equal_to y
          then IntValue 0
          else IntPlus (x, IntComp y)
    | IntAxiom4 x = x;
 
  (* IntAxiom5: x .=. y ==> (- x) + y = 0 *)
 
  fun IntAxiom5 (IntPlus (IntComp x, y)) =
        if x is_equal_to y
          then IntValue 0
          else IntPlus (IntComp x, y)
    | IntAxiom5 x = x;
 
  (* IntAxiom6: x .=. y ==> (- x) + (- y) = ~2x *)
 
  fun IntAxiom6 (IntPlus (IntComp x, IntComp y)) =
        if x is_equal_to y
          then IntTimes (IntValue ~2, x)
          else IntPlus (IntComp x, IntComp y)
    | IntAxiom6 x = x;
 
  (* IntAxiom7: x .=. y ==> x + (y + z) = 2x + z *)
 
  fun IntAxiom7 (IntPlus (x, IntPlus (y, z))) =
        if x is_equal_to y
          then IntPlus (IntTimes (IntValue 2, x), z)
          else IntPlus (x, IntPlus (y, z))
    | IntAxiom7 x = x;
 
  (* IntAxiom8: x .=. y ==> x + (- y + z) = z *)
 
  fun IntAxiom8 (IntPlus (x, IntPlus (IntComp y, z))) =
        if x is_equal_to y
          then z
          else IntPlus (x, IntPlus (IntComp y, z))
    | IntAxiom8 x = x;
 
  (* IntAxiom9: x .=. y ==> (- x) + (y + z) = z *)
 
  fun IntAxiom9 (IntPlus (IntComp x, IntPlus (y, z))) =
        if x is_equal_to y
          then z
          else IntPlus (IntComp x, IntPlus (y, z))
    | IntAxiom9 x = x;
 
  (* IntAxiom10: x .=. y ==> (- x) + (- y + z) = ~2x + z *)
 
  fun IntAxiom10 (IntPlus (IntComp x, IntPlus (IntComp y, z))) =
        if x is_equal_to y
          then IntPlus (IntTimes (IntValue ~2, x), z)
          else IntPlus (IntComp x, IntPlus (y, z))
    | IntAxiom10 x = x;
 
  (* IntAxiom11: x .=. y  ==>  n * x + m * y = (n+m) * x  *)
 
  fun IntAxiom11 (IntPlus (IntTimes (IntValue n, x),
                          IntTimes (IntValue m, y))) =
        if (x is_equal_to y)
          then IntTimes (IntValue (n + m), x)
          else IntPlus (IntTimes (IntValue n, x),
                          IntTimes (IntValue m, y))
    | IntAxiom11 x = x;
 
  (* IntAxiom12: x .=. y  ==>  n * x + (m * y + z) = (n+m) * x + z  *)

  fun IntAxiom12 (IntPlus (IntTimes (IntValue n, x),
                 IntPlus (IntTimes (IntValue m, y), z))) =
        if (x is_equal_to y)
          then IntPlus (IntTimes (IntValue (n + m), x), z)
          else IntPlus (IntTimes (IntValue n, x),
                 IntPlus (IntTimes (IntValue m, y), z))
    | IntAxiom12 x = x;
 
  (* IntAxiom13: x .=. y  ==>  x + n * y = (1+n) * x  *)
 
  fun IntAxiom13 (IntPlus (x, IntTimes (IntValue n, y))) =
        if (x is_equal_to y)
          then IntTimes (IntValue (1 + n), x)
          else IntPlus (x, IntTimes (IntValue n, y))
    | IntAxiom13 x = x;
 
  (* IntAxiom14: x .=. y  ==>  x + (n * y + z) = (1+n) * x + z  *)
 
  fun IntAxiom14 (IntPlus (x, IntPlus (IntTimes (IntValue n, y), z))) =
        if (x is_equal_to y)
          then IntPlus (IntTimes (IntValue (1 + n), x), z)
          else IntPlus (x, IntPlus (IntTimes (IntValue n, y), z))
    | IntAxiom14 x = x;
 
  (* IntAxiom15: x .=. y  ==>  n * x + y = (n+1) * x  *)
 
  fun IntAxiom15 (IntPlus (IntTimes (IntValue n, x), y)) =
        if (x is_equal_to y)
          then IntTimes (IntValue (n + 1), x)
          else IntPlus (IntTimes (IntValue n, x), y)
    | IntAxiom15 x = x;
 
  (* IntAxiom16: x .=. y  ==>  n * x + (y + z) = (n+1) * x + z  *)
 
  fun IntAxiom16 (IntPlus (IntTimes (IntValue n, x), IntPlus (y, z))) =
        if (x is_equal_to y)
          then IntPlus (IntTimes (IntValue (n + 1), x), z)
          else IntPlus (IntTimes (IntValue n, x), IntPlus (y, z))
    | IntAxiom16 x = x;
 
  (* IntAxiom17: x + (- y) = x - y  *)
 
  fun IntAxiom17 (IntPlus (x, IntComp y)) =
        IntMinus (x, y)
    | IntAxiom17 x = x;
 
  (* IntAxiom18: (- x) + y = y - x  *)

  fun IntAxiom18 (IntPlus (IntComp x, y)) =
        IntMinus (y, x)
    | IntAxiom18 x = x;
 
  (* IntAxiom19: x + ((- y) + z) = (x + z) - y  *)
 
  fun IntAxiom19 (IntPlus (x, IntPlus (IntComp y, z))) =
        IntMinus (IntPlus (x, z), y)
    | IntAxiom19 x = x;
 
  (* IntAxiom20: (- x) + (y + z) = (y + z) - x  *)

  fun IntAxiom20 (IntPlus (IntComp x, IntPlus (y, z))) =
        IntMinus (IntPlus (y, z), x)
    | IntAxiom20 x = x;
 

(*** Subtraction axioms ***)
 
  (* IntAxiom30: x - y = x + (- y) *)
 
  fun IntAxiom30 (IntMinus (x, y)) =
        IntPlus (x, IntComp y)
    | IntAxiom30 x = x;
 


(*** Complement axioms ***)
 
  (* IntAxiom40: - (- x) = x *)
 
  fun IntAxiom40 (IntComp (IntComp x)) =
        x
    | IntAxiom40 x = x;
 


(*** Multiplication axioms ***)

  (* IntAxiom50: 0 * x = 0 *)
 
  fun IntAxiom50 (IntTimes (IntValue 0, x)) =
        IntValue 0
    | IntAxiom50 x = x;
 
  (* IntAxiom51: x * 0 = 0  *)
 
  fun IntAxiom51 (IntTimes (x, IntValue 0)) =
        IntValue 0
    | IntAxiom51 x = x;
 
  (* IntAxiom52: 1 * x = x  *)
 
  fun IntAxiom52 (IntTimes (IntValue 1, x)) =
        x
    | IntAxiom52 x = x;
 
  (* IntAxiom53: x * 1 = x  *)
 
  fun IntAxiom53 (IntTimes (x, IntValue 1)) =
        x
    | IntAxiom53 x = x;
 
  (* IntAxiom54: ~1 * x = - x  *)
 
  fun IntAxiom54 (IntTimes (IntValue ~1, x)) =
        IntComp x
    | IntAxiom54 x = x;
 
  (* IntAxiom55: x * ~1 = - x  *)
 
  fun IntAxiom55 (IntTimes (x, IntValue ~1)) =
        IntComp x
    | IntAxiom55 x = x;
 


(*** Division axioms ***)
 
  (* IntAxiom60: x .=. y  &  y .!=. 0  ==>  x / y = 1 *)
 
  fun IntAxiom60 (IntDivide (x, y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntValue 0)))
          then IntValue 1
          else IntDivide (x, y)
    | IntAxiom60 x = x;
 
  (* IntAxiom61: x .=. y  &  y .!=. 0  ==>  (- x) / y = ~1 *)
 
  fun IntAxiom61 (IntDivide (IntComp x, y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntValue 0)))
          then IntValue ~1
          else IntDivide (IntComp x, y)
    | IntAxiom61 x = x;
 
  (* IntAxiom62: x .=. y  &  y .!=. 0  ==>  (x * z) / y = z *)
 
  fun IntAxiom62 (IntDivide (IntTimes (x, z), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntValue 0)))
          then z
          else IntDivide (IntTimes (x, z), y)
    | IntAxiom62 x = x;
 
  (* IntAxiom63: x .=. y  &  y .!=. 0  ==>  (z * x) / y = z *)
 
  fun IntAxiom63 (IntDivide (IntTimes (z, x), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntValue 0)))
          then z
          else IntDivide (IntTimes (z, x), y)
    | IntAxiom63 x = x;

  (* IntAxiom64: x .=. y  &  y .!=. 0  ==>  ((z * x) * z') / y = z * z'  *)
 
  fun IntAxiom64 (IntDivide (IntTimes (IntTimes (z, x), z'), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntValue 0)))
          then IntTimes (z, z')
          else IntDivide (IntTimes (IntTimes (z, x), z'), y)
    | IntAxiom64 x = x;

  (* IntAxiom65: n mod m = 0  ==> (n * x) / m = (n/m) * x  *)

  fun IntAxiom65 (IntDivide (IntTimes (IntValue n, x), IntValue m)) =
        ((if (n mod m) = 0
           then IntTimes (IntValue (n div m), x)
           else IntDivide (IntTimes (IntValue n, x), IntValue m))
         handle e => IntDivide (IntTimes (IntValue n, x), IntValue m))
    | IntAxiom65 x = x;
 
  (* IntAxiom66:  x / 1 = x  *)
 
  fun IntAxiom66 (IntDivide (x, IntValue 1)) =
        x
    | IntAxiom66 x = x;
 
  (* IntAxiom67:  x / ~1 = - x  *)
 
  fun IntAxiom67 (IntDivide (x, IntValue ~1)) =
        IntComp x
    | IntAxiom67 x = x;
 
  (* IntAxiom68:  x .!=. 0  ==>  0 / x = 0 *)
 
  fun IntAxiom68 (IntDivide (IntValue 0, x)) =
        if (not (x is_equal_to (IntValue 0)))
          then IntValue 0
          else IntDivide (IntValue 0, x)
    | IntAxiom68 x = x;



(*** Modulus axioms ***)
 
  (* IntAxiom70: x .=. y  &  y .!=. 0  ==>  x mod y = 0 *)
 
  fun IntAxiom70 (IntModulus (x, y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntValue 0)))
          then IntValue 0
          else IntModulus (x, y)
    | IntAxiom70 x = x;
 
  (* IntAxiom71: x .=. y  &  y .!=. 0  ==>  (x * z) mod y = 0 *)
 
  fun IntAxiom71 (IntModulus (IntTimes (x, z), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntValue 0)))
          then IntValue 0
          else IntModulus (IntTimes (x, z), y)
    | IntAxiom71 x = x;
 
  (* IntAxiom72: x .=. y  &  y .!=. 0  ==>  (z * x) mod y = 0 *)
 
  fun IntAxiom72 (IntModulus (IntTimes (z, x), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntValue 0)))
          then IntValue 0
          else IntModulus (IntTimes (z, x), y)
    | IntAxiom72 x = x;

  (* IntAxiom73: x .=. y  &  y .!=. 0  ==>  ((z * x) * z') mod y = 0 *)
 
  fun IntAxiom73 (IntModulus (IntTimes (IntTimes (z, x), z'), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntValue 0)))
          then IntValue 0
          else IntModulus (IntTimes (IntTimes (z, x), z'), y)
    | IntAxiom73 x = x;

  (* IntAxiom74: n mod m = 0  ==> (n * x) mod m = 0  *)

  fun IntAxiom74 (IntModulus (IntTimes (IntValue n, x), IntValue m)) =
        ((if (n mod m) = 0
           then IntValue 0
           else IntModulus (IntTimes (IntValue n, x), IntValue m))
         handle e => IntModulus (IntTimes (IntValue n, x), IntValue m))
    | IntAxiom74 x = x;
  
  (* IntAxiom75: abs(n).<=.abs(m) & n.!=.0 ==> (x mod n) mod m = x mod n  *)

  fun IntAxiom75 (IntModulus (IntModulus (x, IntValue n), IntValue m)) =
        if ((abs n) <= (abs m)) andalso (not (n = 0))
          then IntModulus (x, IntValue n)
          else IntModulus (IntModulus (x, IntValue n), IntValue m)
    | IntAxiom75 x = x;
  
  (* IntAxiom76: x.=.y & x.!=.0 ==> (z mod x) mod y = z mod y *)
 
  fun IntAxiom76 (IntModulus (IntModulus (z, x), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntValue 0)))
          then IntModulus (z, y)
          else IntModulus (IntModulus (z, x), y)
    | IntAxiom76 x = x;

  (* IntAxiom77: x.=.y & x.!=.0 ==> (z mod x + z') mod y = (z + z') mod y *)
 
  fun IntAxiom77 (IntModulus (IntPlus (IntModulus (z, x), z'), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntValue 0)))
          then IntModulus (IntPlus (z, z'), y)
          else IntModulus (IntPlus (IntModulus (z, x), z'), y)
    | IntAxiom77 x = x;

  (* IntAxiom78: x.=.y & x.!=.0 ==> (z + z' mod x) mod y = (z + z') mod y *)
 
  fun IntAxiom78 (IntModulus (IntPlus (z, IntModulus (z', x)), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntValue 0)))
          then IntModulus (IntPlus (z, z'), y)
          else IntModulus (IntPlus (z, IntModulus (z', x)), y)
    | IntAxiom78 x = x;

  (* IntAxiom79: x.=.y & x.!=.0 ==> (z'' + (z mod x +z')) mod y =
                                    (z'' + (z + z')) mod y *)
 
  fun IntAxiom79 (IntModulus 
                (IntPlus (z'', IntPlus (IntModulus (z, x), z')), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntValue 0)))
          then IntModulus (IntPlus (z'', IntPlus (z, z')), y)
          else IntModulus 
                (IntPlus (z'', IntPlus (IntModulus (z, x), z')), y)
    | IntAxiom79 x = x;

  (* IntAxiom80: abs(n).<=.m & n.!=.0 ==> (x mod n) div m = 0  *)
 
  fun IntAxiom80 (IntDivide (IntModulus (x, IntValue n), IntValue m)) =
        if ((abs n) <= m) andalso (not (n = 0))
          then IntValue 0
          else IntDivide (IntModulus (x, IntValue n), IntValue m)
    | IntAxiom80 x = x;

  (* IntAxiom81: ~(abs(n)).>=.m & n.!=.0 ==> (x mod n) div m = 0  *)
 
  fun IntAxiom81 (IntDivide (IntModulus (x, IntValue n), IntValue m)) =
        if (~(abs n) >= m) andalso (not (n = 0))
          then IntValue ~1
          else IntDivide (IntModulus (x, IntValue n), IntValue m)
    | IntAxiom81 x = x;

  (* IntAxiom82: x.!=.0 ==> 0 mod x = 0  *)
 
  fun IntAxiom82 (IntModulus (IntValue 0, x)) =
        if not (x is_equal_to (IntValue 0))
          then IntValue 0
          else IntModulus (IntValue 0, x)
    | IntAxiom82 x = x;



(*** Exponential Axioms ***)
 
  (* IntAxiom90: x .!=. 0  ==> x ^ 0 = 1 *)

  fun IntAxiom90 (IntPower (x, IntValue 0)) =
        if not (x is_equal_to IntValue 0)
          then IntValue 1
          else IntPower (x, IntValue 0)
    | IntAxiom90 x = x;
 
  (* IntAxiom91: x ^ 1 = x *)

  fun IntAxiom91 (IntPower (x, IntValue 1)) =
        IntValue 1
    | IntAxiom91 x = x;
 
  (* IntAxiom92: 1 ^ x = 1 *)

  fun IntAxiom92 (IntPower (IntValue 1, x)) =
        IntValue 1
    | IntAxiom92 x = x;
 
  (* IntAxiom93: n .<. 0  &  x .!=. 0  ==>  x ^ n = 0 *)

  fun IntAxiom93 (IntPower (x, IntValue n)) =
        if (n < 0) andalso (not (x is_equal_to (IntValue 0)))
          then IntValue 0
          else IntPower (x, IntValue n)
    | IntAxiom93 x = x;
 
  (* IntAxiom94: x .=. y  ==>  x ^ z * y ^ z' = x ^ (z + z') *)

  fun IntAxiom94 (IntTimes (IntPower (x, z), IntPower (y, z'))) =
        if (x is_equal_to y)
          then IntPower (x, IntPlus (z, z'))
          else IntTimes (IntPower (x, z), IntPower (y, z'))
    | IntAxiom94 x = x;

  (* IntAxiom95: x .=. y  ==>  x * y ^ z = x ^ (z + 1) *)

  fun IntAxiom95 (IntTimes (x, IntPower (y, z))) =
        if (x is_equal_to y)
          then IntPower (x, IntPlus (z, IntValue 1))
          else IntTimes (x, IntPower (y, z))
    | IntAxiom95 x = x;

  (* IntAxiom96: x .=. y  ==>  x ^ z * y = x ^ (z + 1) *)

  fun IntAxiom96 (IntTimes (IntPower (x, z), y)) =
        if (x is_equal_to y)
          then IntPower (x, IntPlus (z, IntValue 1))
          else IntTimes (IntPower (x, z), y)
    | IntAxiom96 x = x;

  (* IntAxiom97: x .=. y  ==>  z ^ x * z' ^ y = (z * z') ^ x *)

  fun IntAxiom97 (IntTimes (IntPower (z, x), IntPower (z', y))) =
        if (x is_equal_to y)
          then IntPower (IntTimes (z, z'), x)
          else IntTimes (IntPower (z, x), IntPower (z', y))
    | IntAxiom97 x = x;
 
  (* IntAxiom98: x .=. y  ==>  x ^ z / y ^ z' = x ^ (z + (- z')) *)

  fun IntAxiom98 (IntDivide (IntPower (x, z), IntPower (y, z'))) =
        if (x is_equal_to y)
          then IntPower (x, IntPlus (z, IntComp z'))
          else IntDivide (IntPower (x, z), IntPower (y, z'))
    | IntAxiom98 x = x;
 
  (* IntAxiom99: x .=. y  &  z' .!=. 0  ==>  z ^ x / z' ^ y = (z / z') ^ x *)

  fun IntAxiom99 (IntDivide (IntPower (z, x), IntPower (z', y))) =
        if (x is_equal_to y) andalso (not (z' is_equal_to (IntValue 0)))
          then IntPower (IntDivide (z, z'), x)
          else IntDivide (IntPower (z, x), IntPower (z', y))
    | IntAxiom99 x = x;
 



(*** Assoc., Comm., Distr. axioms ***)
 
  (* IntAxiom200: (x + y) + z =  x + (y + z) *)
 
  fun IntAxiom200 (IntPlus (IntPlus (x, y), z)) =
        IntPlus (x, IntPlus (y, z))
    | IntAxiom200 x = x;
 
  (* IntAxiom201: x * (y * z) = (x * y) * z *)
 
  fun IntAxiom201 (IntTimes (x, IntTimes (y, z))) =
        IntTimes (IntTimes (x, y), z)
    | IntAxiom201 x = x;
 
  (* IntAxiom202: x .<. y ==> x + y = y + x *)
 
  fun IntAxiom202 (IntPlus (x, y)) =
        (case y of
             IntPlus (_, _) => IntPlus (x, y)
           | _ => if x is_right_of y
                    then IntPlus (y, x)
                    else IntPlus (x, y))
    | IntAxiom202 x = x;
 
  (* IntAxiom203: x .<. y ==> x + (y + z) = y + (x + z) *)
 
  fun IntAxiom203 (IntPlus (x, IntPlus (y, z)): IntTerm) =
        (case y of
             IntPlus (_, _) => IntPlus (x, IntPlus (y, z))
           | _ => if x is_right_of y
                    then IntPlus (y, IntPlus (x, z))
                    else IntPlus (x, IntPlus (y, z)))
    | IntAxiom203 x = x;
 
  (* IntAxiom204: x .>. y ==> x * y = x * y  *)
 
  fun IntAxiom204 (IntTimes (x, y): IntTerm) =
        (case x of
             IntTimes (_, _) => IntTimes (x, y)
           | _ => if y is_right_of x
                    then IntTimes (y, x)
                    else IntTimes (x, y))
    | IntAxiom204 x = x;
 
  (* IntAxiom205: x .>. y ==> (z * x) * y = (z * y) * x  *)
 
  fun IntAxiom205 (IntTimes (IntTimes (z, x), y): IntTerm) =
        (case x of
             IntTimes (_, _) => IntTimes (IntTimes (z, x), y)
           | _ => if x is_right_of y
                    then IntTimes (IntTimes (z, y), x)
                    else IntTimes (IntTimes (z, x), y))
    | IntAxiom205 x = x;
 
  (* IntAxiom206: x * (y + z) = x * y + x * z  *)
 
  fun IntAxiom206 (IntTimes (x, IntPlus (y, z)): IntTerm) =
        IntPlus (IntTimes (x, y), IntTimes (x, z))
    | IntAxiom206 x = x;
 
  (* IntAxiom207: (y + z) * x = y * x + z * x  *)
 
  fun IntAxiom207 (IntTimes (IntPlus (y, z), x): IntTerm) =
        IntPlus (IntTimes (y, x), IntTimes (z, x))
    | IntAxiom207 x = x;

  (* IntAxiom208: - (x + y) = (- x) + (- y) *)
 
  fun IntAxiom208 (IntComp (IntPlus (x, y))) =
        IntPlus (IntComp x, IntComp y)
    | IntAxiom208 x = x;
 
  (* IntAxiom209: - (x * y) = (- x) * y  *)
 
  fun IntAxiom209 (IntComp (IntTimes (x, y))) =
        IntTimes (IntComp x, y)
    | IntAxiom209 x = x;
 
  (* IntAxiom210: - (x / y) = (- x) / y  *)
 
  fun IntAxiom210 (IntComp (IntDivide (x, y))) =
        IntDivide (IntComp x, y)
    | IntAxiom210 x = x;

  (* IntAxiom211: - (x mod y) = (- x) mod (- y)  *)
 
  fun IntAxiom211 (IntComp (IntModulus (x, y))) =
        IntModulus (IntComp x, IntComp y)
    | IntAxiom211 x = x;

  (* IntAxiom212: x * (- y) = (- x) * y  *)
 
  fun IntAxiom212 (IntTimes (x, IntComp y)) =
        IntTimes (IntComp x, y)
    | IntAxiom212 x = x;

  (* IntAxiom213: x / (- y) = (- x) / y  *)
 
  fun IntAxiom213 (IntDivide (x, IntComp y)) =
        IntDivide (IntComp x, y)
    | IntAxiom213 x = x;

  (* IntAxiom214: (z ^ x) ^ y = z ^ (x * y)  *)

  fun IntAxiom214 (IntPower (z, IntPower (x, y))) =
        IntPower (z, IntTimes (x, y))
    | IntAxiom214 x = x;
 

(*** end of axioms ***)


  val NatAxioms = DummyAxiom;

  val CalcAxioms = IntAxiom0;

  val AddAxioms = IntAxiom16
                o IntAxiom15
                o IntAxiom14
                o IntAxiom13
                o IntAxiom12
                o IntAxiom11
                o IntAxiom10
                o IntAxiom9
                o IntAxiom8
                o IntAxiom7
                o IntAxiom6
                o IntAxiom5
                o IntAxiom4
                o IntAxiom3
                o IntAxiom2
                o IntAxiom1;

  val DiffAxioms = IntAxiom20
                 o IntAxiom19
                 o IntAxiom18
                 o IntAxiom17;

  val SubAxioms = IntAxiom30;

  val CompAxioms = IntAxiom40;

  val MultAxioms = IntAxiom55
                 o IntAxiom54
                 o IntAxiom53
                 o IntAxiom52
                 o IntAxiom51
                 o IntAxiom50;

  val DivAxioms = IntAxiom68
                o IntAxiom67
                o IntAxiom66
                o IntAxiom65
                o IntAxiom64
                o IntAxiom63
                o IntAxiom62
                o IntAxiom61
                o IntAxiom60;

  val ModAxioms = IntAxiom82
                o IntAxiom81
                o IntAxiom80
                o IntAxiom79
                o IntAxiom78
                o IntAxiom77
                o IntAxiom76
                o IntAxiom75
                o IntAxiom74
                o IntAxiom73
                o IntAxiom72
                o IntAxiom71
                o IntAxiom70;

  val ExpAxioms = IntAxiom99
                o IntAxiom98
                o IntAxiom97
                o IntAxiom96
                o IntAxiom95
                o IntAxiom94
                o IntAxiom93
                o IntAxiom92
                o IntAxiom91
                o IntAxiom90;

  val ACDAxioms = IntAxiom214
                o IntAxiom213
                o IntAxiom212
                o IntAxiom211
                o IntAxiom210
                o IntAxiom209
                o IntAxiom208
                o IntAxiom207
                o IntAxiom206
                o IntAxiom205
                o IntAxiom204
                o IntAxiom203
                o IntAxiom202
                o IntAxiom201
                o IntAxiom200;

  val IntAxioms = ACDAxioms
                o ExpAxioms
                o ModAxioms
                o DivAxioms
                o MultAxioms
                o CompAxioms
                o SubAxioms
                o AddAxioms
                o CalcAxioms;


(* rewrite IntTerms top-down using rewrite axioms for NatTerms and IntTerms
   provided as arguments.
   type (NatTerm -> NatTerm) -> (IntTerm -> IntTerm) -> IntTerm -> IntTerm  *)

  fun rewrite nat_ax int_ax term =
    let
      fun rew (Nat arg) = Nat (nat_rew arg)
        | rew (Int arg) = Int (int_rew arg)

      and nat_rew x =
        let
          val x_rew = nat_ax x;
        in
          if x = x_rew
            then
              case x of
                  NatConstant (root, args) =>
                    NatConstant (root, map rew args)
                | NatValue arg =>
                    NatValue arg
                | NatPlus (arg1, arg2) =>
                    NatPlus (nat_rew arg1, nat_rew arg2)
                | NatMinus (arg1, arg2) =>
                    NatMinus (nat_rew arg1, nat_rew arg2)
                | NatTimes (arg1, arg2) =>
                    NatTimes (nat_rew arg1, nat_rew arg2)
                | NatDivide (arg1, arg2) =>
                    NatDivide (nat_rew arg1, nat_rew arg2)
                | NatModulus (arg1, arg2) =>
                    NatModulus (nat_rew arg1, nat_rew arg2)
                | NatPower (arg1, arg2) =>
                    NatPower (nat_rew arg1, nat_rew arg2)
            else
              nat_rew x_rew
        end

      and int_rew x =
        let
          val x_rew = int_ax x;
        in
          if x = x_rew
            then
              case x of
                  IntNat arg =>
                    IntNat (nat_rew arg)
                | IntConstant (root, args) =>
                    IntConstant (root, map rew args)
                | IntValue arg =>
                    IntValue arg
                | IntAbs arg =>
                    IntAbs (int_rew arg)
                | IntComp arg =>
                    IntComp (int_rew arg)
                | IntPlus (arg1, arg2) =>
                    IntPlus (int_rew arg1, int_rew arg2)
                | IntMinus (arg1, arg2) =>
                    IntMinus (int_rew arg1, int_rew arg2)
                | IntTimes (arg1, arg2) =>
                    IntTimes (int_rew arg1, int_rew arg2)
                | IntDivide (arg1, arg2) =>
                    IntDivide (int_rew arg1, int_rew arg2)
                | IntModulus (arg1, arg2) =>
                    IntModulus (int_rew arg1, int_rew arg2)
                | IntPower (arg1, arg2) =>
                    IntPower (int_rew arg1, int_rew arg2)
                | IntAnd (arg1, arg2) =>
                    IntAnd (int_rew arg1, int_rew arg2)
                | IntOr (arg1, arg2) =>
                    IntOr (int_rew arg1, int_rew arg2)
                | IntXOr (arg1, arg2) =>
                    IntXOr (int_rew arg1, int_rew arg2)
                | IntNot arg =>
                    IntNot (int_rew arg)
                | IntASL arg =>
                    IntASL (int_rew arg)
                | IntASR arg =>
                    IntASR (int_rew arg)
                | IntRoL arg =>
                    IntAbs (int_rew arg)
                | IntRoR arg =>
                    IntRoR (int_rew arg)
            else
              int_rew x_rew
        end;
    in
      int_rew term
    end;


(* rewrites IntTerms with main set of axioms.
   (mapping `x - y' to `x + (- y)')
   type: IntTerm -> IntTerm  *)

  fun main_rew x =
    let
      val x_rew = rewrite NatAxioms IntAxioms x
    in
      if x = x_rew then x else main_rew x_rew
    end;


(* maps `x + (- y)' back to `x - y' for better readability.
   type: IntTerm -> IntTerm  *)

  fun diff_rew x =
    let
      val x_rew = rewrite DummyAxiom DiffAxioms x
    in
      if x = x_rew then x else diff_rew x_rew
    end;

in

(* simplifies an IntTerm as much as possible and then modifies it
   slightly (`x + (- y)' to `x - y') for better readability, before
   giving the result.
   type: IntTerm -> IntTerm  *)

  fun int_simplify x = diff_rew (main_rew x);

end;  (* local *)

end;  (* struct *)
