(* 
   file     : $CAT_DIR/Library/Basics.ML
   platform : PolyML 3.2 and Isabelle94-8/HOL
   version  : 1.0b  (16.07.98)
   notes    : Some basic functions for CAT are provided. This file should
              change with new versions of SML and Isabelle.
*)

signature BASICS =
sig

  exception Error of string

  val PROP: typ
  val BOOL: typ
  val NAT:  typ
  val INT:  typ

  val secl: 'a -> ('a * 'b -> 'c)       -> 'b -> 'c
  val secr:       ('a * 'b -> 'c) -> 'b -> 'a -> 'c

  val pair:      'a -> 'b -> 'a * 'b
  val swap_pair:  'a * 'b -> 'b * 'a

  val the:  'a option -> 'a

  val rm_duplicates: ('a * 'a -> bool) -> 'a list           -> 'a list
  val intersect:     ('a * 'b -> bool) -> 'a list * 'b list -> 'a list
  val flatten:        'a list list                          -> 'a list
  val zip:            'a list * 'b list           -> ('a * 'b) list
  val unzip:         ('a * 'b) list               -> 'a list * 'b list

  val lookupl: ''a -> (''a * 'b) list -> 'b option
  val lookupr: ''a -> ('b * ''a) list -> 'b option

  val pow:  int * int -> int

  val zand: int * int -> int
  val zor:  int * int -> int
  val zxor: int * int -> int
  val znot: int -> int

  val zasl: int -> int
  val zasr: int -> int
  val zrol: int -> int
  val zror: int -> int

  val string_to_int:  string -> int
  val int_to_string:  int    -> string
  val term_to_int:    term   -> int
  val int_to_term:    int    -> term
  val term_to_nat:    term   -> int
  val nat_to_term:    int    -> term

  val term_split:   term -> term * term list
  val name_of_term: term -> string

end;


structure Basics: BASICS =
struct

(* This is the general exception throughout the CAT system. *)
exception Error of string;

(* handy abbreviations ... *)
val PROP = Type ("prop", []);
val BOOL = Type ("bool", []);
val NAT  = Type ("nat", []);
val INT  = Type ("int", []);

(* ... and functors *)
fun secl x g y = g (x, y);
fun secr g y x = g (x, y);

fun pair x y = (x, y);  

fun swap_pair (x, y) = (y, x);

fun the  None    = raise Error "Attempt to dismount None!!!"
  | the (Some x) = x;


(* removes duplictaed elements from a list according to the equivalence
   relation Eq
   type: ('a * 'a -> bool) -> 'a list -> 'a list  *)

fun rm_duplicates Eq l =
  let
    fun rm_duplicates' c []      = c
      | rm_duplicates' c (x::xs) =
          if exists (secl x (op Eq)) c
            then rm_duplicates' c xs
            else rm_duplicates' (x::c) xs;
    in
      rm_duplicates' [] l
    end;


(* all elements in list xs that are equal (w.r.t. Eq) to an element
   in list ys make up resulting list. This is not a commutative operation!
   type: ('a * 'b -> bool) -> ('a list * 'b list) -> 'a list  *)

fun intersect Eq (xs, ys) = filter (fn x => exists (secl x Eq) ys) xs;


(* takes a list of lists to a list of their elements.
   type: 'a list list -> 'a list  *)

fun flatten x = foldl op@ ([], x);


(* two lists of the same length are formed to one list of tuples of their
   elements.
   type: 'a list * 'b list -> ('a * 'b) list  *)

fun zip (x, y) =
  let
    fun zip' l ([], []) = l
      | zip' l (x1'::x1, x2'::x2) = zip' ((x1', x2')::l) (x1, x2)
      | zip' l _ = raise Error "Attempt to zip lists of different length!!!";
  in
    zip' [] (rev x, rev y)
  end;


(* reverse funtion of the above.
   type: ('a * 'b) list -> 'a list * 'b list  *)

fun unzip l =
  let
    fun unzip' (x1, x2) [] = (x1, x2)
      | unzip' (x1, x2) ((x1', x2')::xs) = unzip' (x1'::x1, x2'::x2) xs;
  in
    unzip' ([], []) (rev l)
  end;


(* given a lookup table (list of pairs) and a key, the first matching
   value found is returned.
   keys are assumed to be in the left column of the table.
   type: ''a -> (''a * 'b list) -> 'b option  *)

fun lookupl key []      = None
  | lookupl key (x::xs) = if (fst x) = key
                            then Some (snd x)
                            else lookupl key xs;


(* same as above with keys on the right.
   type: ''a -> ('b * ''a list) -> 'b option  *)

fun lookupr key []      = None
  | lookupr key (x::xs) = if (snd x) = key
                            then Some (fst x)
                            else lookupr key xs;


(* implements the power operation on integers.
   type: int * int -> int  *)

infix pow;

fun  0 pow 0 = raise Error "domain error 0^0"
  |  1 pow y = 1
  | ~1 pow y = if (y mod 2) = 0 then 1 else ~1
  |  x pow 0 = 1
  |  x pow 1 = x
  |  x pow y = if y < 0
                 then 0
                 else let
                        val z = x pow (y div 2);
                      in
                        if (y mod 2) = 0 then z * z else z * z * x
                      end;


(* bitwise `and' operation on integers.
   type: int * int -> int  *)

infix zand;

(*****
fun   0 zand  _ = 0
  |   _ zand  0 = 0
  |  ~1 zand  y = y
  |   x zand ~1 = x
  |   x zand  y =
          ((x mod 2) + (y mod 2)) div 2 + 2 * ((x div 2) zand (y div 2));
... or as a tail recursive version: *)

fun x zand y =
  let
    fun zand' c pot (0, _)  = c
      | zand' c pot (_, 0)  = c
      | zand' c pot (~1, y) = c + (pot * y)
      | zand' c pot (x, ~1) = c + (pot * x)
      | zand' c pot (x, y)  =
          zand' (c + pot * (((x mod 2) + (y mod 2)) div 2))
                (2 * pot)
                (x div 2, y div 2);
  in
    zand' 0 1 (x, y)
  end;


(* bitwise `or' operation on integers.
   type: int * int -> int  *)

infix zor;

(*****
fun   0 zor  y =  y
  |   x zor  0 =  x
  |  ~1 zor  _ = ~1
  |   _ zor ~1 = ~1
  |   x zor  y =
         ((x mod 2) - 1) * ((y mod 2) - 1) + 1 + 2 * ((x div 2) zor (y div 2));
... or as a tail recursive version: *)

fun x zor y =
  let
    fun zor' c pot (0, y)  = c + (pot * y)
      | zor' c pot (x, 0)  = c + (pot * x)
      | zor' c pot (~1, _) = c - pot
      | zor' c pot (_, ~1) = c - pot
      | zor' c pot (x, y)  =
          zor' (c + pot * (((x mod 2) - 1) * ((y mod 2) - 1) + 1))
               (2 * pot)
               (x div 2, y div 2);
  in
    zor' 0 1 (x, y)
  end;


(* bitwise `xor' operation on integers.
   type: int * int -> int  *)

infix zxor;

(*****
fun   0 zxor  y = y
  |   x zxor  0 = x
  |  ~1 zxor  y = ~1 * (y + 1)
  |   x zxor ~1 = ~1 * (x + 1)
  |   x zxor  y =
          (x + y) mod 2 + 2 * ((x div 2) zxor (y div 2));
... or as a tail recursive version: *)

fun x zxor y =
  let
    fun zxor' c pot (0, y)  = c + (pot * y)
      | zxor' c pot (x, 0)  = c + (pot * x)
      | zxor' c pot (~1, y) = c - (pot * (y + 1))
      | zxor' c pot (x, ~1) = c - (pot * (x + 1))
      | zxor' c pot (x, y)  =
          zxor' (c + pot * ((x + y) mod 2))
                (2 * pot)
                (x div 2, y div 2);
  in
    zxor' 0 1 (x, y)
  end;


(* bitwise `xnot' operation on integers.
   type: int -> int  *)

fun znot x = ~1 * (x + 1);


(* arithmetical shift left on integers.
   type: int -> int  *)

fun zasl x = 2 * x;


(* arithmetical shift right on integers.
   type: int -> int  *)

fun zasr x = x div 2;


(* rotate left on integers; we use infinite precision.
   type: int -> int  *)

fun zrol x = if x < 0 then 2 * x + 1 else 2 * x;


(* rotate right on integers; we use infinite precision.
   type: int -> int  *)

fun zror x = if (x mod 2) = 0 then abs (x div 2) else ~(abs (x div 2));


(* takes a string representing an integer to its value, allowing "-" and "~"
   as minus signs.
   type: string -> int  *)

fun string_to_int s =
  let 
    fun sign xs =
      case xs of
          "~" :: Xs => ~1
        | "-" :: Xs => ~1
        | _         =>  1;

    fun convert xs =
      case xs of
          [] => 0
        | ["~"] => 0
        | ["-"] => 0
        | "0" :: Xs => 10 * (convert Xs)
        | "1" :: Xs => 1 + 10 * (convert Xs)
        | "2" :: Xs => 2 + 10 * (convert Xs)
        | "3" :: Xs => 3 + 10 * (convert Xs)
        | "4" :: Xs => 4 + 10 * (convert Xs)
        | "5" :: Xs => 5 + 10 * (convert Xs)
        | "6" :: Xs => 6 + 10 * (convert Xs)
        | "7" :: Xs => 7 + 10 * (convert Xs)
        | "8" :: Xs => 8 + 10 * (convert Xs)
        | "9" :: Xs => 9 + 10 * (convert Xs)
        | _ =>
           raise Error "Attempt to convert ill-formed string to an integer!!!";
  in
    (sign (explode s)) * (convert (rev (explode s)))
  end;


(* converse to the above.
   type: int -> string  *)

fun int_to_string n =
  let 
    val sign = if n >= 0 then "" else "~";

    fun convert_digit m =
      case m of
          0 => "0"
        | 1 => "1"
        | 2 => "2"
        | 3 => "3"
        | 4 => "4"
        | 5 => "5"
        | 6 => "6"
        | 7 => "7"
        | 8 => "8"
        | 9 => "9"
        | _ => raise Error "Impossible error!!!";

   fun convert (m: int) =
     if m <= 9
       then [convert_digit m]
       else (convert (m div 10)) @ [convert_digit (m mod 10)];
  in
    sign ^ (implode (convert (abs n)))
  end;


(* takes an ML integer to a representing Isabelle term in ExtBin.thy.
   type: int -> term  *)

fun int_to_term  0 = Const ("PlusSign",  INT)
  | int_to_term ~1 = Const ("MinusSign", INT)
  | int_to_term  i = 
  let
    fun to_bin P c 0 = c
      | to_bin P c n =
          Const ("Bcons", INT --> BOOL --> INT) $
            (to_bin  P c (n div 2)) $
            (if P n
               then Const ("True",  BOOL)
               else Const ("False", BOOL));
  in
    if i < 0
      then to_bin (fn x => (x mod 2) = 0)
                  (Const ("MinusSign", INT)) ((abs i) - 1)
      else to_bin (fn x => (x mod 2) = 1)
                  (Const ("PlusSign",  INT)) i
  end;


(* converse of the above.
   type: term -> int  *)

fun term_to_int term =
  let
    fun term_to_int' d a (Const ("PlusSign", Type ("int",[]))) =
          a
      | term_to_int' d a (Const ("MinusSign", Type ("int",[]))) =
          ~((2 pow d) - a)
      | term_to_int' d a (Const ("Bcons", Type ("fun",
          [_, Type ("fun", [Type ("bool",[]), Type ("int",[])])])) $ p $ b) =
          if b = (Const ("True", Type ("bool",[])))
            then term_to_int' (d + 1) (a + (2 pow d)) p
            else term_to_int' (d + 1) a p
      | term_to_int' _ _ _ =
          raise Error "Attempt to convert ill-formed term to an integer!!!";
  in
    term_to_int' 0 0 term
  end;


(* takes a non-negative integer to a representing Isabelle term.
   type: int -> term  *)

fun nat_to_term n =
  let
    fun nat_to_term' 0 = Const ("0", NAT)
      | nat_to_term' i = Const ("Suc", NAT --> NAT) $ (nat_to_term' (i - 1));
  in
    if n < 0
      then raise Error "Number is not natural!!!"
      else nat_to_term' n
  end;


(* converse of the above.
   type: term -> int  *)

fun term_to_nat term =
  let
    fun term_to_int' (Const ("0", Type ("nat",[]))) i =
          i
      | term_to_int' (Const ("Suc", Type ("fun",
                           [_, Type ("nat",[])])) $ pred) i =
          term_to_int' pred (i + 1)
      | term_to_int' _ _ =
          raise Error "Attempt to non-number to an integer!!!";
  in
    term_to_int' term 0
  end;


(* takes an Isabelle term to its outmost function symbol and its arguments.
   type: term -> term * term list  *)

fun term_split term =
  let
    fun term_split' terms (arg1 $ arg2) = term_split' (arg2 :: terms) arg1
      | term_split' terms arg           = arg :: terms;

    val split = term_split' [] term;
  in
    (hd split, tl split)
  end;


(* gives the name string of an Isabelle constant or variable.
   type: type -> string  *)

fun name_of_term (Const (nm, _))    = nm
  | name_of_term (Free (nm, _))     = nm
  | name_of_term (Var ((nm, _), _)) = nm
  | name_of_term (Bound i)          = "Bound " ^ (int_to_string i)
  | name_of_term (Abs (nm, _, _))   = "Lambda " ^ nm
  | name_of_term _                  = raise Error "illformed argument";


end; (* struct *)


