(* 
   file     : $CAT_DIR/SupInf/SupInf.ML
   platform : PolyML 3.2 and Isabelle94-8/HOL
   version  : 1.0b  (16.07.98)
   notes    : This file defines the language for the treatment of
              the extended part of Shostak's method. Types are being
              checked when translating from Isabelle term to BoolTerm.
*)

signature ArithSEM =
sig

(*
IntegConstant:
  - user defined constants:                     (<name>, 0)
  - uninterpreted arithmetical operation symb.: (<operator>, 1)
  - funct. symbols derived from preds.:         (<function>, 2)
  - constants of type nat. number:              (<term>, 3)
  - introduced constants for `zbool':           ("Bool", n) with n > 3
  - introduced variables:                       ("X", n) with n > 3
*)

  datatype IntegTerm =
      IntegConstant of (string * int) * IntegTerm list
    | IntegValue    of int
    | IntegComp     of IntegTerm
    | IntegPlus     of IntegTerm * IntegTerm
    | IntegMinus    of IntegTerm * IntegTerm
    | IntegTimes    of IntegTerm * IntegTerm
    | IntegDivide   of IntegTerm * IntegTerm
    | IntegModulus  of IntegTerm * IntegTerm
    | IntegPower    of IntegTerm * IntegTerm

  datatype BoolTerm =
      BoolTrue
    | BoolFalse
    | BoolPred    of string * IntegTerm list
    | BoolLess    of IntegTerm * IntegTerm
    | BoolLEq     of IntegTerm * IntegTerm
    | BoolEq      of IntegTerm * IntegTerm
    | BoolGEq     of IntegTerm * IntegTerm
    | BoolGreater of IntegTerm * IntegTerm
    | BoolNot     of BoolTerm
    | BoolAnd     of BoolTerm * BoolTerm
    | BoolOr      of BoolTerm * BoolTerm
    | BoolImp     of BoolTerm * BoolTerm


  val term_to_BoolTerm:  term -> BoolTerm

end;


structure ArithSemantics: ArithSEM =
struct

local

  open Basics;


(* provides unique numbers to generate new constant names *)

  val bool_count = ref 3;
  fun read_bc () = (bool_count := (!bool_count) + 1; !bool_count);


(* concatenate all elements of a list l1 to another list l2 which do not
   occure in l2; duplications of elements in l1 will be removed.
   type: ''a list -> ''a list -> ''a list  *)

  fun merge_to l []      = l
    | merge_to l (x::xs) = if exists (fn y => x = y) l
                             then merge_to l xs
                             else merge_to (x::l) xs;


(* reduces the argument list (and as such the arity) of a function
   symbol to expressions of type INT; choosing the name of the
   function be that; e.g., f(x+1,"s",8) becomes f_s_(x+1,8)
   type: term list -> string * term list  *)

  fun clean args =
    let
      fun clean' (i, a)  []     = (i, rev a)
        | clean' (i, a) (x::xs) =
            if (type_of x) = INT
              then clean' (i ^ "_", x::a) xs
              else clean' (i ^ (name_of_term x), a) xs;
    in
      clean' ("", []) args
    end;

in

(*
IntegConstant:
  - user defined constants:                     (<name>, 0)
  - uninterpreted arithmetical operation symb.: (<operator>, 1)
  - funct. symbols derived from preds.:         (<function>, 2)
  - constants of type nat. number:              (<term>, 3)
  - introduced constants for `zbool':           ("Bool", n) with n > 3
  - introduced variables:                       ("X", n) with n > 3
*)

  datatype IntegTerm =
      IntegConstant of (string * int) * IntegTerm list
    | IntegValue    of int
    | IntegComp     of IntegTerm
    | IntegPlus     of IntegTerm * IntegTerm
    | IntegMinus    of IntegTerm * IntegTerm
    | IntegTimes    of IntegTerm * IntegTerm
    | IntegDivide   of IntegTerm * IntegTerm
    | IntegModulus  of IntegTerm * IntegTerm
    | IntegPower    of IntegTerm * IntegTerm;

  datatype BoolTerm =
      BoolTrue
    | BoolFalse
    | BoolPred    of string * IntegTerm list
    | BoolLess    of IntegTerm * IntegTerm
    | BoolLEq     of IntegTerm * IntegTerm
    | BoolEq      of IntegTerm * IntegTerm
    | BoolGEq     of IntegTerm * IntegTerm
    | BoolGreater of IntegTerm * IntegTerm
    | BoolNot     of BoolTerm
    | BoolAnd     of BoolTerm * BoolTerm
    | BoolOr      of BoolTerm * BoolTerm
    | BoolImp     of BoolTerm * BoolTerm;


(* a natural number is mapped to an integer; a variable of type NAT
   is mapped to a new IntegConstant and an appropriate constraint;
   e.g., Suc(Suc(x)) becomes x' and the constraint ~2 <= x'.
   type: term -> IntegTerm * BoolTerm list  *)

  fun natTerm_to_IntegTerm term =
      let
        fun natTerm_to_IntegTerm' n (Const ("0", Type ("nat",[]))) = (n, None)
          | natTerm_to_IntegTerm' n (Const ("Suc", Type ("fun",
                              [Type ("nat",[]), Type ("nat",[])])) $ arg') =
              natTerm_to_IntegTerm' (n + 1) arg'
          | natTerm_to_IntegTerm' n arg' = (n, Some arg');

        val (v, r) = natTerm_to_IntegTerm' 0 term;
      in
        if r = None
          then
            (IntegValue v, [])
          else 
            let
              val nm = Sign.string_of_term (sign_of Nat.thy) (the r);
            in
              (IntegConstant ((nm, 3), []),
                [BoolLEq (IntegValue (~v), IntegConstant ((nm, 3), []))])
            end
      end;


(* maps Isabelle int terms to IntegTerm, where a list of constraints
   is generated.
   type: term -> IntegTerm * BoolTerm list  *)

  fun term_to_IntegTerm x =
    let
      fun to_IntegTerm constraints term =
          case term of
	    Const ("znat",
              Type ("fun",
                [_, Type ("int",[])])) $ arg =>
              let
                val (iterm, constrs) = natTerm_to_IntegTerm arg;
              in
                (iterm, merge_to constraints constrs)
              end
          | Const ("PlusSign", Type ("int",[])) =>
              (IntegValue 0, constraints)
          | Const ("MinusSign", Type ("int",[])) =>
              (IntegValue ~1, constraints)
          | Const ("zbool",
              Type ("fun",
                [Type ("bool",[]), Type ("int",[])])) $ arg =>
              let
                val iconst = IntegConstant (("Bool", read_bc ()), []);
                val bterm  = term_to_BoolTerm arg;
              in
                (iconst,
                merge_to constraints
                  [BoolAnd
                    (BoolImp (        bterm, BoolEq (iconst, IntegValue 1)),
                     BoolImp (BoolNot bterm, BoolEq (iconst, IntegValue 0)))])
              end
	  | Const ("zsuc",
              Type ("fun",
                [_, Type ("int",[])])) $ arg =>
              let
                val (iterm, constrs) = to_IntegTerm constraints arg;
              in
                (IntegPlus (iterm, IntegValue 1), constrs)
              end
	  | Const ("zpred",
              Type ("fun",
                [_, Type ("int",[])])) $ arg =>
              let
                val (iterm, constrs) = to_IntegTerm constraints arg;
              in
                (IntegPlus (iterm, IntegValue ~1), constrs)
              end
	  | Const ("zmagnitude",
              Type ("fun",
                [Type ("int",[]), Type ("int",[])])) $ arg =>
              let
              in
                case arg of
                    Const ("PlusSign", Type ("int",[])) =>
                      (IntegValue 0, constraints)
                  | Const ("MinusSign", Type ("int",[])) =>
                      (IntegValue 1, constraints)
                  | Const ("Bcons", Type ("fun",[
                                      Type ("int",[]),
                                      Type ("fun",[
                                        Type ("bool",[]),
                                        Type ("int",[])])])) $ a =>
                      (IntegValue (abs (term_to_int a)), constraints)
                  | _ => 
                      let
                        val (iterm, constr) = to_IntegTerm [] arg;
                        val constrs = merge_to constraints constr;
                      in
                        (IntegConstant (("zmagnitude:_:", 0), [iterm]),constrs)
                      end
              end
	  | Const ("zminus",
              Type ("fun",
                [_, Type ("int",[])])) $ arg =>
              let
                val (iterm, constrs) = to_IntegTerm constraints arg;
              in
                (IntegComp iterm, constrs)
              end
	  | Const ("op +",
              Type ("fun",
                [_, Type ("fun",
                  [_, Type ("int",[])])])) $ arg1 $ arg2 =>
              let
                val (iterm1, constrs1) = to_IntegTerm [] arg1;
                val (iterm2, constrs2) = to_IntegTerm [] arg2;
                val constrs = merge_to constraints (constrs1 @ constrs2);
              in
                (IntegPlus (iterm1, iterm2), constrs)
              end
	  | Const ("op -",
              Type ("fun",
                [_, Type ("fun",
                  [_, Type ("int",[])])])) $ arg1 $ arg2 =>
              let
                val (iterm1, constrs1) = to_IntegTerm [] arg1;
                val (iterm2, constrs2) = to_IntegTerm [] arg2;
                val constrs = merge_to constraints (constrs1 @ constrs2);
              in
                (IntegMinus (iterm1, iterm2), constrs)
              end
	  | Const ("op *",
              Type ("fun",
                [_, Type ("fun",
                  [_, Type ("int",[])])])) $ arg1 $ arg2 =>
              let
                val (iterm1, constrs1) = to_IntegTerm [] arg1;
                val (iterm2, constrs2) = to_IntegTerm [] arg2;
                val constrs = merge_to constraints (constrs1 @ constrs2);
              in
                (IntegTimes (iterm1, iterm2), constrs)
              end
	  | Const ("op zdiv",
              Type ("fun",
                [_, Type ("fun",
                  [_, Type ("int",[])])])) $ arg1 $ arg2 =>
              let
                val (iterm1, constrs1) = to_IntegTerm [] arg1;
                val (iterm2, constrs2) = to_IntegTerm [] arg2;
                val constrs = merge_to constraints (constrs1 @ constrs2);
              in
                (IntegDivide (iterm1, iterm2), constrs)
              end
	  | Const ("op zmod",
              Type ("fun",
                [_, Type ("fun",
                  [_, Type ("int",[])])])) $ arg1 $ arg2 =>
              let
                val (iterm1, constrs1) = to_IntegTerm [] arg1;
                val (iterm2, constrs2) = to_IntegTerm [] arg2;
                val constrs = merge_to constraints (constrs1 @ constrs2);
              in
                (IntegModulus (iterm1, iterm2), constrs)
              end
	  | Const ("op zexp",
              Type ("fun",
                [_, Type ("fun",
                  [_, Type ("int",[])])])) $ arg1 $ arg2 =>
              let
                val (iterm1, constrs1) = to_IntegTerm [] arg1;
                val (iterm2, constrs2) = to_IntegTerm [] arg2;
                val constrs = merge_to constraints (constrs1 @ constrs2);
              in
                (IntegPower (iterm1, iterm2), constrs)
              end
          | _ =>
              if (type_of term) = INT
                then
                  let
                    val (root, args) = term_split term;
                  in
                    case root of
                        Const ("Bcons", _) =>
                          (IntegValue (term_to_int term), constraints)
                      | _ =>
                          let
                            val (idxs, arg_list) = clean args;
                            val (iterm_list, constr_list) = 
                                  unzip (map (to_IntegTerm []) arg_list);
                            val constrs =
                                  merge_to constraints (flatten constr_list);
                            val nm = (name_of_term root) ^ ":" ^ idxs ^ ":";
                          in
                            (IntegConstant ((nm, 0), iterm_list), constrs)
                          end
                  end
                else
                  raise Error "term has illegal format!!!";
    in
      to_IntegTerm [] x
    end


(* maps an Isabelle bool term to BoolTerm, where certain integer operators
   are interpreted others are treated as arbitrary constants.
   type: term -> BoolTerm  *)

  and term_to_BoolTerm x =
    let
      fun conjoin []      = BoolTrue
        | conjoin [c]     = c
        | conjoin (c::cs) = BoolAnd (c, conjoin cs);
    in
        case x of
            Const ("True", Type ("bool",[])) =>
              BoolTrue
          | Const ("False", Type ("bool",[])) =>
              BoolFalse
          | Const ("op <",
              Type ("fun",
                [_, Type ("fun",
                  [_, Type ("bool",[])])])) $ arg1 $ arg2 =>
              let
                val (iterm1, constrs1) = term_to_IntegTerm arg1;
                val (iterm2, constrs2) = term_to_IntegTerm arg2;
                val constrs = merge_to constrs1 constrs2;
              in
                conjoin ((BoolLess (iterm1, iterm2))::constrs)
              end
          | Const ("op <=",
              Type ("fun",
                [_, Type ("fun",
                  [_, Type ("bool",[])])])) $ arg1 $ arg2 =>
              let
                val (iterm1, constrs1) = term_to_IntegTerm arg1;
                val (iterm2, constrs2) = term_to_IntegTerm arg2;
                val constrs = merge_to constrs1 constrs2;
              in
                conjoin ((BoolLEq (iterm1, iterm2))::constrs)
              end
          | Const ("op =",
              Type ("fun",
                [Type ("bool",[]), Type ("fun",
                  [Type ("bool",[]), Type ("bool",[])])])) $ arg1 $ arg2 =>
              let
                val bool1 = term_to_BoolTerm arg1;
                val bool2 = term_to_BoolTerm arg2;
              in
                BoolAnd (BoolImp (bool1, bool2), BoolImp (bool2, bool1))
              end
          | Const ("op =",
              Type ("fun",
                [_, Type ("fun",
                  [_, Type ("bool",[])])])) $ arg1 $ arg2 =>
              let
                val (iterm1, constrs1) = term_to_IntegTerm arg1;
                val (iterm2, constrs2) = term_to_IntegTerm arg2;
                val constrs = merge_to constrs1 constrs2;
              in
                conjoin ((BoolEq (iterm1, iterm2))::constrs)
              end
          | Const ("zgt",
              Type ("fun",
                [_, Type ("fun",
                  [_, Type ("bool",[])])])) $ arg1 $ arg2 =>
              let
                val (iterm1, constrs1) = term_to_IntegTerm arg1;
                val (iterm2, constrs2) = term_to_IntegTerm arg2;
                val constrs = merge_to constrs1 constrs2;
              in
                conjoin ((BoolGreater (iterm1, iterm2))::constrs)
              end
          | Const ("zge",
              Type ("fun",
                [_, Type ("fun",
                  [_, Type ("bool",[])])])) $ arg1 $ arg2 =>
              let
                val (iterm1, constrs1) = term_to_IntegTerm arg1;
                val (iterm2, constrs2) = term_to_IntegTerm arg2;
                val constrs = merge_to constrs1 constrs2;
              in
                conjoin ((BoolGEq (iterm1, iterm2))::constrs)
              end
          | Const ("Not",
              Type ("fun",
                [Type ("bool", []), Type ("bool",[])])) $ arg =>
              BoolNot (term_to_BoolTerm arg)
          | Const ("op &",
              Type ("fun",
                [_, Type ("fun",
                  [_, Type ("bool",[])])])) $ arg1 $ arg2 =>
              BoolAnd (term_to_BoolTerm arg1, term_to_BoolTerm arg2)
          | Const ("op |",
              Type ("fun",
                [_, Type ("fun",
                  [_, Type ("bool",[])])])) $ arg1 $ arg2 =>
              BoolOr (term_to_BoolTerm arg1, term_to_BoolTerm arg2)
          | Const ("op -->",
              Type ("fun",
                [_, Type ("fun",
                  [_, Type ("bool",[])])])) $ arg1 $ arg2 =>
              BoolImp (term_to_BoolTerm arg1, term_to_BoolTerm arg2)
          | Const ("Trueprop",
              Type ("fun", [Type ("bool",[]), Type ("prop",[])])) $ arg =>
              term_to_BoolTerm arg
          | Const ("Eps",
              Type ("fun",
                [Type ("fun", [_, Type ("bool",[])]), _])) $ _ =>
              raise Error "Attempt to use illegal term!!!"
          | Const ("All",
              Type ("fun",
                [Type ("fun", 
                   [_, Type ("bool",[])]), Type ("bool",[])])) $ _ =>
              raise Error "Attempt to use illegal term!!!"
          | Const ("Ex",
              Type ("fun",
                [Type ("fun", 
                   [_, Type ("bool",[])]), Type ("bool",[])])) $ _ =>
              raise Error "Attempt to use illegal term!!!"
          | Const ("Ex1",
              Type ("fun",
                [Type ("fun", 
                   [_, Type ("bool",[])]), Type ("bool",[])])) $ _ =>
              raise Error "Attempt to use illegal term!!!"
          | Const ("Least",
              Type ("fun",
                [Type ("fun", [_, Type ("bool",[])]), _])) $ _ =>
              raise Error "Attempt to use illegal term!!!"
          | Const ("op o",
              Type ("fun", [Type ("fun", [_, _]),
                Type ("fun", [Type ("fun", [_, _]),
                  Type ("fun", [_, _])])])) $ _ $ _ =>
              raise Error "Attempt to use illegal term!!!"
          | Const ("If",
              Type ("fun", [Type ("bool",[]),
                Type ("fun", [_, Type ("fun", [_, _])])])) =>
              raise Error "Attempt to use illegal term!!!"
          | Const ("Let",
              Type ("fun", [_,
                Type ("fun", [Type ("fun", [_, _]), _])])) $ _ $ _ =>
              raise Error "Attempt to use illegal term!!!"
          | Abs (a, T, u) =>
              let
                val (_, term) = variant_abs (a, T, u);
              in
                if (type_of term) = BOOL
                  then
                    term_to_BoolTerm term
                  else
                    raise Error "Attempt to use illegal term!!!"
              end
          | _ =>
              if (type_of x) = BOOL
                then
                  let
                    val (root, args) = term_split x;
                    val (iterms, constrs) = unzip (map term_to_IntegTerm args);
                  in
                    conjoin ((BoolPred (name_of_term root, iterms)) ::
                             (flatten constrs))
                  end
                else
                  raise Error "Attempt to use illegal term!!!"
    end;


end; (* local *)

end; (* struct *)
