(* 
   file     : $CAT_DIR/SupInf/ExtILPSolver.ML
   platform : PolyML 3.2 and Isabelle94-8/HOL
   version  : 1.0b  (16.07.98)
   notes    : This file implements Shostak's decision procedure for
              extended, quantifier-free Presburger arithmetic.
              See: JACM 26(2), 1979, pp. 351-360.
*)

signature ExtILPSOLV =
sig

  val solve: ILPSemantics.ILP list -> bool

end;


structure ExtILP : ExtILPSOLV =
struct

local
  
  open Basics;
  open ArithSemantics;
  open ILPSemantics;
  open ILPSolver;


(* exception for search for solutions *)
  exception NoSolution;

(* auxiliary function - equality of pairs
  type: (''a * ''b) * (''a * ''b) -> bool  *) 

  infix equals;
  fun (x1, y1) equals (x2, y2) = (x1 = x2) andalso (y1 = y2);


(* auxiliary function - just a patch up to get the `ceil' we actually want.
   type: real -> int  *)

  fun my_ceil n = if (real (Real.floor n) = n)
                    then Real.floor n
                    else Real.ceil n;


(* given an intervall [i, k] `Some j' is returned if i <= j <= k
   and j is an integer; we get `None' if no integer in [i, k] is found.
   type: Expr * Expr -> int  *)

  fun get_solution (Number m, Number n)   =
        if (Real.floor n) < (my_ceil m)
          then raise NoSolution
          else my_ceil m
    | get_solution (Number m, pInfinity)  = my_ceil m
    | get_solution (mInfinity, pInfinity) = 0
    | get_solution (mInfinity, Number n)  = Real.floor n
    | get_solution  _                     = raise NoSolution;


(* substitutes `Number n' for a `Variable v' in expression `e'.
   type: string -> real -> Expr -> Expr  *)

  fun num_for_var v n e =
        case e of
            pInfinity    => pInfinity
          | mInfinity    => mInfinity
          | Number m     => Number m
          | Variable s   => if (s = v) then (Number n) else (Variable s)
          | Times (m, x) => Times (m, num_for_var v n x)
          | Plus (x, y)  => Plus (num_for_var v n x, num_for_var v n y)
          | Max (x, y)   => Max (num_for_var v n x, num_for_var v n y)
          | Min (x, y)   => Min (num_for_var v n x, num_for_var v n y);


(* given a conjunction of inequalities a valid assignment for each variable
   is searched for by calculating the supremum and infimum and testing for
   an integer value between them; the supremum and infimum for some variable x
   is calculated in dependence of other variables that are `greater' than x
   with respect to some well-ordering (position in list `tbles'); a list
   of (<IntegConstant represented by variable>, <IntegValue of solution>)
   tuples is returned for each variable, a valid assignment could be found for.
   type: ILP -> (IntegTerm * IntegTerm) list *)

  fun solve_ilp (S, tbles) =
    let
      (* calculates supremum and infimum for a list of variables `j::H'
         constrained by `S'; for each variable `j' the following tuple
         is returned: (j, (<inf of j>, <sup of j>)).
         type: Ineq list -> string list -> (string * (Expr * Expr)) list  *)

      fun supinf c []     = c
        | supinf c (j::H) = supinf ((j, (Inf S j H, Sup S j H)) :: c) H;

      (* calculates inf and sup in `S' for every variable in `tbles' *)

      val ranges = supinf [] (fst (unzip tbles));

      (* given a list of solutions `sols' (tuples of a variable name and an
         integer assigned to it) and an expression `e', all variables of
         `sols' that occur in `e' are replaced by their assigned values
         and the resulting expression is simplified.
         type: (string * int) list -> Expr -> Expr  *)

      fun simp sols e =
            foldl (fn (x, (v, n)) => Simp (num_for_var v (real n) x))
                  (e, sols);

      (* given a list of variables `sols', which has already got a solution,
         the solution (if it exists) for variable `j' is calculated, depending
         on `inf' and `sup' of `j'; the solution of `j' is added to `sols'.
         type: (string * int) list * (string * (Expr * Expr)) ->
                 (string * int) list  *)

      fun calc_sol (sols, (j, (inf, sup))) =
            (j, get_solution (simp sols inf, simp sols sup)) :: sols;

      (* calculates solutions from a list of infimums and supremums *)

      val solutions = foldl calc_sol ([], ranges);

      (* maps a solution (a tuple of a variable name and an integer
         assigned to it) to an IntegVariable and IntegValue.
         type: string * int -> IntegTerm * IntegTerm  *)

      fun format (j, n) = (the (lookupl j tbles), IntegValue n);

    in
      map format solutions
    end;


(* given a list of IntegTerms (only IntegConstants and IntegValues allowed),
   a list of all subterms is returned; all duplicate terms are removed.
   type: IntegTerm list -> IntegTerm list  *)

  fun subterms_of terms =
    let
      fun subterms_of' (IntegConstant (nmi, l)) =
            (IntegConstant (nmi, l)) :: (flatten (map subterms_of' l))
        | subterms_of' (IntegValue _) =
            []
        | subterms_of' _ =
             raise Error "Attempt to solve non-normalized ILP!!!";
    in
      rm_duplicates op= (flatten (map subterms_of' terms))
    end;

  fun arity_of (IntegConstant (_, x)) = length x
    | arity_of  _  = raise Error "Internal error while solving ILPs!!!";

  fun args_of (IntegConstant (_, x)) = x
    | args_of  _  = raise Error "Internal error while solving ILPs!!!";


(* searches for substituitivity violations in the ILP solution `TS' with
   the variables `tbl'.
   type: (IntegTerm * IntegTerm) list -> Transl list ->
           ((IntegTerm * IntegTerm) *
           ((IntegTerm * IntegTerm) * Transl list) list) option  *)

  fun get_violations TS tbl =
    let
      (* separates the solutions `S' from the terms `T' it assigns values to
         type: IntegTerm list * IntegTerm list  *)
      val (T, S) = unzip TS;

      (* the set of all subterms of `T'
         type: IntegTerm list  *)
      val U = subterms_of T;

      (* consists of a list `e' of IntegTerms for each element `t' in U;
         the head of `e' is `t' itself and the tail of `e' consists
         of all terms in `T' that have the same solution as `t' has;
         the tail of `e' is the empty list otherwise.
         type: IntegTerm list list  *)
      val E =
        let
          (* gets terms with an equal solution *)
          fun collect_eqs term =
            let
              (* get solution for `term', if any *)
              val sol = lookupl term TS;
            in
              if sol = None
                then [term]
                else term :: 
                       (flatten
                         (map (fn (x, y) =>
                                 if (y = the sol) andalso (not (x = term))
                                   then [x]
                                   else [])
                               TS))
            end;
        in
          map collect_eqs U
        end;


      (* consists of those elements of `E' that have an actual solution.
         type: IntegTerm list list  *)
      val ET = filter (fn x => exists (fn y => (hd x) = y) T) E;

      fun not_eqs []      = []
        | not_eqs (x::xs) =
            x :: (filter (fn y => not (exists (secl y op=) (x::xs))) T);

      fun root_of (IntegConstant (nmi, _)) = nmi
        | root_of  _  = raise Error "Internal error while solving ILPs!!!";

      fun same_root []      = []
        | same_root (x::xs) =
            x :: (filter (fn y => (root_of x) equals (root_of y)) xs);

      (* consists of lists of IntegTerms which have the same outermost
         function symbol, but different solutions.
         type: IntegTerm list list  *)
      val N = map (same_root o not_eqs) ET;

      (* consists of IntegTerm tuples with the same outermost symbol and
         different solutions.
         type: IntegTerm list list  *)
      val M = flatten (map (fn l => map (pair (hd l)) (tl l))
                           (filter ((secl 1 op<) o length) N));

      (* consists of lists of expression pairs potentially violating
         substituitivity
         type: (IntegTerm * IntegTerm) list list  *)
      val P = rm_duplicates
                (fn (x, y) => (x equals y) orelse (x equals (swap_pair y)))
                M;

      fun get_E tx = hd (filter ((secl tx op=) o hd) E);

      fun there_are_no_holes_in []       = true
        | there_are_no_holes_in ([]:: _) = false
        | there_are_no_holes_in (_ ::xs) = (there_are_no_holes_in xs);

      (* tries to find terms that is equal to `t1' and `t2'.
         type: IntegTerm * IntegTerm * (IntegTerm * IntegTerm) list ->
                 (IntegTerm * IntegTerm) list  *)
      fun EQPAIRS' (t1, t2, alreadytried) =
        let
          val E_t1 = get_E t1;
          val E_t2 = get_E t2;
          val roots1 = rm_duplicates (op equals) (map root_of E_t1); 
          val roots2 = rm_duplicates (op equals) (map root_of E_t2);

          (* list of function symbols which applications are equal to
             both `t1' and `t2' *)
          val common_roots = intersect (op equals) (roots1, roots2);

          (* chooses first pair with same arity of terms and
             all their arguments are equal.
             type: (IntegTerm * IntegTerm) list ->
                   ((IntegTerm * IntegTerm) * (IntegTerm * IntegTerm) list)
                   option  *)
          fun choose' []             = None
            | choose' ((f1, f2)::fs) =
              let
                fun eqs_of args =
                      map (fn (a, b) => EQPAIRS (a, b, (t1, t2)::alreadytried))
                          (zip args);
              in
                if (arity_of f1) = (arity_of f2)
                  then 
                    let
                      (* checks for arguments to be equal *)
                      val eqs = eqs_of (args_of f1, args_of f2);
                    in
                      (* are all arguments equal? *)
                      if there_are_no_holes_in eqs
                        then Some ((f1, f2),
                                   rm_duplicates (op equals) (flatten eqs))
                        else choose' fs
                    end
                  else choose' fs
              end;

          (* gives first pair of IntegTerms which shows that `t1' and `t2'
             are equal.
             type: (string, int) list -> (IntegTerm * IntegTerm) list  *)
          fun choose []      = None
            | choose (r::rs) =
              let
                val rs1 = filter (fn x => ((root_of x) equals r) andalso
                                          ((arity_of x) > 0))
                                  E_t1;
                val rs2 = filter (fn x => ((root_of x) equals r) andalso
                                          ((arity_of x) > 0))
                                  E_t2;

                val choice' =
                      choose' (flatten (map (fn x => map (pair x) rs2) rs1));
              in
                if choice' = None
                  then choose rs
                  else choice'
              end;

          (* list of IntegTerm pairs which are equal and which contains `t1'
             and `t2', if any solution exists. *)
          val choice = choose common_roots;

          fun P1 x = if t1 = x then [] else [(t1, x)];
          fun P2 x = if t2 = x then [] else [(t2, x)];
        in
          if choice = None
            then []
            else (P1 (fst (fst (the choice))))
               @ (P2 (snd (fst (the choice))))
               @ (snd (the choice))
        end


      (* checks violation of substituitivity for two IntegTerms `t1'
         and `t2'; returns IntegTerm pairs that prove that `t1' and `t2'
         are equal.
         type: IntegTerm * IntegTerm * (IntegTerm * IntegTerm) list ->
                 (IntegTerm * IntegTerm) list  *)
      and EQPAIRS (IntegValue n, IntegValue m, _) =
            if n = m then [(IntegValue n, IntegValue m)] else []
        | EQPAIRS (IntegValue n, _, _) =
            []
        | EQPAIRS (_, IntegValue m, _) =
            []
        | EQPAIRS (t1, t2, alreadytried) =
            if (exists (secl (t1, t2) (op equals)) alreadytried)
              then []  (* no violation of substituitivity exists *)
              (* Is `t1' in the set of terms equal to `t2'? *)
              else if (exists (secl t1 op=) (get_E t2))
                     then
                       [(t1, t2)]
                     else (* tries to find a term which is equal to `t1'
                             and `t2' *)
                       EQPAIRS' (t1, t2, alreadytried);

      (* checks whether there IntegTerms `t1' and `t2' appear in the
         translation table, and returns their entries if necessary.
         type: IntegTerm * IntegTerm ->
               (IntegTerm * IntegTerm) * Transl list  *)
      fun tbl_annotation (t1, t2) =
        let
          val a1  = lookupr t1 tbl;
          val as1 = if a1 = None then [(IntegConst_to_string t1, t1)] else [];
          val a2  = lookupr t2 tbl;
          val as2 = if a2 = None then [(IntegConst_to_string t2, t2)] else [];
        in
          ((t1, t2), as1 @ as2)
        end;

      (* the first IntegTerm pair that is found violating substituitivity
         (`t1', `t2') is given as result together with the terms that
         prove that `t1' should be equal to `t2' and the appropriate additions
         to their translation tables.
         type: (IntegTerm * IntegTerm) list ->
               ((IntegTerm * IntegTerm) *
                ((IntegTerm * IntegTerm) * Transl list) list) option  *)
      fun get_violations'  []            = None
        | get_violations' ((t1, t2)::ts) =
          let
            val eqpairs = EQPAIRS (t1, t2, []);
          in
            if eqpairs = []
              then get_violations' ts
              else Some ((t1, t2), map tbl_annotation eqpairs)
          end;
    in
      (* `P' consists of IntegTerm pairs that potentially violate
         substituitivity. *)
      get_violations' P
    end;


(* tests a solution `TS' for a conjunction of inequalities `ineqs'
   for possible substituitivity violations; it returns a new linear
   programming problem which removes the violations by adding new
   constraints.
   type: Ineq list * Transl list -> (IntegTerm * IntegTerm) list -> ILP list *)

  fun violations (ineqs, tbl) TS =
    let
      (* list of IntegTerm pairs which prove a violation of substituitivity
         in the solution TS, if any *)
      val vs = get_violations TS tbl;

      fun trans term ttbl = Variable (the (lookupr term ttbl))
            handle e => raise Error "Internal error while solving ILPs!!!";

      fun mk_ilps ((t1, t2), probs) = 
        let

          val ilp1 = ((trans t1 tbl, trans t2 tbl) ::
                      (trans t2 tbl, trans t1 tbl) :: ineqs, tbl);

          fun g ((x1, x2), tx) =
            let
              val table = tbl @ tx;
            in
               (((Plus (trans x1 table, Number 1.0), trans x2 table) :: ineqs),
                 table) :: 
              [(((Plus (trans x2 table, Number 1.0), trans x1 table) :: ineqs),
                 table)]
            end;
        in
          ilp1 :: (flatten (map g probs))
        end;
    in
      if vs = None
        then (* no violation exists *)
          []
        else (* ILPs are built that encode violations *)
          mk_ilps (the vs)
    end;

in

(* taking a list of conjunctions this function tests whether any of these
   is satisfiable; each conjunction is represented as a list of
   inequalities and a translation table that relates variable names
   with the corresponding IntegConstant.
   type: ILP list -> bool  *)

  fun solve ilps =
    let

      (* if no variables exist, we just check for integer solutions.
         type: ILP -> bool  *)
      fun solve' (conj, []) =
        let
          fun dummy_function (x, y) =
            let
              val dummy = get_solution (fst y, snd y);
            in
              x
            end;
        in
          (foldl dummy_function (true, conj)
           handle NoSolution => false)
        end

      (* if a valid assignment for every variable is found, it'll be checked
         for possible substituitivity violations, if there aren't any,
         we've got a counter example and the investigated formula is not
         a tautology. *)

        | solve' (conj, tbl) =
        let
          (* `TS' is a list of (<IntegConstant>, <solution>) tuples
             type: (IntegTerm * IntegTerm) list  *)

          val TS = solve_ilp (conj, tbl) handle NoSolution => [];

        in
          if (length TS) < (length tbl)
            then
              false
            else
              let
                val violns = violations (conj, tbl) TS;
              in
                if violns = []
                  then
                    true
                  else
                    foldl (fn (x, y) => x orelse (solve' y))
                          (false, violns)
              end
        end;
    in
      foldl (fn (x, y) => x orelse (solve' y)) (false, ilps)
    end;

end;  (* local *)

end;  (* struct *)
