(* 
   file     : $CAT_DIR/SupInf/ILPSemantics.ML
   platform : PolyML 3.2 and Isabelle94-8/HOL
   version  : 1.0b  (16.07.98)
   notes    : This file defines the language for the treatment of
              Shostak's method.
*)

signature ILPSEM =
sig

  datatype Expr =
      pInfinity
    | mInfinity
    | Number   of real
    | Variable of string
    | Times    of real * Expr
    | Plus     of Expr * Expr
    | Max      of Expr * Expr
    | Min      of Expr * Expr

  type Ineq
  type Transl
  type ILP

  val IntegConst_to_string: ArithSemantics.IntegTerm -> string

  val IntegTerm_to_Expr: ArithSemantics.IntegTerm -> Expr * Transl list

  val DNF_to_ILPs: ArithSemantics.BoolTerm -> ILP list

end;


structure ILPSemantics: ILPSEM =
struct

local

  open Basics;
  open ArithSemantics;

in

  datatype Expr =
      pInfinity
    | mInfinity
    | Number   of real
    | Variable of string
    | Times    of real * Expr
    | Plus     of Expr * Expr
    | Max      of Expr * Expr
    | Min      of Expr * Expr;


(* less-then-or-equal inequality *)

  type Ineq = Expr * Expr;


(* relates a string to the IntegConstant represented by it *)

  type Transl = string * IntegTerm;


(* Integer Linear Programming problems *)

  type ILP = Ineq list * Transl list;


  fun Expr_to_string e =
        case e of
            pInfinity => "+inf"
          | mInfinity => "-inf"
          | Number n => PolyML.makestring n
          | Variable nm => nm
          | Times (nf, e) => (PolyML.makestring nf) ^ " * " ^
                             (Expr_to_string e)
          | Plus (e1, e2) => (Expr_to_string e1) ^ " + " ^ (Expr_to_string e2)
          | Max (e1, e2) => "Max (" ^ (Expr_to_string e1) ^ ", " ^
                            (Expr_to_string e2) ^ ")"
          | Min (e1, e2) => "Min (" ^ (Expr_to_string e1) ^ ", " ^
                            (Expr_to_string e2) ^ ")";

(* forms a string from an inequality
   type: Ineq -> string  *)

  fun ineq_to_string (e1, e2) =
        (Expr_to_string e1) ^ " <= " ^ (Expr_to_string e2);


(* forms a string from a list of inequalities
   type: Ineq list -> string  *)

  fun conj_to_string []  = ""
    | conj_to_string [c] = "(" ^ (ineq_to_string c) ^ ")"
    | conj_to_string (c::xs) = "(" ^ (ineq_to_string c) ^ ") &\n" ^
                               (conj_to_string xs);


(* forms a string from constants (partial map over Integterm)
   type: IntegTerm -> string  *)

  fun IntegConst_to_string (IntegConstant ((nm, i), l)) =
      let
        fun IntegConst_to_string' [] =
               ""
          | IntegConst_to_string' [x] =
               IntegConst_to_string x
          | IntegConst_to_string' (x::xs) =
              (IntegConst_to_string x) ^ "," ^ (IntegConst_to_string' xs)
      in
        nm ^ "_" ^ (int_to_string i) ^ "(" ^ (IntegConst_to_string' l) ^ ")"
      end
    | IntegConst_to_string (IntegValue n) =
        int_to_string n
    | IntegConst_to_string _ =
        raise Error "Attempt to use non-nomalized term!!!";


(* map from IntegConstant or IntegValue to string (returns "???" else).
   type: IntegTerm -> string  *)

  fun IntegAtom_to_string (IntegValue n) =
        int_to_string n
    | IntegAtom_to_string (IntegConstant c) =
        IntegConst_to_string (IntegConstant c)
    | IntegAtom_to_string _ = "???";


(* map from IntegTerm to Expr
   type: IntegTerm -> Expr  *)

  fun IntegTerm_to_Expr e =
    let
      fun IntegTerm_to_Expr' l x =
            case x of
                IntegConstant _ =>
                  let
                    val str = IntegConst_to_string x;
                  in
                    (Variable str, (str, x) :: l)
                  end
              | IntegValue n =>
                  (Number (real n), l)
              | IntegComp (IntegConstant y) =>
                  let
                    val (foo, boo) = IntegTerm_to_Expr' [] (IntegConstant y);
                  in
                    (Times (~1.0, foo), l @ boo)
                  end
              | IntegPlus (y, z) =>
                  let
                    val (foo1, boo1) = IntegTerm_to_Expr' [] y;
                    val (foo2, boo2) = IntegTerm_to_Expr' [] z;
                  in
                    (Plus (foo1, foo2), l @ boo1 @ boo2)
                  end
              | IntegTimes (IntegValue y, z) => 
                  let
                    val (foo, boo) = IntegTerm_to_Expr' [] z;
                  in
                    (Times (real y, foo), boo)
                  end
              | _ => raise Error "Attempt to use non-nomalized term!!!";

      val (gueck, stueck) = IntegTerm_to_Expr' [] e;
    in
      (gueck, rm_duplicates (fn (g, s) => (fst g) = (fst s)) stueck)
    end;


(* takes formula in DNF to list representation, e.g.,
     ((x1 <= x2) | (x3 <= x4)) & (x5 <= x6)  =>
       [[(x1, x2), (x3, x4)], [(x5, x6)]]
   type: BoolTerm -> (IntegTerm * IntegTerm) list list  *)

  fun BoolTerm_to_NFInteg term =
    let
      fun brk_ineq (BoolLEq (e, z)) =
            (e, z)
        | brk_ineq _ =
            raise Error "Attempt to use non-normalized term!!!";

      fun brk_conj x =
        let
          fun brk_conj' l (BoolAnd (e, z)) =
                (brk_conj' [] e) @ (brk_conj' [] z)
            | brk_conj' l w  =
                [w];
        in
          map brk_ineq (brk_conj' [] x)
        end;

      fun brk_disj x =
        let
          fun brk_disj' l (BoolOr (e, z)) =
                (brk_disj' [] e) @ (brk_disj' [] z)
            | brk_disj' l w  =
                [w];
        in
          map brk_conj (brk_disj' [] x)
        end;
    in
      brk_disj term
    end;


(* translates IntegTerm to Expr in a list representation of DNF formula,
   and also attaches a translation table between Variable names and
   IntegConstants to each conjunction
   type: (IntegTerm * IntegTerm) list list  ->
           ((Expr * Expr) list * (string * IntegTerm) list) list  *)

  fun NFInteg_to_NFExpr nf =
    let
      fun map_pair (x, y) =
        let
          val (x', tx) = IntegTerm_to_Expr x;
          val (y', ty) = IntegTerm_to_Expr y;
        in
          ((x', y'), rm_duplicates (fn (g, s) => (fst g) = (fst s)) (tx @ ty))
        end;

      fun map_disj x =
        let
          val (ps, ts) = unzip (map map_pair x);
        in
          (ps, rm_duplicates (fn (g, s) => (fst g) = (fst s)) (flatten ts))
        end;
    in
      map map_disj nf
    end;


(* translates a DNF formula to a list of ILPs.
   type: BoolTerm -> ILP list  *)

  fun DNF_to_ILPs formula = (NFInteg_to_NFExpr o BoolTerm_to_NFInteg) formula;


end;  (* local *)

end;  (* struct *)




