(* 
   file     : $CAT_DIR/SupInf/ILPSolver.ML
   platform : PolyML 3.2 and Isabelle94-8/HOL
   version  : 1.0b  (16.07.98)
   notes    : This file implements Shostak's decision procedure for
              quantifier-free Presburger arithmetic.
              See: "On the SUP-INF Method for Proving Presburger Formulas",
                   by R.E.Shostak JACM 24(4), 1977, pp. 529-543.
*)

signature ILPSOLV =
sig

  val Simp: ILPSemantics.Expr -> ILPSemantics.Expr

  val Sup: (ILPSemantics.Expr * ILPSemantics.Expr) list -> string -> 
             string list -> ILPSemantics.Expr

  val Inf: (ILPSemantics.Expr * ILPSemantics.Expr) list -> string ->
             string list -> ILPSemantics.Expr

end;

structure ILPSolver: ILPSOLV =
struct

  open Basics;
  open ILPSemantics;


(* the occurence of variable `v' in expression `e' is tested.
   type: string * Expr -> bool  *)

  infixr occurs_in;
  fun v occurs_in e =
    case e of
        Number n =>
          false
      | pInfinity =>
          false
      | mInfinity =>
          false
      | Variable v' =>
          if v = v' 
            then true
            else false
      | Times (_, e') =>
          v occurs_in e'
      | Plus (e1, e2) =>
          (v occurs_in e1) orelse (v occurs_in e2)
      | Max (e1, e2) =>
          (v occurs_in e1) orelse (v occurs_in e2)
      | Min (e1, e2) =>
          (v occurs_in e1) orelse (v occurs_in e2);


(* some rewriting to simplify and normalize expressions *)

      (* Assoc., Commut., and Distr. axioms *)
      fun
          (*  x(y + z) => xy + xz  *)
          DistAxiom (Times (x, Plus (y, z))) =
            Plus (Times (x, y), Times (x, z))
          (*  n + xy => xy + n  *)
        | DistAxiom (Plus (Number n, Times (x, y))) =
            Plus (Times (x, y), Number n)
          (*  n + (xy + z) => xy + (n + z)  *)
        | DistAxiom (Plus (Number n, Plus (Times (x, y), z))) =
            Plus (Times (x, y), Plus (Number n, z))
	  (*  (x + y) + z => x + (y + z)  *)
        | DistAxiom (Plus (Plus (x, y), z)) =
            Plus (x, Plus (y, z))
	  (*  x1*v1 + x2*v2 and v1 < v2 => x2*v2 + x1*v1  *)
        | DistAxiom (Plus (Times (x1, Variable v1), Times (x2, Variable v2))) =
            if v1 < v2
              then Plus (Times (x2, Variable v2), Times (x1, Variable v1))
              else Plus (Times (x1, Variable v1), Times (x2, Variable v2))
	  (*  x1*v1 + (x2*v2 + z) and v1 < v2 => x2*v2 + (x1*v1 + z)  *)
        | DistAxiom (Plus (Times (x1, Variable v1), 
                           Plus (Times (x2, Variable v2), z)))=
            if v1 < v2
              then Plus (Times (x2, Variable v2),
                         Plus (Times (x1, Variable v1), z))
              else Plus (Times (x1, Variable v1),
                         Plus (Times (x2, Variable v2), z))
          (*  +inf + (x*v + y) => x*v + (+inf + y)  *)
        | DistAxiom (Plus (pInfinity, Plus (Times (x, v), y))) =
            Plus (Times (x, v), Plus (pInfinity, y))
          (*  -inf + (x*v + y) => x*v + (-inf + y)  *)
        | DistAxiom (Plus (mInfinity, Plus (Times (x, v), y))) =
            Plus (Times (x, v), Plus (mInfinity, y))
          (*  -inf + +inf => +inf + -inf  *)
        | DistAxiom (Plus (mInfinity, pInfinity)) =
            Plus (pInfinity, mInfinity)
          (*  else  *)
        | DistAxiom x = x;


      (* `Max' axioms *)
      fun 
          (*  a + Max (b, c) => Max (a + b, a + c)  *)
	  MaxAxiom (Plus (a, Max (b, c))) =
            Max (Plus (a, b), Plus (a, c))
	  (*  Max (a, b) + c => Max (a + c, b + c)  *)
        | MaxAxiom (Plus (Max (a, b), c)) =
            Max (Plus (a, c), Plus (b, c))
	  (*  a * Max (b, c) and a >= 0  => Max (a * b, a * c) else 
                                            Min (a * b, a * c)  *)
        | MaxAxiom (Times (a, Max (b, c))) =
            if a >= 0.0
              then Max (Times (a, b), Times (a, c))
              else Min (Times (a, b), Times (a, c))
	  (*  Max (Max (a, b), c) => Max (a, Max (b, c))  *)
        | MaxAxiom (Max (Max (a, b), c)) =
            Max (a, Max (b, c))
          (*  else  *)
        | MaxAxiom x = x;


      (* `Min' axioms *)
      fun 
	  (*  a + Min (b, c) => Min (a + b, a + c)  *)
	  MinAxiom (Plus (a, Min (b, c))) =
            Min (Plus (a, b), Plus (a, c))
	  (*  Min (a, b) + c => Min (a + c, b + c)  *)
        | MinAxiom (Plus (Min (a, b), c)) =
            Min (Plus (a, c), Plus (b, c))
	  (*  a * Min (b, c) and a >= 0  => Min (a * b, a * c) else
                                            Max (a * b, a * c)  *)
        | MinAxiom (Times (a, Min (b, c))) =
            if a >= 0.0
              then Min (Times (a, b), Times (a, c))
              else Max (Times (a, b), Times (a, c))
	  (*  Min (Min (a, b), c) => Min (a, Min (b, c))  *)
        | MinAxiom (Min (Min (a, b), c)) =
            Min (a, Min (b, c))
          (*  else  *)
        | MinAxiom x = x;


      (* calculation axioms *)
      fun 
          (*  c1 * (c2 * a) => c1c2 * a  *)
	  DiagAxiom (Times (c1, Times (c2, a))) =
            Times (c1 * c2, a)
	  (*  c * n => cn  *)
        | DiagAxiom (Times (c, Number n)) =
            Number (c * n)
          (*  n1 + n2 => n1+n2  *)
        | DiagAxiom(Plus (Number n1, Number n2)) =
            Number (n1 + n2)
          (*  Min (n1, n2) and n1 < n2 => n1 else n2  *)
        | DiagAxiom (Min (Number n1, Number n2)) =
            if n1 < n2
              then Number n1
              else Number n2
          (*  Max (n1, n2) and n1 < n2 => n2 else n1  *)
        | DiagAxiom (Max (Number n1, Number n2)) =
            if n1 < n2
              then Number n2
              else Number n1
          (*  (x1 * v1) + (x2 * v2) and v1 = v2  =>  (x1+x2) * v1  *)
        | DiagAxiom (Plus (Times (x1, Variable v1), Times (x2, Variable v2))) =
            if v1 = v2
              then Times (x1 + x2, Variable v1)
              else Plus (Times (x1, Variable v1), Times (x2, Variable v2))
          (*  (x1 * v1) + ((x2 * v2) + z) and
                                  v1 = v2  =>  ((x1+x2) * v1) + z) *)
        | DiagAxiom (Plus (Times (x1, Variable v1),
                           Plus (Times (x2, Variable v2), z))) =
            if v1 = v2
              then Plus (Times (x1 + x2, Variable v1), z)
              else Plus (Times (x1, Variable v1),
                         Plus (Times (x2, Variable v2), z))
          (*  0 * a => 0  *)
        | DiagAxiom (Times (0.0, a)) =
            Number 0.0
          (*  Max (a, -inf)  =>  a  *)
        | DiagAxiom (Max (a, mInfinity)) =
            a
          (*  Max (-inf, a)  =>  a  *)
        | DiagAxiom (Max (mInfinity, a)) =
            a
          (*  Max (a, +inf)  =>  +inf  *)
        | DiagAxiom (Max (a, pInfinity)) =
            pInfinity
          (*  Max (+inf, a)  =>  +inf  *)
        | DiagAxiom (Max (pInfinity, a)) =
            pInfinity
          (*  Min (a, +inf)  =>  a  *)
        | DiagAxiom (Min (a, pInfinity)) =
            a
          (*  Min (+inf, a)  =>  a  *)
        | DiagAxiom(Min(pInfinity,a)) =
            a
          (*  Min (a, -inf)  =>  -inf  *)
        | DiagAxiom (Min (a, mInfinity)) =
            mInfinity
          (*  Min (-inf, a)  =>  -inf  *)
        | DiagAxiom (Min (mInfinity, a)) =
            mInfinity
          (*  a * +inf  and a > 0  =>  +inf  elseif a < 0  => -inf
                                                 else  a * +inf  *) 
        | DiagAxiom (Times (a, pInfinity)) =
            if a > 0.0
              then pInfinity
              else if a < 0.0
                     then mInfinity
                     else Times (a, pInfinity)
          (*  a * -inf  and a > 0  =>  -inf  elseif a < 0  => +inf
                                                 else  a * -inf  *) 
        | DiagAxiom (Times (a, mInfinity)) =
            if a > 0.0
              then mInfinity
              else if a < 0.0
                     then pInfinity
                     else Times (a, mInfinity)
          (*  +inf + a  =>  +inf  *)
        | DiagAxiom (Plus (pInfinity, Number a)) =
            pInfinity
          (*  a + +inf  =>  +inf  *)
        | DiagAxiom (Plus (Number a, pInfinity)) =
            pInfinity
          (*  -inf + a  =>  -inf  *)
        | DiagAxiom (Plus (mInfinity, Number a)) =
            mInfinity
          (*  a + -inf  =>  -inf  *)
        | DiagAxiom (Plus (Number a, mInfinity)) =
            mInfinity
          (*  +inf + +inf => +inf  *)
        | DiagAxiom (Plus (pInfinity, pInfinity)) =
            pInfinity
          (*  -inf + -inf => -inf  *)
        | DiagAxiom (Plus (mInfinity, mInfinity)) =
            mInfinity
	(*  else  *)
        | DiagAxiom x = x;


        val Axioms =
              DistAxiom
            o MaxAxiom
            o MinAxiom
            o DiagAxiom;


(* distributes `*' over `+' and pulles `Max' and `Min' out.
   type: Expr -> Expr  *)

  fun Simp J = 
    let
        (* top-down application of above axioms *)
        fun applyAxioms J =
              if J = Axioms J
                then case J of
                         Number _      => J
                       | pInfinity     => J
                       | mInfinity     => J
                       | Variable _    => J
                       | Times (x, y)  => Times (x, applyAxioms y)
                       | Plus (x, y)   => Plus (applyAxioms x, applyAxioms y)
                       | Max (x, y)    => Max (applyAxioms x, applyAxioms y)
                       | Min (x, y)    => Min (applyAxioms x, applyAxioms y)
                else applyAxioms (Axioms J);

        val J' = applyAxioms J;
    in
      if J = J'
        then J
        else Simp J'
    end;


(* simplifies expressions
   type: Expr -> Expr  *)
  fun SIMP J = 
        if J mem [pInfinity, mInfinity]
          then J
          else Simp (Plus (Times (1.0, J), Number 0.0));


(* brings Variable `x' to the left in some expression; the following rules
   are applied: av + bx becomes bx + av  and  av + (bx + e) becomes
   bx + (av + e),  where e is an expression, v is a variable, and
   a, b are constants.
   type: string * Expr -> Expr  *)

  fun PermAxiom x (Plus (Times (a, v), Times (b, Variable v2)))=
        if v2 = x
          then Plus (Times (b, Variable x), Times (a, v))
          else Plus (Times (a, v),Times (b, Variable v2))
    | PermAxiom x (Plus (Times (a, v),
                         Plus (Times (b, Variable v2), e))) =
        if v2 = x
          then Plus (Times (b, Variable x), Plus (Times (a, v), e))
          else Plus (Times (a, v), Plus (Times (b, Variable v2), e))
    | PermAxiom x y = y;


(* applies `PermAxiom' top-down on `J'.
   type: string * Expr -> Expr  *)

  fun applyPerm x J =
        if J = PermAxiom x J
          then case J of
                   Number _     => J
                 | pInfinity    => J
                 | mInfinity    => J
                 | Variable _   => J
                 | Times (a, b) => Times (a, applyPerm x b)
                 | Plus (a, b)  => Plus (applyPerm x a, applyPerm x b)
                 | Max (a, b)   => Max (applyPerm x a, applyPerm x b)
                 | Min (a, b)   => Min (applyPerm x a, applyPerm x b)
         else applyPerm x (PermAxiom x J);


(* brings Variable `x' to the left in `J'.
   type: string * Expr -> Expr  *)

  fun applyPerms x J =
        if J = applyPerm x J
          then J
          else applyPerms x (applyPerm x J);


(* solves a list of inequalities `S' for Variable `x'.
   type: (Expr * Expr) list * string -> (Expr * Expr) list  *)

  infixr solve_for;
  fun S solve_for x = 
    let
      (* brings Variable `x' to the left (if occuring). *)
      val permS = map (fn (U1, U2) => (applyPerms x U1, applyPerms x U2))
                      (filter (fn (U1, U2) =>
                                 (x occurs_in U1) orelse (x occurs_in U2)) S)

      (* solves an inequality U1 <= U2 for x.
         type: (Expr * Expr) -> Expr list  *)
      fun solve (U1, U2) =
            case (U1, U2) of
                (*  n <= cv + a  *)
                (Number n, Plus (Times (c, Variable v), a)) =>
                  if c < 0.0 andalso v = x
                    then (*  x <= (n - a)/c  *)
                      [(Plus (Times (1.0, Variable x), Number 0.0),
		       SIMP (Times (1.0 / c, Plus (Number n, Times(~1.0,a)))))]
                    else
                      if c > 0.0 andalso v = x
                        then (*  (n - a)/c <= x  *)
                          [(SIMP (Times (1.0 / c,
                                        Plus (Number n, Times (~1.0, a)))), 
		           Plus (Times (1.0, Variable x), Number 0.0))]
                        else (*  x is not in RHS  *)
                          raise Error "Cannot solve inequation!!!"
                (*  cv + a <= n  *)
              | (Plus (Times (c, Variable v), a), Number n) =>
                  if c < 0.0 andalso v = x
                    then (*  (n - a)/c <= x  *)
                     [(SIMP (Times (1.0 / c, Plus (Number n,Times (~1.0, a)))),
		       Plus (Times (1.0, Variable x), Number 0.0))]
                    else
                      if c > 0.0 andalso v = x
                        then (*  x <= (n - a)/c  *)
                          [(Plus (Times (1.0, Variable x), Number 0.0),
		            SIMP (Times (1.0 / c,
                                        Plus (Number n, Times (~1.0 ,a)))))]
                        else (*  x is not in RHS  *)
                          raise Error "Cannot solve inequation!!!"
               (*  c1*v1 + a1 <= c2*v2 + a2  *)
             | (Plus (Times (c1, Variable v1), a1),
                Plus (Times (c2, Variable v2), a2)) =>
                  if x = v1 andalso x = v2
                    then
                      if c1 > c2
                        then (*  x <= (a2 - a1)/(c1 - c2)  *)
                          [(Plus (Times (1.0, Variable x), Number 0.0),
		           SIMP (Times (1.0 / (c1 - c2),
                                        Plus (a2, Times (~1.0, a1)))))]
                        else
                          if c1 < c2
                            then (*  (a2 - a1)/(c1 - c2) <= x  *)
                              [(SIMP (Times (1.0 / (c1 - c2),
                                            Plus (a2, Times (~1.0, a1)))),
		               Plus (Times (1.0, Variable x), Number 0.0))]
                            else 
                              if c1 < 0.0
                                then
                                  [(Plus (Times (1.0, Variable x), Number 0.0),
                                    SIMP (Plus (Times (1.0, Variable x),
                                            Times (1.0 / c1,
                                              Plus (a1, Times (~1.0, a2)))))),
                                   (SIMP (Plus (Times (1.0, Variable x),
                                            Times (1.0 / c1,
                                              Plus (a2, Times (~1.0, a1))))),
                                    Plus (Times (1.0, Variable x),Number 0.0))]
                                else
                                  if c1 > 0.0
                                    then
                                      [(Plus (Times (1.0, Variable x),
                                              Number 0.0),
                                        SIMP (Plus (Times (1.0, Variable x),
                                            Times (1.0 / c1,
                                              Plus (a2, Times (~1.0, a1)))))),
                                       (SIMP (Plus (Times (1.0, Variable x),
                                            Times (1.0 / c1,
                                              Plus (a1, Times (~1.0, a2))))),
                                        Plus (Times (1.0, Variable x),
                                              Number 0.0))]
                                    else
                                      raise Error "Cannot solve inequation!!!"
                    else
                      if x = v1
                        then
                          if c1 < 0.0
                            then (*  (U2 - a1)/c1 <= x  *)
                              [(SIMP (Times (1.0 / c1,
                                            Plus (U2, Times (~1.0, a1)))),
		               Plus (Times (1.0, Variable x), Number 0.0))]
                            else (*  x <= (U2 - a1)/c1  *)
                              [(Plus (Times (1.0, Variable x), Number 0.0),
		               SIMP (Times (1.0 / c1,
                                            Plus (U2, Times (~1.0, a1)))))]
                        else
                          if x = v2
                            then
                              if c2 < 0.0
                                then (*  x <= (U1 - a2)/c2  *)
                                  [(Plus (Times (1.0, Variable x), Number 0.0),
			           SIMP (Times (1.0 / c2,
                                                Plus (U1, Times (~1.0, a2)))))]
                                else (*  (U1 - t2)/r2 <= x  *)
                                  [(SIMP (Times (1.0 / c2,
                                                Plus (U1, Times (~1.0, a2)))),
			           Plus (Times (1.0, Variable x), Number 0.0))]
                            else (*  x is neither in U1 nor in U2  *)
                              raise Error "Cannot solve inequation!!!"
             | _ =>
                 raise Error "Cannot solve inequation!!!"
    in
      filter (fn (U1,U2) => (x occurs_in U1) orelse (x occurs_in U2))
             (flatten (map solve permS))
    end;


(* greatest value for Variable `x' constraint by `S'
   type: (Expr * Expr) list -> string -> Expr  *)

  fun UPPER S x = 
    let
      (* list of inequalities containing `x' and solved for `x'
         type: (Expr * Expr) list  *)
      val soln = S solve_for x;

      (* `x <= _' ?
         type: Expr * Expr -> Bool  *)
      fun upper (U1, U2) = (U1 = Plus (Times (1.0, Variable x), Number 0.0));

      (* list of RHS of `x <= _'
        type: Expr list  *)
      val soln_set = map (fn  (U1, U2) => U2) (filter upper soln);

      (* takes the least upper restriction on `x'
         type: Expr list -> Expr  *)
      fun MIN S =
            case S of
                []      => raise Error "impossible error!!!"
              | [s]     => s
              | s :: S' => Min (s, MIN S');
    in
      if soln_set = []
        then pInfinity
        else MIN soln_set
    end;


(* lowest value for Variable `x' constraint by `S'
   type: (Expr * Expr) list -> string -> Expr  *)

  fun LOWER S x = 
    let
      (* list of inequalities containing `x' and solved for `x'
         type: (Expr * Expr) list *)
      val soln = S solve_for x;

      (* `_ <= x' ?
         type: Expr * Expr -> Bool  *)
      fun lower (U1, U2) = (U2 = Plus (Times (1.0, Variable x), Number 0.0))

      (* list of LHS of `_ <= x'
         type: Expr list  *)
      val soln_set = map (fn  (U1, U2) => U1) (filter lower soln);

      (* greatest lower restriction on `x'
         type: Expr list -> Expr  *)
      fun MAX S =
            case S of
                []      => raise Error "impossible error!!!"
              | [s]     => s
              | s :: S' => Max (s, MAX S');
    in
      if soln_set = []
        then mInfinity
        else MAX soln_set
    end;


(* auxiliary function called by SUP.
   type: string -> Expr -> Expr  *)

  fun SUPP x y =
    case y of
        Number n  => y
      | pInfinity => y
      | Plus (pInfinity, c) => pInfinity
      | mInfinity => y
      | Plus (mInfinity, c) => mInfinity
      | _         =>
          if Variable x = y
            then pInfinity
            else case y of
                     Min (A, B) => Min (SUPP x A, SUPP x B)
                   | _          =>
                       if x occurs_in y
                         then
                           case (applyPerms x y) of
                               Plus (Times (b, x'), c) =>
                                 if (Variable x = x') andalso
                                    not (x occurs_in c)
                                   then if b > 1.0
                                          then pInfinity
                                          else if b < 1.0
                                                 then Times (1.0 /(1.0 - b), c)
                                                 else case c of
                                                          Number n =>
                                                            if n < 0.0
                                                              then mInfinity
                                                              else pInfinity
                                                        | pInfinity =>
                                                            c
                                                        | mInfinity =>
                                                            c
                                                        | _ =>
                                                            pInfinity
                                   else raise Error
                                          "error in SUPP (should not occur)"
                             | _ => raise Error
                                          "error in SUPP (should not occur)"
                         else y;


(* auxiliary function called by INF.
   type: string -> Expr -> Expr  *)

  fun INFF x y =
    case y of
        Number n  => y
      | pInfinity => y
      | Plus (pInfinity, c) => pInfinity
      | mInfinity => y
      | Plus (mInfinity, c) => mInfinity
      | _         =>
          if Variable x = y
            then mInfinity
            else case y of
                Max (A, B) => Max (INFF x A, INFF x B)
              | _          =>
                  if x occurs_in y
                    then
                      case (applyPerms x y) of
                          Plus (Times (b, x'), c) =>
                            if (Variable x = x') andalso 
                               not (x occurs_in c)
                              then if b > 1.0
                                     then mInfinity
                                     else if b < 1.0
                                            then Times (1.0/(1.0 - b), c)
                                            else case c of
                                                     Number n =>
                                                       if n > 0.0
                                                         then pInfinity
                                                         else mInfinity
                                                   | pInfinity =>
                                                       c
                                                   | mInfinity =>
                                                       c
                                                   | _ =>
                                                       mInfinity
                              else raise Error
                                          "error in INFF (should not occur)"
                        | _ => raise Error "error in INFF (should not occur)"
                    else y;


(* computes the supremum of `J' for a list `S' of inequalities
   `H' is bound to a list of variables occuring in `S' for recursive
   function calls; initially empty.
   type: (Expr * Expr) list -> Expr -> string list -> Expr  *)

  fun SUP S J H =
        case J of
            Number _ =>
              J
          | pInfinity =>
              J
          | Plus (pInfinity, c) =>
              pInfinity
          | mInfinity =>
              J
          | Plus (mInfinity, c) =>
              mInfinity
          | Variable j =>
              if j mem H
                then J
                else let
                       val Q = UPPER S j;
                       val Z = SUP S Q (j :: H);
                     in
                       SUPP j (SIMP Z)
                     end
          | Times (r, A) =>
              if r < 0.0
                then Times (r, INF S A H)
                else Times (r, SUP S A H)
          | Plus (Times (r, Variable v), B) =>
              let
                val B' = SIMP (SUP S B (v :: H));
              in
                if (B' = pInfinity) orelse (B' = mInfinity)
                  then
                    B'
                  else
                    if v occurs_in B'
                      then let
                             val J' = SIMP (Plus (Times (r, Variable v), B'))
                           in
                             SUP S J' H
                           end
                      else Plus (SUP S (Times (r, Variable v)) H, B')
              end
          | Min (A, B) =>
              Min (SUP S A H, SUP S B H)
          | _ =>
              raise Error "error in SUP (should not occur)"

(* computes the infimum of `J' for a list `S' of inequalities
   `H' is bound to a list of variables occuring in `S' for recursive
   function calls; initially empty.
   type:: (Expr * Expr) list -> Expr -> string list -> Expr  *)

  and INF S J H =
        case J of
            Number _ =>
              J
          | pInfinity =>
              J
          | Plus (pInfinity, c) =>
              pInfinity
          | mInfinity =>
              J
          | Plus (mInfinity, c) =>
              mInfinity
          | Variable j =>
              if j mem H
                then J
                else let
                       val Q = LOWER S j;
                       val Z = INF S Q (j :: H);
                     in
                       INFF j (SIMP Z)
                     end
          | Times (r, A) =>
              if r < 0.0
                then Times (r, SUP S A H)
                else Times (r, INF S A H)
          | Plus (Times (r, Variable v), B) =>
              let
                val B' = SIMP (INF S B (v :: H))
              in
                if (B' = pInfinity) orelse (B' = mInfinity)
                  then
                    B'
                  else
                    if v occurs_in B'
                      then let
                             val J' = SIMP (Plus (Times (r, Variable v), B'))
                           in
                             INF S J' H
                           end
                      else Plus (INF S (Times (r, Variable v)) H, B')
              end
          | Max (A, B) =>
              Max (INF S A H, INF S B H)
          | _ =>
              raise Error "error in INF (should not occur)";


(* computes the supremum of `j' for a list `S' of inequalities depending
   on variables in `H'.
   type: (Expr * Expr) list -> string -> string list -> Expr  *)

  fun Sup S j H = 
    SIMP (SUP (map (fn (U1, U2) => (SIMP U1, SIMP U2)) S) (Variable j) H);


(* computes the infimum of `J' for a list `S' of inequalities depending
   on variables in `H'.
   type: (Expr * Expr) list -> string -> string list -> Expr  *)

  fun Inf S j H = 
    SIMP (INF (map (fn (U1, U2) => (SIMP U1, SIMP U2)) S) (Variable j) H);

end;  (* struct *)
