(* 
   file     : $CAT_DIR/SupInf/PASimplifier.ML
   platform : PolyML 3.2 and Isabelle94-8/HOL
   version  : 1.0b  (16.07.98)
   notes    : A Simplifier to prepare a term for its treatment in
              Presburger arithmetic is defined here.
*)

signature PASIMP =
sig

  val pa_simplify:     ArithSemantics.BoolTerm -> ArithSemantics.BoolTerm
  val map_to_pa:       ArithSemantics.BoolTerm -> ArithSemantics.BoolTerm
  val eliminate_preds: ArithSemantics.BoolTerm -> ArithSemantics.BoolTerm
  val pull_args_out:   ArithSemantics.BoolTerm -> ArithSemantics.BoolTerm
  val normal_ineqs:    ArithSemantics.BoolTerm -> ArithSemantics.BoolTerm
  val map_to_DNF:      ArithSemantics.BoolTerm -> ArithSemantics.BoolTerm
  val ineq_simp:       ArithSemantics.BoolTerm -> ArithSemantics.BoolTerm

end;

structure PASimplifier: PASIMP =
struct

local

  open Basics;
  open ArithSemantics;

  infix pow;


(* provides unique numbers to generate new constant names *)

  val var_count = ref 3;
  fun read_vc () = (var_count := (!var_count) + 1; !var_count);


(* auxiliary functions *)

(* type: BoolTerm list -> BoolTerm *)

  fun mk_conj []      = BoolTrue
    | mk_conj [x]     = x
    | mk_conj (x::xs) = BoolAnd (x, mk_conj xs);


(* type: BoolTerm list -> BoolTerm *)

  fun mk_imp []      = BoolFalse
    | mk_imp [x]     = x
    | mk_imp (x::xs) = BoolImp (x, mk_imp xs);


(* pulles out arguments of IntegConstants that contain `+' and `*'.
   type: BoolTerm -> BoolTerm * BoolTerm  *)

  fun pull_args term =
    let
      fun check l = 
        let
          fun check' moo noo []      = (rev moo, noo)
            | check' moo noo (x::xs) =
                case x of
                    IntegConstant (nmi, args) =>
                      let
                        val (aoo, yoo) = check args;
                      in
                        check' ((IntegConstant (nmi, aoo))::moo) (noo @ yoo) xs
                      end
                  | IntegValue _ =>
                      check' (x::moo) noo xs
                  | _ =>
                      let
                        val m = IntegConstant (("X", read_vc ()), []);
                      in
                        check' (m::moo) ((BoolEq (m, x))::noo) xs
                      end;
        in
          check' [] [] l
        end;

      fun bull (x, xs) =
            case x of
                IntegConstant (nmi, []) =>
                  (IntegConstant (nmi, []), [])
              | IntegConstant (nmi, args) =>
                  let
                    val (new_args, ooo) = check args;
                  in
                    (IntegConstant (nmi, new_args), ooo)
                  end
              | IntegValue arg =>
                  (IntegValue arg, [])
              | IntegComp arg =>
                  let
                    val (voo, loo) = bull (arg, []);
                  in
                    (IntegComp voo, loo)
                  end
              | IntegPlus (arg1, arg2) =>
                  let
                    val (voo1, loo1) = bull (arg1, []);
                    val (voo2, loo2) = bull (arg2, []);
                  in
                    (IntegPlus (voo1, voo2), loo1 @ loo2)
                  end
              | IntegMinus (arg1, arg2) =>
                  let
                    val (voo1, loo1) = bull (arg1, []);
                    val (voo2, loo2) = bull (arg2, []);
                  in
                    (IntegMinus (voo1, voo2), loo1 @ loo2)
                  end
              | IntegTimes (arg1, arg2) =>
                  let
                    val (voo1, loo1) = bull (arg1, []);
                    val (voo2, loo2) = bull (arg2, []);
                  in
                    (IntegTimes (voo1, voo2), loo1 @ loo2)
                  end
              | IntegDivide (arg1, arg2) =>
                  let
                    val (voo1, loo1) = bull (arg1, []);
                    val (voo2, loo2) = bull (arg2, []);
                  in
                    (IntegDivide (voo1, voo2), loo1 @ loo2)
                  end
              | IntegModulus (arg1, arg2) =>
                  let
                    val (voo1, loo1) = bull (arg1, []);
                    val (voo2, loo2) = bull (arg2, []);
                  in
                    (IntegModulus (voo1, voo2), loo1 @ loo2)
                  end
              | IntegPower (arg1, arg2) =>
                  let
                    val (voo1, loo1) = bull (arg1, []);
                    val (voo2, loo2) = bull (arg2, []);
                  in
                    (IntegPower (voo1, voo2), loo1 @ loo2)
                  end;


      fun pull (x, xs) = 
            case x of
                BoolTrue =>
                  (BoolTrue, [])
              | BoolFalse =>
                  (BoolFalse, [])
              | BoolPred (nm, args) =>
                  (BoolPred (nm, args), [])
              | BoolLess (arg1, arg2) =>
                  let
                    val (voo1, loo1) = bull (arg1, []);
                    val (voo2, loo2) = bull (arg2, []);
                  in
                    (BoolLess (voo1, voo2), loo1 @ loo2)
                  end
              | BoolLEq (arg1, arg2) =>
                  let
                    val (voo1, loo1) = bull (arg1, []);
                    val (voo2, loo2) = bull (arg2, []);
                  in
                    (BoolLEq (voo1, voo2), loo1 @ loo2)
                  end
              | BoolEq (arg1, arg2) =>
                  let
                    val (voo1, loo1) = bull (arg1, []);
                    val (voo2, loo2) = bull (arg2, []);
                  in
                    (BoolEq (voo1, voo2), loo1 @ loo2)
                  end
              | BoolGEq (arg1, arg2) =>
                  let
                    val (voo1, loo1) = bull (arg1, []);
                    val (voo2, loo2) = bull (arg2, []);
                  in
                    (BoolGEq (voo1, voo2), loo1 @ loo2)
                  end
              | BoolGreater (arg1, arg2) =>
                  let
                    val (voo1, loo1) = bull (arg1, []);
                    val (voo2, loo2) = bull (arg2, []);
                  in
                    (BoolGreater (voo1, voo2), loo1 @ loo2)
                  end
              | BoolNot arg =>
                  let
                    val (voo, loo) = pull (arg, []);
                  in
                    (BoolNot voo, loo)
                  end
              | BoolAnd (arg1, arg2) =>
                  let
                    val (voo1, loo1) = pull (arg1, []);
                    val (voo2, loo2) = pull (arg2, []);
                  in
                    (BoolAnd (voo1, voo2), loo1 @ loo2)
                  end
              | BoolOr (arg1, arg2) =>
                  let
                    val (voo1, loo1) = pull (arg1, []);
                    val (voo2, loo2) = pull (arg2, []);
                  in
                    (BoolOr (voo1, voo2), loo1 @ loo2)
                  end
              | BoolImp (arg1, arg2) =>
                  let
                    val (voo1, loo1) = pull (arg1, []);
                    val (voo2, loo2) = pull (arg2, []);
                  in
                    (BoolImp (voo1, voo2), loo1 @ loo2)
                  end;

      (* replaces `+' and `*' expressions in `term' by new variables and
         gives list pulled out expressions in `l'.
         type: Boolterm * BoolTerm list  *)
      val (boo, l) = pull (term, []);

      val doo = mk_conj l;
    in
      (* `boo' is argument term, where expressions involving `+' and `*'
         are replaced by new variables; `doo' is conjunction of pulled out
         constraints, e.g., f(1+x, g(2*y)) becomes
         (f(X1, g(X2)), (X1 = 1+x) & (X2 = 2*y)) *)
      (boo, doo)
    end;


(* pulls all expressions involving `+' or `*' out of a term and its subterms.
   type: BoolTerm -> BoolTerm list  *)

  fun pull_out l =
    let
      val (boo, foo) = pull_args l
    in
      boo :: (if foo = BoolTrue then [] else pull_out foo)
    end;


(* the rewrite system follows *)


(* auxiliary functions for an ordering relation *)

(* type: IntegTerm * unit -> bool *)

  infix is_number;
  fun (IntegValue _) is_number () = true
    | _              is_number () = false;
 
(* type: IntegTerm * unit -> bool *)

  infix is_const;
  fun (IntegConstant _)             is_const () = true
    | (IntegComp (IntegConstant _)) is_const () = true
    | _                             is_const () = false;
 
(* type: IntegTerm -> IntegTerm *)

  fun get_const (IntegConstant x)             = IntegConstant x
    | get_const (IntegComp (IntegConstant x)) = IntegConstant x
    | get_const _ = raise Error "Attempt to get non-constant!!!";


(* This is an ordering function over IntegTerm constants and values.
   type: IntegTerm * IntegTerm -> bool  *)

  infix is_less_then;
  fun (IntegConstant ((nm1, i1), _)) is_less_then
      (IntegConstant ((nm2, i2), _)) =
        (nm1 < nm2) orelse ((nm1 = nm2) andalso (i1 < i2))
    | (IntegValue n) is_less_then (IntegValue m) = n < m
    | _ is_less_then _ =
        raise Error "Attempt to compare uncomparable terms!!!";
 

(* imposes a relation over IntegTerm which sets IntegValues right of
   IntegConstants and IntegConstants right of other IntegTerms;
   IntegConstants themself are ordered by `is_less_then', defined above.
   type: IntegTerm * IntegTerm -> bool  *)

  infix is_right_of;
  fun (x is_right_of y) =
        if y is_const ()
          then (if x is_number ()
                 then true
                 else (if x is_const ()
                        then (get_const x) is_less_then (get_const y)
                        else false))
          else (if y is_number ()
                  then (if x is_number ()
                          then x is_less_then y
                          else false)
                  else (if (x is_number ()) orelse (x is_const ())
                          then true
                          else false));
 

(* syntactical equality relation over IntegTerm;
   could be changed for a better one.
   type: IntegTerm * IntegTerm -> bool  *)

  infix is_equal_to;
  fun x is_equal_to y = x = y;


(* auxiliary axiom; identity.
   type: IntegTerm -> IntegTerm  *)

  fun DummyAxiom x = x;


(*** Calculation axioms ***)
 
  (* Axiom0: n Integ<op> m = n' <op> m' *)
 
  fun Axiom0 (IntegPlus (IntegValue n, IntegValue m)) =
        (IntegValue (n + m) 
          handle e => IntegPlus (IntegValue n, IntegValue m))
    | Axiom0 (IntegMinus (IntegValue n, IntegValue m)) =
        (IntegValue (n - m) 
          handle e => IntegMinus (IntegValue n, IntegValue m))
    | Axiom0 (IntegTimes (IntegValue n, IntegValue m)) =
        (IntegValue (n * m)
          handle e => IntegTimes (IntegValue n, IntegValue m))
    | Axiom0 (IntegDivide (IntegValue n, IntegValue m)) =
        (IntegValue (n div m)
          handle e => IntegDivide (IntegValue n, IntegValue m))
    | Axiom0 (IntegModulus (IntegValue n, IntegValue m)) =
        (IntegValue (n mod m)
          handle e => IntegModulus (IntegValue n, IntegValue m))
    | Axiom0 (IntegPower (IntegValue n, IntegValue m)) =
        (IntegValue (n pow m)
          handle e => IntegPower (IntegValue n, IntegValue m))
    | Axiom0 x = x;


(*** Addition axioms ***)
 
  (* Axiom1: 0 + x = x *)
 
  fun Axiom1 (IntegPlus (IntegValue 0, x)) =
        x
    | Axiom1 x = x;
 
  (* Axiom2: x + 0 = x *)
 
  fun Axiom2 (IntegPlus (x, IntegValue 0)) =
        x
    | Axiom2 x = x;
 
  (* Axiom3: x .=. y ==> x + y = 2x *)
 
  fun Axiom3 (IntegPlus (x, y)) =
        if x is_equal_to y
          then IntegTimes (IntegValue 2, x)
          else IntegPlus (x, y)
    | Axiom3 x = x;
 
  (* Axiom4: x .=. y ==> x + (- y) = 0 *)
 
  fun Axiom4 (IntegPlus (x, IntegComp y)) =
        if x is_equal_to y
          then IntegValue 0
          else IntegPlus (x, IntegComp y)
    | Axiom4 x = x;
 
  (* Axiom5: x .=. y ==> (- x) + y = 0 *)
 
  fun Axiom5 (IntegPlus (IntegComp x, y)) =
        if x is_equal_to y
          then IntegValue 0
          else IntegPlus (IntegComp x, y)
    | Axiom5 x = x;
 
  (* Axiom6: x .=. y ==> (- x) + (- y) = ~2x *)
 
  fun Axiom6 (IntegPlus (IntegComp x, IntegComp y)) =
        if x is_equal_to y
          then IntegTimes (IntegValue ~2, x)
          else IntegPlus (IntegComp x, IntegComp y)
    | Axiom6 x = x;
 
  (* Axiom7: x .=. y ==> x + (y + z) = 2x + z *)
 
  fun Axiom7 (IntegPlus (x, IntegPlus (y, z))) =
        if x is_equal_to y
          then IntegPlus (IntegTimes (IntegValue 2, x), z)
          else IntegPlus (x, IntegPlus (y, z))
    | Axiom7 x = x;
 
  (* Axiom8: x .=. y ==> x + (- y + z) = z *)
 
  fun Axiom8 (IntegPlus (x, IntegPlus (IntegComp y, z))) =
        if x is_equal_to y
          then z
          else IntegPlus (x, IntegPlus (IntegComp y, z))
    | Axiom8 x = x;
 
  (* Axiom9: x .=. y ==> (- x) + (y + z) = z *)
 
  fun Axiom9 (IntegPlus (IntegComp x, IntegPlus (y, z))) =
        if x is_equal_to y
          then z
          else IntegPlus (IntegComp x, IntegPlus (y, z))
    | Axiom9 x = x;
 
  (* Axiom10: x .=. y ==> (- x) + (- y + z) = ~2x + z *)
 
  fun Axiom10 (IntegPlus (IntegComp x, IntegPlus (IntegComp y, z))) =
        if x is_equal_to y
          then IntegPlus (IntegTimes (IntegValue ~2, x), z)
          else IntegPlus (IntegComp x, IntegPlus (y, z))
    | Axiom10 x = x;
 
  (* Axiom11: x .=. y  ==>  n * x + m * y = (n+m) * x  *)
 
  fun Axiom11 (IntegPlus (IntegTimes (IntegValue n, x),
                          IntegTimes (IntegValue m, y))) =
        if (x is_equal_to y)
          then IntegTimes (IntegValue (n + m), x)
          else IntegPlus (IntegTimes (IntegValue n, x),
                          IntegTimes (IntegValue m, y))
    | Axiom11 x = x;
 
  (* Axiom12: x .=. y  ==>  n * x + (m * y + z) = (n+m) * x + z  *)

  fun Axiom12 (IntegPlus (IntegTimes (IntegValue n, x),
                 IntegPlus (IntegTimes (IntegValue m, y), z))) =
        if (x is_equal_to y)
          then IntegPlus (IntegTimes (IntegValue (n + m), x), z)
          else IntegPlus (IntegTimes (IntegValue n, x),
                 IntegPlus (IntegTimes (IntegValue m, y), z))
    | Axiom12 x = x;
 
  (* Axiom13: x .=. y  ==>  x + n * y = (1+n) * x  *)
 
  fun Axiom13 (IntegPlus (x, IntegTimes (IntegValue n, y))) =
        if (x is_equal_to y)
          then IntegTimes (IntegValue (1 + n), x)
          else IntegPlus (x, IntegTimes (IntegValue n, y))
    | Axiom13 x = x;
 
  (* Axiom14: x .=. y  ==>  x + (n * y + z) = (1+n) * x + z  *)
 
  fun Axiom14 (IntegPlus (x, IntegPlus (IntegTimes (IntegValue n, y), z))) =
        if (x is_equal_to y)
          then IntegPlus (IntegTimes (IntegValue (1 + n), x), z)
          else IntegPlus (x, IntegPlus (IntegTimes (IntegValue n, y), z))
    | Axiom14 x = x;
 
  (* Axiom15: x .=. y  ==>  n * x + y = (n+1) * x  *)
 
  fun Axiom15 (IntegPlus (IntegTimes (IntegValue n, x), y)) =
        if (x is_equal_to y)
          then IntegTimes (IntegValue (n + 1), x)
          else IntegPlus (IntegTimes (IntegValue n, x), y)
    | Axiom15 x = x;
 
  (* Axiom16: x .=. y  ==>  n * x + (y + z) = (n+1) * x + z  *)
 
  fun Axiom16 (IntegPlus (IntegTimes (IntegValue n, x), IntegPlus (y, z))) =
        if (x is_equal_to y)
          then IntegPlus (IntegTimes (IntegValue (n + 1), x), z)
          else IntegPlus (IntegTimes (IntegValue n, x), IntegPlus (y, z))
    | Axiom16 x = x;
 
  (* Axiom17: x - y = x + (- y) *)
 
  fun Axiom17 (IntegMinus (x, y)) =
        IntegPlus (x, IntegComp y)
    | Axiom17 x = x;



(*** Complement axioms ***)
 
  (* Axiom20: - (- x) = x *)
 
  fun Axiom20 (IntegComp (IntegComp x)) =
        x
    | Axiom20 x = x;
 
  (* Axiom21: - n = ~n *)
 
  fun Axiom21 (IntegComp (IntegValue n)) =
        IntegValue (~n)
    | Axiom21 x = x;
  


(*** Multiplication axioms ***)
 
  (* Axiom30: 0 * x = 0 *)
 
  fun Axiom30 (IntegTimes (IntegValue 0, x)) =
        IntegValue 0
    | Axiom30 x = x;
 
  (* Axiom31: x * 0 = 0  *)
 
  fun Axiom31 (IntegTimes (x, IntegValue 0)) =
        IntegValue 0
    | Axiom31 x = x;
 
  (* Axiom32: 1 * x = x  *)
 
  fun Axiom32 (IntegTimes (IntegValue 1, x)) =
        x
    | Axiom32 x = x;
 
  (* Axiom33: x * 1 = x  *)
 
  fun Axiom33 (IntegTimes (x, IntegValue 1)) =
        x
    | Axiom33 x = x;
 
  (* Axiom34: ~1 * x = - x  *)
 
  fun Axiom34 (IntegTimes (IntegValue ~1, x)) =
        IntegComp x
    | Axiom34 x = x;
 
  (* Axiom35: x * ~1 = - x  *)
 
  fun Axiom35 (IntegTimes (x, IntegValue ~1)) =
        IntegComp x
    | Axiom35 x = x;
 


(*** Division axioms ***)
 
  (* Axiom40: x .=. y  &  y .!=. 0  ==>  x / y = 1 *)
 
  fun Axiom40 (IntegDivide (x, y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntegValue 0)))
          then IntegValue 1
          else IntegDivide (x, y)
    | Axiom40 x = x;
 
  (* Axiom41: x .=. y  &  y .!=. 0  ==>  (- x) / y = ~1 *)
 
  fun Axiom41 (IntegDivide (IntegComp x, y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntegValue 0)))
          then IntegValue ~1
          else IntegDivide (IntegComp x, y)
    | Axiom41 x = x;
 
  (* Axiom42: x .=. y  &  y .!=. 0  ==>  (x * z) / y = z *)
 
  fun Axiom42 (IntegDivide (IntegTimes (x, z), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntegValue 0)))
          then z
          else IntegDivide (IntegTimes (x, z), y)
    | Axiom42 x = x;
 
  (* Axiom43: x .=. y  &  y .!=. 0  ==>  (z * x) / y = z *)
 
  fun Axiom43 (IntegDivide (IntegTimes (z, x), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntegValue 0)))
          then z
          else IntegDivide (IntegTimes (z, x), y)
    | Axiom43 x = x;

  (* Axiom44: x .=. y  &  y .!=. 0  ==>  ((z * x) * z') / y = z * z'  *)
 
  fun Axiom44 (IntegDivide (IntegTimes (IntegTimes (z, x), z'), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntegValue 0)))
          then IntegTimes (z, z')
          else IntegDivide (IntegTimes (IntegTimes (z, x), z'), y)
    | Axiom44 x = x;

  (* Axiom45: n mod m = 0  ==> (n * x) / m = (n/m) * x  *)

  fun Axiom45 (IntegDivide (IntegTimes (IntegValue n, x), IntegValue m)) =
        ((if (n mod m) = 0
           then IntegTimes (IntegValue (n div m), x)
           else IntegDivide (IntegTimes (IntegValue n, x), IntegValue m))
         handle e => IntegDivide (IntegTimes (IntegValue n, x), IntegValue m))
    | Axiom45 x = x;
 
  (* Axiom46:  x / 1 = x  *)
 
  fun Axiom46 (IntegDivide (x, IntegValue 1)) =
        x
    | Axiom46 x = x;
 
  (* Axiom47:  x / ~1 = - x  *)
 
  fun Axiom47 (IntegDivide (x, IntegValue ~1)) =
        IntegComp x
    | Axiom47 x = x;
 
  (* Axiom48:  x .!=. 0  ==>  0 / x = 0 *)
 
  fun Axiom48 (IntegDivide (IntegValue 0, x)) =
        if (not (x is_equal_to (IntegValue 0)))
          then IntegValue 0
          else IntegDivide (IntegValue 0, x)
    | Axiom48 x = x;



(*** Modulus axioms ***)
 
  (* Axiom50: x .=. y  &  y .!=. 0  ==>  x mod y = 0 *)
 
  fun Axiom50 (IntegModulus (x, y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntegValue 0)))
          then IntegValue 0
          else IntegModulus (x, y)
    | Axiom50 x = x;
 
  (* Axiom51: x .=. y  &  y .!=. 0  ==>  (x * z) mod y = 0 *)
 
  fun Axiom51 (IntegModulus (IntegTimes (x, z), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntegValue 0)))
          then IntegValue 0
          else IntegModulus (IntegTimes (x, z), y)
    | Axiom51 x = x;
 
  (* Axiom52: x .=. y  &  y .!=. 0  ==>  (z * x) mod y = 0 *)
 
  fun Axiom52 (IntegModulus (IntegTimes (z, x), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntegValue 0)))
          then IntegValue 0
          else IntegModulus (IntegTimes (z, x), y)
    | Axiom52 x = x;

  (* Axiom53: x .=. y  &  y .!=. 0  ==>  ((z * x) * z') mod y = 0 *)
 
  fun Axiom53 (IntegModulus (IntegTimes (IntegTimes (z, x), z'), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntegValue 0)))
          then IntegValue 0
          else IntegModulus (IntegTimes (IntegTimes (z, x), z'), y)
    | Axiom53 x = x;

  (* Axiom54: n mod m = 0  ==> (n * x) mod m = 0  *)

  fun Axiom54 (IntegModulus (IntegTimes (IntegValue n, x), IntegValue m)) =
        ((if (n mod m) = 0
           then IntegValue 0
           else IntegModulus (IntegTimes (IntegValue n, x), IntegValue m))
         handle e => IntegModulus (IntegTimes (IntegValue n, x), IntegValue m))
    | Axiom54 x = x;
  
  (* Axiom55: abs(n).<=.abs(m) & n.!=.0 ==> (x mod n) mod m = x mod n  *)

  fun Axiom55 (IntegModulus (IntegModulus (x, IntegValue n), IntegValue m)) =
        if ((abs n) <= (abs m)) andalso (not (n = 0))
          then IntegModulus (x, IntegValue n)
          else IntegModulus (IntegModulus (x, IntegValue n), IntegValue m)
    | Axiom55 x = x;
  
  (* Axiom56: x.=.y & x.!=.0 ==> (z mod x) mod y = z mod y *)
 
  fun Axiom56 (IntegModulus (IntegModulus (z, x), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntegValue 0)))
          then IntegModulus (z, y)
          else IntegModulus (IntegModulus (z, x), y)
    | Axiom56 x = x;

  (* Axiom57: x.=.y & x.!=.0 ==> (z mod x + z') mod y = (z + z') mod y *)
 
  fun Axiom57 (IntegModulus (IntegPlus (IntegModulus (z, x), z'), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntegValue 0)))
          then IntegModulus (IntegPlus (z, z'), y)
          else IntegModulus (IntegPlus (IntegModulus (z, x), z'), y)
    | Axiom57 x = x;

  (* Axiom58: x.=.y & x.!=.0 ==> (z + z' mod x) mod y = (z + z') mod y *)
 
  fun Axiom58 (IntegModulus (IntegPlus (z, IntegModulus (z', x)), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntegValue 0)))
          then IntegModulus (IntegPlus (z, z'), y)
          else IntegModulus (IntegPlus (z, IntegModulus (z', x)), y)
    | Axiom58 x = x;

  (* Axiom59: x.=.y & x.!=.0 ==> (z''+(z mod x +z'))mod y=(z''+(z+z'))mod y *)
 
  fun Axiom59 (IntegModulus 
                (IntegPlus (z'', IntegPlus (IntegModulus (z, x), z')), y)) =
        if (x is_equal_to y) andalso (not (y is_equal_to (IntegValue 0)))
          then IntegModulus (IntegPlus (z'', IntegPlus (z, z')), y)
          else IntegModulus 
                (IntegPlus (z'', IntegPlus (IntegModulus (z, x), z')), y)
    | Axiom59 x = x;

  (* Axiom60: abs(n).<=.m & n.!=.0 ==> (x mod n) div m = 0  *)
 
  fun Axiom60 (IntegDivide (IntegModulus (x, IntegValue n), IntegValue m)) =
        if ((abs n) <= m) andalso (not (n = 0))
          then IntegValue 0
          else IntegDivide (IntegModulus (x, IntegValue n), IntegValue m)
    | Axiom60 x = x;

  (* Axiom61: ~(abs(n)).>=.m & n.!=.0 ==> (x mod n) div m = 0  *)
 
  fun Axiom61 (IntegDivide (IntegModulus (x, IntegValue n), IntegValue m)) =
        if (~(abs n) >= m) andalso (not (n = 0))
          then IntegValue ~1
          else IntegDivide (IntegModulus (x, IntegValue n), IntegValue m)
    | Axiom61 x = x;

  (* Axiom62: x.!=.0 ==> 0 mod x = 0  *)
 
  fun Axiom62 (IntegModulus (IntegValue 0, x)) =
        if not (x is_equal_to (IntegValue 0))
          then IntegValue 0
          else IntegModulus (IntegValue 0, x)
    | Axiom62 x = x;



(*** Exponential axioms ***)
 
  (* Axiom70: x .!=. 0  ==> x ^ 0 = 1 *)

  fun Axiom70 (IntegPower (x, IntegValue 0)) =
        if not (x is_equal_to IntegValue 0)
          then IntegValue 1
          else IntegPower (x, IntegValue 0)
    | Axiom70 x = x;
 
  (* Axiom71: x ^ 1 = x *)

  fun Axiom71 (IntegPower (x, IntegValue 1)) =
        IntegValue 1
    | Axiom71 x = x;
 
  (* Axiom72: 1 ^ x = 1 *)

  fun Axiom72 (IntegPower (IntegValue 1, x)) =
        IntegValue 1
    | Axiom72 x = x;
 
  (* Axiom73: n .<. 0  &  x .!=. 0  ==>  x ^ n = 0 *)

  fun Axiom73 (IntegPower (x, IntegValue n)) =
        if (n < 0) andalso (not (x is_equal_to (IntegValue 0)))
          then IntegValue 0
          else IntegPower (x, IntegValue n)
    | Axiom73 x = x;
 
  (* Axiom74: x .=. y  ==>  x ^ z * y ^ z' = x ^ (z + z') *)

  fun Axiom74 (IntegTimes (IntegPower (x, z), IntegPower (y, z'))) =
        if (x is_equal_to y)
          then IntegPower (x, IntegPlus (z, z'))
          else IntegTimes (IntegPower (x, z), IntegPower (y, z'))
    | Axiom74 x = x;
 
  (* Axiom75: x .=. y  ==>  z ^ x * z' ^ y = (z * z') ^ x *)

  fun Axiom75 (IntegTimes (IntegPower (z, x), IntegPower (z', y))) =
        if (x is_equal_to y)
          then IntegPower (IntegTimes (z, z'), x)
          else IntegTimes (IntegPower (z, x), IntegPower (z', y))
    | Axiom75 x = x;
 
  (* Axiom76: x .=. y  ==>  x ^ z / y ^ z' = x ^ (z + (- z')) *)

  fun Axiom76 (IntegDivide (IntegPower (x, z), IntegPower (y, z'))) =
        if (x is_equal_to y)
          then IntegPower (x, IntegPlus (z, IntegComp z'))
          else IntegDivide (IntegPower (x, z), IntegPower (y, z'))
    | Axiom76 x = x;
 
  (* Axiom77: x .=. y  &  z' .!=. 0  ==>  z ^ x / z' ^ y = (z / z') ^ x *)

  fun Axiom77 (IntegDivide (IntegPower (z, x), IntegPower (z', y))) =
        if (x is_equal_to y) andalso (not (z' is_equal_to (IntegValue 0)))
          then IntegPower (IntegDivide (z, z'), x)
          else IntegDivide (IntegPower (z, x), IntegPower (z', y))
    | Axiom77 x = x;
 


(*** Assoc., Comm., Distr. axioms ***)
 
  (* Axiom100: (x + y) + z =  x + (y + z) *)
 
  fun Axiom100 (IntegPlus (IntegPlus (x, y), z)) =
        IntegPlus (x, IntegPlus (y, z))
    | Axiom100 x = x;
 
  (* Axiom100a: x + (y + z) =  (x + y) + z *)
 
  fun Axiom100a (IntegPlus (x, IntegPlus (y, z))) =
        IntegPlus (IntegPlus (x, y), z)
    | Axiom100a x = x;
 
  (* Axiom101: x * (y * z) = (x * y) * z *)
 
  fun Axiom101 (IntegTimes (x, IntegTimes (y, z))) =
        IntegTimes (IntegTimes (x, y), z)
    | Axiom101 x = x;
 
  (* Axiom102: x .<. y ==> x + y = y + x *)
 
  fun Axiom102 (IntegPlus (x, y)) =
        (case y of
             IntegPlus (_, _) => IntegPlus (x, y)
           | _ => if x is_right_of y
                    then IntegPlus (y, x)
                    else IntegPlus (x, y))
    | Axiom102 x = x;
 
  (* Axiom103: x .<. y ==> x + (y + z) = y + (x + z) *)
 
  fun Axiom103 (IntegPlus (x, IntegPlus (y, z)): IntegTerm) =
        (case y of
             IntegPlus (_, _) => IntegPlus (x, IntegPlus (y, z))
           | _ => if x is_right_of y
                    then IntegPlus (y, IntegPlus (x, z))
                    else IntegPlus (x, IntegPlus (y, z)))
    | Axiom103 x = x;
 
  (* Axiom104: x .>. y ==> x * y = x * y  *)
 
  fun Axiom104 (IntegTimes (x, y): IntegTerm) =
        (case x of
             IntegTimes (_, _) => IntegTimes (x, y)
           | _ => if y is_right_of x
                    then IntegTimes (y, x)
                    else IntegTimes (x, y))
    | Axiom104 x = x;
 
  (* Axiom105: x .>. y ==> x * (y * z) = y * (x * z)  *)
 
  fun Axiom105 (IntegTimes (x, IntegTimes (y, z)): IntegTerm) =
        (case x of
             IntegTimes (_, _) => IntegTimes (x, y)
           | _ => if x is_right_of y
                    then IntegTimes (y, IntegTimes (x, z))
                    else IntegTimes (x, IntegTimes (y, z)))
    | Axiom105 x = x;
 
  (* Axiom106: x * (y + z) = x * y + x * z  *)
 
  fun Axiom106 (IntegTimes (x, IntegPlus (y, z)): IntegTerm) =
        IntegPlus (IntegTimes (x, y), IntegTimes (x, z))
    | Axiom106 x = x;
 
  (* Axiom107: (y + z) * x = y * x + z * x  *)
 
  fun Axiom107 (IntegTimes (IntegPlus (y, z), x): IntegTerm) =
        IntegPlus (IntegTimes (y, x), IntegTimes (z, x))
    | Axiom107 x = x;

  (* Axiom108: - (x + y) = (- x) + (- y) *)
 
  fun Axiom108 (IntegComp (IntegPlus (x, y))) =
        IntegPlus (IntegComp x, IntegComp y)
    | Axiom108 x = x;
 
  (* Axiom109: - (x * y) = (- x) * y  *)
 
  fun Axiom109 (IntegComp (IntegTimes (x, y))) =
        IntegTimes (IntegComp x, y)
    | Axiom109 x = x;
 
  (* Axiom110: - (x / y) = (- x) / y  *)
 
  fun Axiom110 (IntegComp (IntegDivide (x, y))) =
        IntegDivide (IntegComp x, y)
    | Axiom110 x = x;

  (* Axiom111: - (x mod y) = (- x) mod (- y)  *)
 
  fun Axiom111 (IntegComp (IntegModulus (x, y))) =
        IntegModulus (IntegComp x, IntegComp y)
    | Axiom111 x = x;

  (* Axiom112: (- x) * y = x * (- y)  *)
 
  fun Axiom112 (IntegTimes (IntegComp x, y)) =
        IntegTimes (x, IntegComp y)
    | Axiom112 x = x;

  (* Axiom113: x / (- y) = (- x) / y  *)
 
  fun Axiom113 (IntegDivide (x, IntegComp y)) =
        IntegDivide (IntegComp x, y)
    | Axiom113 x = x;

  (* Axiom114: (z ^ x) ^ y = z ^ (x * y)  *)

  fun Axiom114 (IntegPower (z, IntegPower (x, y))) =
        IntegPower (z, IntegTimes (x, y))
    | Axiom114 x = x;
 

(*** Convertion axioms ***)

  fun AxiomA (IntegTimes (x, y)) =
        (case x of
             IntegValue _ => IntegTimes (x, y)
           | _            => IntegConstant (("*", 1), [x, y]))
    | AxiomA x = x;
 
  fun AxiomB (IntegDivide (x, y)) =
        IntegConstant (("/", 1), [x, y])
    | AxiomB x = x;
 
  fun AxiomC (IntegModulus (x, y)) =
        IntegConstant (("%", 1), [x, y])
    | AxiomC x = x;
 
  fun AxiomD (IntegPower (x, y)) =
        IntegConstant (("^", 1), [x, y])
    | AxiomD x = x;
 

(*** Predicate Elimination axiom ***)

  (* PredElimAxiom: P(X) = (f(X) = 0)  *)

  fun PredElimAxiom (BoolPred (nm, args)) =
        BoolEq (IntegConstant ((nm, 2), args), IntegValue 0)
    | PredElimAxiom x = x;
 

(*** Inequality axioms ***)

  (* AxiomE1: (x = y) = (x <= y) & (y <= x)  *)

  fun AxiomE1 (BoolEq (x, y)) =
        BoolAnd (BoolLEq (x, y), BoolLEq (y, x))
    | AxiomE1 x = x;
 
  (* AxiomE2: (x >= y) = (y <= x)  *)

  fun AxiomE2 (BoolGEq (x, y)) =
        BoolLEq (y, x)
    | AxiomE2 x = x;
 
  (* AxiomE3: (x < y) = ((x + 1) <= y)  *)

  fun AxiomE3 (BoolLess (x, y)) =
        BoolLEq (IntegPlus (x, IntegValue 1), y)
    | AxiomE3 x = x;
 
  (* AxiomE4: (x > y) = ((y + 1) <= x)  *)

  fun AxiomE4 (BoolGreater (x, y)) =
        BoolLEq (IntegPlus (y, IntegValue 1), x)
    | AxiomE4 x = x;
 

(*** And/Or axioms ***)

  (* AxiomF1: x & T  =  x  *)

  fun AxiomF1 (BoolAnd (x, BoolTrue)) =
        x
    | AxiomF1 x = x;
 
  (* AxiomF2: T & x  =  x  *)

  fun AxiomF2 (BoolAnd (BoolTrue, x)) =
        x
    | AxiomF2 x = x;
 
  (* AxiomF3: x & F  =  F  *)

  fun AxiomF3 (BoolAnd (x, BoolFalse)) =
        BoolFalse
    | AxiomF3 x = x;
 
  (* AxiomF4: F & x  =  F  *)

  fun AxiomF4 (BoolAnd (BoolFalse, x)) =
        BoolFalse
    | AxiomF4 x = x;
 
  (* AxiomF5: x | F  =  x  *)

  fun AxiomF5 (BoolOr (x, BoolFalse)) =
        x
    | AxiomF5 x = x;
 
  (* AxiomF6: F | x  =  x  *)

  fun AxiomF6 (BoolOr (BoolFalse, x)) =
        x
    | AxiomF6 x = x;
 
  (* AxiomF7: x | T  =  T  *)

  fun AxiomF7 (BoolOr (x, BoolTrue)) =
        BoolTrue
    | AxiomF7 x = x;
 
  (* AxiomF8: T | x  =  T  *)

  fun AxiomF8 (BoolOr (BoolTrue, x)) =
        BoolTrue
    | AxiomF8 x = x;
 

(*** Imp axioms ***)

  (* AxiomF10: x -> y  =  !x | y  *)

  fun AxiomF10 (BoolImp (x, y)) =
        BoolOr (BoolNot x, y)
    | AxiomF10 x = x;
   
  (*** Not axioms ***)

  (* AxiomF20: !!x  =  x  *)

  fun AxiomF20 (BoolNot (BoolNot x)) =
        x
    | AxiomF20 x = x;
   
  (* AxiomF21: !(x <= y)  =  ((y + 1) <= x)  *)

  fun AxiomF21 (BoolNot (BoolLEq (x, y))) =
        BoolLEq (IntegPlus (y, IntegValue 1), x)
    | AxiomF21 x = x;
   

(*** AD axioms ***)

  (* AxiomF100: x & (y & z)  =  (x & y) & z  *)

  fun AxiomF100 (BoolAnd (x, BoolAnd (y, z))) =
        BoolAnd (BoolAnd (x, y), z)
    | AxiomF100 x = x;
   
  (* AxiomF101: x | (y | z)  =  (x | y) | z  *)

  fun AxiomF101 (BoolOr (x, BoolOr (y, z))) =
        BoolOr (BoolOr (x, y), z)
    | AxiomF101 x = x;

  (* AxiomF102: x & (y | z)  =  (x & y) | (x & z)  *)

  fun AxiomF102 (BoolAnd (x, BoolOr (y, z))) =
        BoolOr (BoolAnd (x, y), BoolAnd (x, z))
    | AxiomF102 x = x;

  (* AxiomF103: (y | z) & x  =  (y & x) | (z & x)  *)

  fun AxiomF103 (BoolAnd (BoolOr (y, z), x)) =
        BoolOr (BoolAnd (y, x), BoolAnd (z, x))
    | AxiomF103 x = x;

  (* AxiomF104: !(x & y)  =  !x | !y  *)

  fun AxiomF104 (BoolNot (BoolAnd (x, y))) =
        BoolOr (BoolNot x, BoolNot y)
    | AxiomF104 x = x;

  (* AxiomF105: !(x | y)  =  !x & !y  *)

  fun AxiomF105 (BoolNot (BoolOr (x, y))) =
        BoolAnd (BoolNot x, BoolNot y)
    | AxiomF105 x = x;

  (* AxiomF106: abs(n) < abs(m) ==> n <= (y + m)  = 0 <= (y + (m-n))
                     n = m      ==> n <= (y + m)  = 0 <= y
                                                   else  (n-m) <= y  *)

  fun AxiomF106 (BoolLEq (IntegValue n, IntegPlus (y, IntegValue m))) =
        if n = m
          then BoolLEq (IntegValue 0, y)
          else (if (abs n) < (abs m)
                  then BoolLEq (IntegValue 0,IntegPlus (y, IntegValue (m - n)))
                  else BoolLEq (IntegValue (n - m), y))
    | AxiomF106 x = x;

  (* AxiomF107: abs(n) < abs(m) ==> (x + n) <= m  = x <= (m-n)
                     n = m      ==> (x + n) <= m  = x <= 0
                                                   else  (x + (n-m)) <= 0  *)

  fun AxiomF107 (BoolLEq (IntegPlus (x, IntegValue n), IntegValue m)) =
        if n = m
          then BoolLEq (x, IntegValue 0)
          else (if (abs n) < (abs m)
                  then BoolLEq (x, IntegValue (m - n))
                  else BoolLEq (IntegPlus (x, IntegValue (n - m)),
                                IntegValue 0))
    | AxiomF107 x = x;

  (* AxiomF108: abs(n) < abs(m) ==> (x + n) <= (y + m)  = x <= (y + (m-n))
                     n = m      ==> (x + n) <= (y + m)  = x <= y
                                                   else  (x + (n-m)) <= y  *)

  fun AxiomF108 (BoolLEq (IntegPlus (x, IntegValue n),
                          IntegPlus (y, IntegValue m))) =
        if n = m
          then BoolLEq (x, y)
          else (if (abs n) < (abs m)
                  then BoolLEq (x, IntegPlus (y, IntegValue (m - n)))
                  else BoolLEq (IntegPlus (x, IntegValue (n - m)), y))
    | AxiomF108 x = x;



(*** end of axioms ***)


  val CalcAxioms = Axiom0;

  val AddAxioms = Axiom17
                o Axiom16
                o Axiom15
                o Axiom14
                o Axiom13
                o Axiom12
                o Axiom11
                o Axiom10
                o Axiom9
                o Axiom8
                o Axiom7
                o Axiom6
                o Axiom5
                o Axiom4
                o Axiom3
                o Axiom2
                o Axiom1;

  val CompAxioms = Axiom20
                 o Axiom21;

  val MultAxioms = Axiom35
                 o Axiom34
                 o Axiom33
                 o Axiom32
                 o Axiom31
                 o Axiom30;

  val DivAxioms = Axiom48
                o Axiom47
                o Axiom46
                o Axiom45
                o Axiom44
                o Axiom43
                o Axiom42
                o Axiom41
                o Axiom40;

  val ModAxioms = Axiom62
                o Axiom61
                o Axiom60
                o Axiom59
                o Axiom58
                o Axiom57
                o Axiom56
                o Axiom55
                o Axiom54
                o Axiom53
                o Axiom52
                o Axiom51
                o Axiom50;

  val ExpAxioms = Axiom77
                o Axiom76
                o Axiom75
                o Axiom74
                o Axiom73
                o Axiom72
                o Axiom71
                o Axiom70;

  val ACDAxioms = Axiom114
                o Axiom113
                o Axiom112
                o Axiom111
                o Axiom110
                o Axiom109
                o Axiom108
                o Axiom107
                o Axiom106
                o Axiom105
                o Axiom104
                o Axiom103
                o Axiom102
                o Axiom101
                o Axiom100;

  val IntegAxioms = ACDAxioms
                  o ExpAxioms
                  o ModAxioms
                  o DivAxioms
                  o MultAxioms
                  o CompAxioms
                  o AddAxioms
                  o CalcAxioms;

  val ConvAxioms = AxiomA
                 o AxiomB
                 o AxiomC
                 o AxiomD;

  val IncAxioms = CalcAxioms
                o Axiom100;

  val IneqsAxioms = AxiomE4
                  o AxiomE3
                  o AxiomE2
                  o AxiomE1;

  val AndOrAxioms = AxiomF8
                  o AxiomF7
                  o AxiomF6
                  o AxiomF5
                  o AxiomF4
                  o AxiomF3
                  o AxiomF2
                  o AxiomF1;

  val ImpAxioms = AxiomF10;

  val NotAxioms = AxiomF20
                o AxiomF21;

  val ADAxioms = AxiomF105
               o AxiomF104
               o AxiomF103
               o AxiomF102
               o AxiomF101
               o AxiomF100;

  val DNFAxioms = AndOrAxioms
                o ImpAxioms
                o NotAxioms
                o ADAxioms;

  val LAssocAxiom = Axiom100a;

  val IneqSimpAxioms = AxiomF106
                     o AxiomF107
                     o AxiomF108;


(* rewrite terms top-down using rewrite axioms for BoolTerms and IntegTerms
   provided as arguments.
   type (BoolTerm -> BoolTerm) -> (IntegTerm -> IntegTerm) -> IntwgTerm ->
           IntegTerm  *)

  fun pa_rew bool_ax int_ax term =
    let
      fun pa_int_rew x =
        let
          val x_rew = int_ax x;
        in
          if x = x_rew
            then
              case x of
                  IntegConstant (nmi, args) =>
                    IntegConstant (nmi, map pa_int_rew args)
                | IntegValue arg =>
                    IntegValue arg
                | IntegComp arg =>
                    IntegComp (pa_int_rew arg)
                | IntegPlus (arg1, arg2) =>
                    IntegPlus (pa_int_rew arg1, pa_int_rew arg2)
                | IntegMinus (arg1, arg2) =>
                    IntegMinus (pa_int_rew arg1, pa_int_rew arg2)
                | IntegTimes (arg1, arg2) =>
                    IntegTimes (pa_int_rew arg1, pa_int_rew arg2)
                | IntegDivide (arg1, arg2) =>
                    IntegDivide (pa_int_rew arg1, pa_int_rew arg2)
                | IntegModulus (arg1, arg2) =>
                    IntegModulus (pa_int_rew arg1, pa_int_rew arg2)
                | IntegPower (arg1, arg2) =>
                    IntegPower (pa_int_rew arg1, pa_int_rew arg2)
            else
              pa_int_rew x_rew
        end;

      fun pa_bool_rew x =
        let
          val x_rew = bool_ax x;
        in
          if x = x_rew
            then
              case x of
                  BoolTrue =>
                    BoolTrue
                | BoolFalse =>
                    BoolFalse
                | BoolPred (nm, args) =>
                    BoolPred (nm, map pa_int_rew args)
                | BoolLess (arg1, arg2) =>
                    BoolLess (pa_int_rew arg1, pa_int_rew arg2)
                | BoolLEq (arg1, arg2) =>
                    BoolLEq (pa_int_rew arg1, pa_int_rew arg2)
                | BoolEq (arg1, arg2) =>
                    BoolEq (pa_int_rew arg1, pa_int_rew arg2)
                | BoolGEq (arg1, arg2) =>
                    BoolGEq (pa_int_rew arg1, pa_int_rew arg2)
                | BoolGreater (arg1, arg2) =>
                    BoolGreater (pa_int_rew arg1, pa_int_rew arg2)
                | BoolNot arg =>
                    BoolNot (pa_bool_rew arg)
                | BoolAnd (arg1, arg2) =>
                    BoolAnd (pa_bool_rew arg1, pa_bool_rew arg2)
                | BoolOr (arg1, arg2) =>
                    BoolOr (pa_bool_rew arg1, pa_bool_rew arg2)
                | BoolImp (arg1, arg2) =>
                    BoolImp (pa_bool_rew arg1, pa_bool_rew arg2)
          else
            pa_bool_rew x_rew
        end;
    in
      pa_bool_rew term
    end;

in

  (* integer arithmetical simplifications *)

  fun pa_simplify x =
    let
      val x_rew = pa_rew DummyAxiom IntegAxioms x;
    in
      if x = x_rew then x else pa_simplify x_rew
    end;

 
  (* mapping to extended Presburger arithmetic by treating
     uninterpreted arithmetical operators, like `div', `mod', and
     `exp', as constants. *)

  fun map_to_pa x =
    let
      val x_rew = pa_rew DummyAxiom ConvAxioms x;
    in
      if x = x_rew then x else map_to_pa x_rew
    end;

  (* replaces uninterpreted predicates, P(X), by
     f_P(X) = 0, where f_P is a new function symbol with
     same arity as P and X is the argument vector of P. *)

  fun eliminate_preds x =
    let
      val x_rew = pa_rew PredElimAxiom DummyAxiom x;
    in
      if x = x_rew then x else eliminate_preds x_rew
    end;


  (* expressions involving `+' or `*' as arguments in uninterpreted
     function symbols are replaced by new variables; these
     variables are constrained by the replaced expressions and
     appear as premise of the formula, e.g.,
     f(x+1) = y becomes (z = x+1) --> (f(z) = y) *)

  fun pull_args_out x = mk_imp (rev (pull_out x));


  (* rewriting all inequalities and equalities into
     less-or-equal form, e.g.,
     x > y becomes (y + 1) <= x. *)

  fun normal_ineqs x =
    let
      val x_rew = pa_rew IneqsAxioms IncAxioms x;
    in
      if x = x_rew then x else normal_ineqs x_rew
    end;


  (* maps to Disjoint Normal Form *)

  fun map_to_DNF x =
    let
      val x_rew = pa_rew DNFAxioms IncAxioms x;
    in
      if x = x_rew then x else map_to_DNF x_rew
    end;


  (* simplifies inequalities if possible *)

  fun ineq_simp x =
    let
      val x_rew = pa_rew IneqSimpAxioms LAssocAxiom x;
    in
      if x = x_rew then x else ineq_simp x_rew
    end;


end;  (* local *)

end;  (* struct *)