(* 
   file     : $CAT_DIR/SupInf/SupInf.ML
   platform : PolyML 3.2 and Isabelle94-8/HOL
   version  : 1.0b  (16.07.98)
   notes    : This file prepares the argument for the decision procedure
              and invokes it.
*)

signature SUPINF =
sig

  val SupInf: term -> term

end;


structure SupInf: SUPINF =
struct

(* this function decides the validity of an Isabelle formula in
   extended Presburger arithmetic using Shostak's method;
   type: term -> term  *)

  fun SupInf term =
    let

      (* mapping to internal language *)

      val form1 = ArithSemantics.term_to_BoolTerm  term;


      (* simplification of arithmetical expressions *)

      val form2 = PASimplifier.pa_simplify form1;


      (* mapping to extended Presburger arithmetic by treating
         uninterpreted arithmetical operators, like `div', `mod', and
         `exp', as constants. *)

      val form3 = PASimplifier.map_to_pa form2;


      (* replaces uninterpreted predicates, P(X), by
         f_P(X) = 0, where f_P is a new function symbol with
         same arity as P and X is the argument vector of P. *)

      val form4 = PASimplifier.eliminate_preds form3;


      (* expressions involving + or * as arguments in uninterpreted
         function symbols are replaced by new variables; these
         variables are constrained by the replaced expressions and
         appear as premise of the formula, e.g.,
         f(x+1) = y becomes (z = x+1) --> (f(z) = y) *)

      val form5 = PASimplifier.pull_args_out form4;


      (* formula is negated *)

      val form6 = ArithSemantics.BoolNot form5;


      (* rewriting all inequalities and equalities into
         less-or-equal form, e.g.,
         x > y becomes (y + 1) <= x. *)

      val form7 = PASimplifier.normal_ineqs form6;


      (* maps to Disjoint Normal Form *)

      val form8 = PASimplifier.map_to_DNF form7;


      (* simplifies inequalities if possible *)

      val form9 = PASimplifier.ineq_simp form8;


      (* maps DNF formula to integer linear programming problems, where
         uninterpreted function expressions are treated as constants,
         e.g., f(g,8) becomes a constant with the name "f(g,8)". *)

      val form10 = ILPSemantics.DNF_to_ILPs form9;

    in
      (* tries to solve linear programming problem for negated formula,
         if there is a solution and no violations of substitutivity
         arise, the formula is not valid, otherwise it is. *)

      if ExtILP.solve form10
        then Const ("False", Type ("bool", []))
        else Const ("True",  Type ("bool", []))
    end;

end;  (* struct *)

